# include <stdio.h>
# include <math.h>

# include "apm.h"
# include "apmUtil.h"
# include "converse.h"
# include "pi.h"

	/*  PostScript parameters for the drawing */
# define   MAYBE_GREY    1.0
# define   NOTOR_GREY    0.5
# define   SYMMTRC_GREY  0.8
# define   LINE_WIDTH	 0.25
# define   FONT_NAME	 "Times-Italic"
# define   TYPE_SIZE	 12

	/* Size of the drawing and margins in inches */
# define   TOR_SIZE	 6.0
# define   LEFT_MARGIN	 1.5
# define   BOTTOM_MARGIN 3.0

# define   VZERO_OFFSET  (STAID_LEN + (MAT_DIM+1) * DEG_FREE + N_PARMS )
# define   VONE_OFFSET   (VZERO_OFFSET + MAT_DIM + 1)


# define   TOR_SCL	   (72 * TOR_SIZE )
# define   X_OFF	   (72 * LEFT_MARGIN) 
# define   Y_OFF	   (72 * BOTTOM_MARGIN)
# define   x_coord( dx )   (( dx * TOR_SCL / TWO_PI) + X_OFF )
# define   y_coord( dy )   (( dy * TOR_SCL / TWO_PI) + Y_OFF )

/* Variables shared via converse.h */
char *graf_file = DFLT_GRAF_FILE ;

/* Functions private to this file */
static void fprintHeader(FILE *gfp);

/* Variables private to this file */
static APM    gScratch ;
static FILE   *ps_file ;
/* +++++++++++++++++++++++++++ */

void 
startGraphics (void)
{
    ps_file = fopen( graf_file, "w" ) ;
    if( ps_file == NULL ) {
	fprintf( stderr, 
		"Trouble opening the PostScript file, %s.  Pressing on. \n",
							  graf_file ) ;
	do_graph = NO ;
    }
    else
	fprintHeader( ps_file ) ;

    gScratch = apmNew( BASE ) ;
}
/* +++++++++++++++++++++++++++ */

void 
restartGraphics (void)
{
    ps_file = fopen( graf_file, "a" ) ;
    if( ps_file == NULL ) {
	fprintf( stderr, 
		"Trouble reopening the PostScript file, %s.  Pressing on. \n",
							  graf_file ) ;
	do_graph = NO ;
    }

    gScratch = apmNew( BASE ) ;
}
/* ++++++++++++++++++++++++++++++ */

void 
finishGraphics (void)
{
    fprintf( ps_file, "showpage\n" ) ;
    fclose( ps_file ) ;
}
/* ++++++++++++++++++++++++++ */

void 
graphPrism (RPrism *rpt, int style)
{
    double   xll, yll, xur, yur ;

    apmAdd( gScratch, rpt->center->z.v[0], rpt->matrix[VZERO_OFFSET] ) ; 
    xur = apmtodbl( gScratch ) ;

    apmSubtract( gScratch, rpt->center->z.v[0], rpt->matrix[VZERO_OFFSET] ) ; 
    xll = apmtodbl( gScratch ) ;

    apmAdd( gScratch, rpt->center->z.v[1], rpt->matrix[VONE_OFFSET] ) ; 
    yur = apmtodbl( gScratch ) ;

    apmSubtract( gScratch, rpt->center->z.v[1], rpt->matrix[VONE_OFFSET] ) ; 
    yll = apmtodbl( gScratch ) ;

    switch( style ) {
	case  NO_TORI :
	    fprintf( ps_file, 
			 "%.3f %.3f %.3f %.3f %.3f shadedRectangle \n", 
					x_coord( xll ), y_coord( yll ),
					x_coord( xur ), y_coord( yur ),
							  NOTOR_GREY ) ;
	    break ;

	case  MAYBE :
	    fprintf( ps_file, 
			 "%.3f %.3f %.3f %.3f %.3f shadedRectangle \n",
					x_coord( xll ), y_coord( yll ),
					x_coord( xur ), y_coord( yur ),
							  MAYBE_GREY ) ;
	    break ;

	case  ACTIVE :
	    break ;

	case  UNTRIED :
	    break ;

	case  SYMMTRC :
	    fprintf( ps_file, 
			 "%.3f %.3f %.3f %.3f %.3f shadedRectangle \n",
					x_coord( xll ), y_coord( yll ),
					x_coord( xur ), y_coord( yur ),
							SYMMTRC_GREY ) ;
	    break ;

	default :
	    break ;
    }

    fflush( ps_file ) ;
}
/* ++++++++++++++++++++++++++++++++++ */

void 
fprintHeader (FILE *gfp)
{
/*
		Print things to set up the graphics state.
*/
    fprintf( gfp, "%%!PS-Adobe-1.0 \n" ) ;
    fprintf( gfp,"%%%%Title: Converse KAM output header \n" ) ;
    fprintf( gfp,"%%%%EndProlog \n" ) ;
    fprintf( gfp,"%%%%Page : 1 1 \n" ) ;
    fprintf( gfp,"\ninitgraphics \n" ) ;
    fprintf( gfp,"%.3f setlinewidth \n\n", LINE_WIDTH ) ;

/*
	Define some useful procedures 
*/
	/*  An argument snarfing procedure */
    fprintf( gfp, "%%  A procedure to snarf an argument off the stack \n" ) ;
    fprintf( gfp,"/getarg { exch \n\t   def } def \n" ) ;

	/*   A rectangle drawing procedure  */
    fprintf( gfp, "%% \n" ) ;
    fprintf( gfp,"%% A procedure to draw a grey filled rectangle \n" ) ;
    fprintf( gfp,"%% with a boundary.  Its lower-left corner is at \n" ) ;
    fprintf( gfp,"%% (xll, yll), its upper right is at (xur, yur). \n" ) ;
    fprintf( gfp,"%% The inside is shaded the color given by grey_val. \n" ) ;
    fprintf( gfp,"%% \n" ) ;
    fprintf( gfp,"/shadedRectangle { \n" ) ;
    fprintf( gfp,"    gsave \n" ) ;
    fprintf( gfp,"\t/grey_val   getarg \n" ) ;
    fprintf( gfp,"\t    /yur    getarg \n" ) ;
    fprintf( gfp,"\t    /xur    getarg \n" ) ;
    fprintf( gfp,"\t    /yll    getarg \n" ) ;
    fprintf( gfp,"\t    /xll    getarg \n\n" ) ;

    fprintf( gfp,"\txll  yll    moveto \n" ) ;
    fprintf( gfp,"\txur  yll    lineto \n" ) ;
    fprintf( gfp,"\txur  yur    lineto \n" ) ;
    fprintf( gfp,"\txll  yur    lineto \n" ) ;
    fprintf( gfp,"\tclosepath \n\n" ) ;

    fprintf( gfp,"\tgsave\n" ) ;
    fprintf( gfp,"\t    grey_val \tsetgray \n" ) ;
    fprintf( gfp,"\t\t\tfill \n" ) ;
    fprintf( gfp,"\tgrestore \n\n" ) ;

    fprintf( gfp,"\t0.0\t\tsetgray \n" ) ;
    fprintf( gfp,"\t\t\tstroke \n" )  ;
    fprintf( gfp,"    grestore\n" ) ;
    fprintf( gfp,"} def\n\n" ) ;

    fflush( gfp ) ;
}
