# include <stdio.h>
# include <strings.h>
# include <stdlib.h>
# include <ctype.h>
# include <math.h>

# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"

# define  DEBUG	 	NO

# define  BACK_COUNT	5 
# define  GBUF_SZ		81
# define  BIG_BUF		256

# define  END_MARK	"#-30-#"

/* For fseek() */
# define  FROM_START	0
# define  FROM_HERE		1
# define  FROM_END		2

/* Variables shared via converse.h */
char	*back_name, *rest_name, *parm_names[] = { "a", "b", "c" } ;

/* Functions private to this file */
static long write_hdr(RPrism *priz);
static void note_options(void);
static void note_ephemera(void);
static void restore_ephemera(void);
static void note_performance(void);
static void note_edge(void);
static void backupRPrism(RPrism *priz);
static void translate_cuts(RPrism *priz, RPrism *base);
static char *get_cuts(RPrism *priz);
static char *get_good_line(void);
static void skip_fluff(void);

/* Variables private to this file */
static int		count ;
static APM		back_scratch, Rcenter, Rwidth ;
static char		graf_buf[GBUF_SZ], line_buf[BIG_BUF] ;
static long		end_o_hdr ;
static FILE		*back_file, *rest_file ;
static RPrism 	*start ;
static char		*stars = "*************************************" ;
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
startBackup (RPrism *priz)
{
    if( restoration == YES )
	priz = start ;

    back_scratch = apmNew( BASE ) ;
    if( do_backup == YES ) {
	back_file = fopen( back_name, "w+" ) ;
	if( back_file == NULL ) {
	    fprintf( stderr, 
		"Unable to open %s for writing, cannot do backups. \n", 
								back_name ) ;
	    do_backup = NO ;
	}

	count = 0 ;
	end_o_hdr = write_hdr( priz ) ;
    }	
}
/* ++++++++++++++++++++++++++++++ */

long 
write_hdr (RPrism *priz)
{
    int	 j, vj ;

    note_options() ;
    fprintf( back_file, "\n" ) ;

    for( j=0 ; j < N_PARMS ; j++ ) 
	fprintf( back_file, "%.14e \t %.14e \n", 
			    apmtodbl( priz->center->p[j] ), 
			    apmtodbl( priz->matrix[j*(MAT_DIM + 1)] ) ) ;

    fprintf( back_file, "\n" ) ;
    for( j=0 ; j < DEG_FREE ; j++ )  {
	apmDivide( back_scratch, precision, (APM) NULL, 
					    priz->center->z.v[j], pi ) ;
	fprintf( back_file, "%.14e \t ",apmtodbl( back_scratch ) ) ;
	vj = j + DEG_FREE + N_PARMS ;
	apmDivide( back_scratch, precision, (APM) NULL, 
			priz->matrix[vj * (MAT_DIM+1)], pi ) ;
	fprintf( back_file, "%.14e \n",apmtodbl( back_scratch ) ) ;
    }

    fprintf( back_file, "%s\n\n", stars ) ;
    fflush( back_file ) ;

    return( ftell(back_file) ) ;
}
/* ++++++++++++++++++++++++++++++ */

void 
note_options (void)
{
/*
	Record the various options on the first few lines of 
	the header.
*/
	/* graphics stuff */
    fprintf( back_file, "%d", do_graph ) ;
    if( do_graph == YES ) 
	fprintf( back_file, "\t%s\n", graf_file ) ;
    else 
	fprintf( back_file, "\n" ) ;

	/* depth, precision, furthest, terseness, and stubbornness */
    fprintf( back_file, "%d \t %d \t %d \t %d \t %d \n", depth,
				(precision - SAFETY_PRECIS),
						   furthest,
						   terse,
						   stubborn ) ;
}
/* ++++++++++++++++++++++++ */

void 
make_backup (RPrism *priz)
{
    RPrism  *pz_pt ;

    if( ++count < BACK_COUNT )
       return ;
    else {
	count = 0 ;

	fseek( back_file, end_o_hdr, FROM_START ) ;
	note_ephemera() ;
	for( pz_pt = priz ; pz_pt != NULL ; pz_pt = pz_pt->next ) 
	    backupRPrism( pz_pt ) ;

	fprintf( back_file, "%s\n", END_MARK ) ;
	fflush( back_file ) ;
# if DEBUG
	cease() ;
# endif
    }
}
/* +++++++++++++++++++++++++++ */

void 
note_ephemera (void)
{
    note_edge() ;
    note_performance() ;
}
/* +++++++++++++++++++++++++++ */

void 
restore_ephemera (void)
{
    restore_edge() ;
    restore_performance() ;
}
/* +++++++++++++++++++++++++++ */

void 
note_performance (void)
{
    fprintf( back_file, "%d \n", most_cuts ) ;
    fprintf( back_file, "%d \t %d \t %d \n", tries, quick_tries, Rtries ) ;
    fprintf( back_file, "%d \t %d \n", max_steps, max_NTsteps ) ;
    fprintf( back_file, "%d \t %d \n", HermSuccess, LLSuccess ) ;
    fprintf( back_file, "%d \t %d \t %d \n", ll_used[0], ll_used[1], 
							 ll_used[2] ) ;
    fprintf( back_file, "%s\n\n", stars ) ;
}
/* ++++++++++++++++++++++++++ */

void 
backupRPrism (RPrism *priz)
{
    int  j ;

    fprintf( back_file, "%d  ", priz->n_cuts ) ;
    for( j=0 ; j < MAT_DIM ; j++ )
	fprintf( back_file, "%s ", priz->cuts[j] ) ;

    fprintf( back_file, "\n" ) ;
}
/* +++++++++++++++++++++++++ */

void 
restore_options (void)
{
/*
	Find the various options on the first few lines of 
	the header.
*/
	/* graphics stuff */
    fscanf( rest_file, "%d", &do_graph ) ;
    if( do_graph == YES ) {
		fscanf( rest_file, "\t%s\n", graf_buf ) ;
		graf_file = graf_buf ;
    }


	/* depth, precision, furthest, terseness, and stubbornness */
    fscanf( rest_file, "%d \t %d \t %d \t %d \t %d \n", 
						  &depth,
						  &precision,
						  &furthest,
						  &terse,
						  &stubborn ) ;
	/*
		Set the working precision and
		the rigorous precision - the former is
		SAFETY_PRECIS + the users precision,
		the latter is the user's precision
	*/
    apmAssignLong( max_error, 1L, -precision, BASE ) ;
    precision += SAFETY_PRECIS ;
# if DEBUG
    printf( "restored the options \n" ) ;
# endif
}
/* +++++++++++++++++++++++ */

void 
restore_start (void)
{
    int	     j, vj ;
    double   center, width ;

    start = conjureRPrism() ;

    for( j=0 ; j < N_PARMS ; j++ )  {
	fscanf( rest_file, "%lf \t %lf \n", &center, &width ) ;
	dbltoapm( start->center->p[j], BASE, center )  ;
	dbltoapm( start->matrix[j*(MAT_DIM + 1)], BASE, width ) ;
    }

    for( j=0 ; j < DEG_FREE ; j++ )  {
	fscanf( rest_file, "%lf \t %lf", &center, &width ) ;
	dbltoapm( back_scratch, BASE, center ) ;
	apmMultiply( start->center->z.v[j], back_scratch, pi ) ;

	vj = j + DEG_FREE + N_PARMS ;	
	dbltoapm( back_scratch, BASE, width ) ;
	apmMultiply( start->matrix[vj * (MAT_DIM+1)], back_scratch, pi ) ;
    }

    echo_parameters( start ) ;
# if DEBUG
    printf( "restored the initial prism \n" ) ;
# endif
}
/* ++++++++++++++++++++++ */

void 
restore_chain (RPrism *first)
{
    int	    length ;
    RPrism  *next, *link ;

    for( next=first, length=0 ; get_cuts(next) != NULL ; length++ ) {  
	link = next ;
	translate_cuts( link, start ) ;

	next = conjureRPrism() ;
	link->next = next ;
    }

    if( length > 0 ) {
	releaseRPrism( next ) ;
	link->next = NULL ;
    }
    else {
	fprintf( stderr, "Cannot restore any prisms. Died. \n" ) ;
	exit(0) ;
    }
# if DEBUG
    startBackup( first ) ;
    count = BACK_COUNT ; 
    make_backup( first ) ;
# endif
}
/* ++++++++++++++++++++++++++++ */

void 
translate_cuts (RPrism *priz, RPrism *base)
{
    int	   j, vj ;
    char   *cpt ;

    for( j=0 ; j < N_PARMS ; j++ ) {
	apmAssign( Rcenter, base->center->p[j] ) ;
	apmAssign( Rwidth, base->matrix[j * (MAT_DIM + 1)] ) ;

	cpt = &priz->cuts[j][1] ;
	while( *cpt != '\0' ) {
	    apmDivide( back_scratch, precision, (APM)NULL, Rwidth, two ) ;
	    if( *cpt == 'r' ) 
		apmCalc( Rcenter, Rcenter, back_scratch, APM_ADD, NULL ) ;
	    else
		apmCalc( Rcenter, Rcenter, back_scratch, APM_SUB, NULL ) ;

	    apmAssign( Rwidth, back_scratch ) ;
	    cpt++ ;
	}

	apmAssign( priz->center->p[j], Rcenter ) ;
	apmAssign( priz->matrix[j * (MAT_DIM+1)], Rwidth ) ;
    }

    for( vj=(N_PARMS + DEG_FREE), j=0 ; j < DEG_FREE ; j++, vj++ ) {
	apmAssign( Rcenter, base->center->z.v[j] ) ;
	apmAssign( Rwidth, base->matrix[vj * (MAT_DIM+1)] ) ;

	cpt = &priz->cuts[vj][1] ;
	while( *cpt != '\0' ) {
	    apmDivide( back_scratch, precision, (APM) NULL, Rwidth, two ) ;
	    if( *cpt == 'r' ) 
		apmCalc( Rcenter, Rcenter, back_scratch, APM_ADD, NULL ) ;
	    else
		apmCalc( Rcenter, Rcenter, back_scratch, APM_SUB, NULL ) ;

	    apmAssign( Rwidth, back_scratch ) ;
	    cpt++ ;
	}

	apmAssign( priz->center->z.v[j], Rcenter ) ;
	apmAssign( priz->matrix[vj * (MAT_DIM+1)], Rwidth ) ;
    }
}
/* +++++++++++++++++++++++ */

void 
begin_restoration (void)

{
    rest_file = fopen( rest_name, "r" ) ;
    if( rest_file == NULL ) {
	fprintf( stderr, 
	    "Unable to open %s for reading, cannot do restoration. \n", 
							  rest_name ) ;
	exit(0) ;
    }

	/* Do up some APM's */
    Rcenter = apmNew( BASE ) ;
    Rwidth  = apmNew( BASE ) ;
    if( apmValidate( back_scratch ) != APM_OK )  
	back_scratch = apmNew( BASE ) ;
}
/* ++++++++++++++++++++++++++ */

void 
note_edge (void)
{    
    int	    j ;

    for( j=0 ; j < N_PARMS ; j++ ) 
	fprintf( back_file, "%.14e \t %.14e \n", parm_floor[j], parm_roof[j] ) ;
}
/* ++++++++++++++++++++++++ */

void 
restore_edge (void)
{
    int	    j ;

    skip_fluff() ;

    for( j=0 ; j < N_PARMS ; j++ ) 
		fscanf( rest_file, "%lf \t %lf \n", &parm_floor[j], &parm_roof[j] ) ;
}
/* +++++++++++++++++++++++++++ */

void 
restore_performance (void)
{
    skip_fluff() ;

    fscanf( rest_file, "%d \n", &most_cuts ) ;
    fscanf( rest_file, "%d \t %d \t %d \n", &tries, &quick_tries, &Rtries ) ;
    fscanf( rest_file, "%d \t %d \n", &max_steps, &max_NTsteps ) ;
    fscanf( rest_file, "%d \t %d \n", &HermSuccess, &LLSuccess ) ;
    fscanf( rest_file, "%d \t %d \t %d \n", &ll_used[0], &ll_used[1], 
							 &ll_used[2] ) ;
}
/* ++++++++++++++++++++++++++ */

char *
get_cuts (RPrism *priz)
{
    int	  j ;
    char  *result, *cut_str, *num_str, *strtok() ;

    result = get_good_line() ;
# if DEBUG
    printf( "get_cuts : %s \n", line_buf ) ;
# endif
    if( result != NULL ) {
		num_str = strtok( line_buf, " " ) ;
		priz->n_cuts = atoi( num_str ) ;

		for( j=0 ; j < MAT_DIM ; j++ ) {
			cut_str = strtok( NULL, " " ) ;
			strcpy( priz->cuts[j], cut_str ) ;
	# if DEBUG
			printf( "get_cuts : %s \n", cut_str ) ;
	# endif 
		}
    }

    return( result ) ;
}
/* ++++++++++++++++++++++++++++ */

char *
get_good_line (void)
{
    int   dot_count ;
    char  *cpt, *result ;

    dot_count = 0 ;
    result = line_buf ;
    while( (dot_count != MAT_DIM) && (result != NULL) ) {
		result = fgets( line_buf, BIG_BUF, rest_file ) ;
		if( strcmp(line_buf, END_MARK) == 0 ) 
			result = NULL ;

		dot_count = 0 ;
		if( result != NULL ) {
	# if DEBUG
			printf( "echo : %s \n", line_buf ) ;
	# endif
			for( cpt = line_buf ; *cpt != '\0' ; cpt++ )
			if( *cpt == '.' ) 
				dot_count++ ;
		}
    }

    return( result ) ;
}
/* ++++++++++++++++++++++++++++++++++++++++++ */

void 
skip_fluff (void)
{
    int  c ;

    while( !isdigit( (c = fgetc(rest_file)) ) ) ;
    ungetc( c, rest_file ) ;
}
