# include  <stdio.h>
# include  <math.h>

# include  "apm.h"
# include  "apmSpecial.h"
# include  "converse.h"
# include  "bounding.h"
# include  "pi.h"

# define   DEBUG   NO 

/* Variables private to this file */
static APM	half, three_halfs ;
static APM	Rdelta, Rmax_cos, Rmin_cos  ;
static APM	Rmax_x, Rmin_x, Rfloor_x, Rlft_val, Rrght_val ;

static Bdd_apm	Rnew_theta ;
/* -------------------------------- */

void initTrigBd( void )
/*
	Set up the APM's defined above.
*/
{
    Rdelta = apmNew( BASE ) ;
    Rmin_x = apmNew( BASE ) ;
    Rmax_x = apmNew( BASE ) ;
    Rfloor_x = apmNew( BASE ) ;
    Rmax_cos = apmNew( BASE ) ;
    Rmin_cos = apmNew( BASE ) ;
    Rlft_val = apmNew( BASE ) ;
    Rrght_val = apmNew( BASE ) ;

    Rnew_theta.ub = apmNew( BASE ) ;
    Rnew_theta.lb = apmNew( BASE ) ;

    half = apmInit( 2L, 0, BASE ) ;
    three_halfs = apmInit( 3L, 0, BASE ) ;
    apmCalc( half, half, APM_RECIP( precision ), NULL ) ; 
    apmCalc( three_halfs, half, three_halfs, APM_MUL, NULL ) ;
}
/* ++++++++++++++++++++++++++++++++++++++ */

void 
bd_cos ( 
/*
	Obtain bounds for the cosine function over
	a certain given range of angles.
*/
    Bdd_dbl *bound,
    Bdd_dbl *theta
)
{
    double    max_cos, min_cos  ;
    double    max_x, min_x, floor_x, lft_val, rght_val ;

/*
	Get some variables equal to theta / TWO_PI.  These will
	help decide whether the interval under consideration
	contains any extrema.
*/
    min_x = theta->lb / TWO_PI ;
    max_x = theta->ub / TWO_PI ;

    floor_x = floor( min_x ) ;
    min_x -= floor_x ;
    max_x -= floor_x ;
    if( (max_x - min_x) > 1.0 ) {
	bound->ub = 1.0 ;
	bound->lb = -1.0 ;
    }

    else {
	lft_val = cos( theta->lb ) ;
	rght_val = cos( theta->ub ) ;
	if( lft_val > rght_val ) {
	    max_cos = lft_val ;
	    min_cos = rght_val ;
	}
	else {
	    max_cos = rght_val ;
	    min_cos = lft_val ;
	}
/*
	Check for extrema.
*/
	if( max_x > 1.0 ) 
	    max_cos = 1.0 ;

	if( (max_x > 1.5) || 
	   ((min_x < 0.5) & (max_x > 0.5)) ) min_cos = -1.0 ;

	bound->ub = max_cos + DBL_ERR ;
	bound->lb = min_cos - DBL_ERR ;
    }
# if DEBUG
    printf( "bd_cos : %.5f < theta < %.5f -> ( %.5f, %.5f ) \n", 
					   theta->lb, theta->ub,
					   bound->lb, bound->ub ) ;
# endif
    return ;
}
/* +++++++++++++++++++++++++++++ */

void 
bd_sin ( 
/*
    Use the relation sin( x - HALF_PI ) = cos( x ) 
    and the function bd_cos() to obtain a bound on 
    the sines of angles lying in a given range.
*/
    Bdd_dbl *bound,
    Bdd_dbl *theta
)
{
    Bdd_dbl  new_theta ;

    new_theta.ub = theta->ub - HALF_PI ;
    new_theta.lb = theta->lb - HALF_PI ;

# if DEBUG
    printf( "bd_sin %.5f < theta < %.5f uses ", theta->lb, theta->ub ) ;
# endif
    bd_cos( bound, &new_theta ) ;
    return ;
}
/* ++++++++++++++++++++++++++++++++++++++ */

void 
Rbd_cos ( 
/*
	Obtain bounds for the cosine function over
	a certain given range of angles.
*/
    Bdd_apm *bound,
    Bdd_apm *theta
)
{
/*
	An APM partner to the function above.  The variables 
	used here are static, and are defined at the top 
	of the file.
*/
/*
	Get some variables equal to theta / TWO_PI.  These will
	help decide whether the interval under consideration
	contains any extrema.
*/
    apmDivide( Rmin_x, precision, (APM)NULL, theta->lb, two_pi ) ;  
    apmDivide( Rmax_x, precision, (APM)NULL, theta->ub, two_pi ) ;  

    apmFloor( Rfloor_x, Rmin_x, BASE ) ;
    apmCalc( Rmin_x, Rmin_x, Rfloor_x, APM_SUB, NULL ) ;
    apmCalc( Rmax_x, Rmax_x, Rfloor_x, APM_SUB, NULL ) ;
    apmSubtract( Rdelta, Rmax_x, Rmin_x ) ;
    if( apmCompare( Rdelta, one ) == 1 ) {
	apmAssign( bound->ub, one ) ;
	apmNegate( bound->lb, one ) ; 
    }

    else {
	apmCos( Rlft_val, theta->lb ) ;
	apmCos( Rrght_val, theta->ub ) ;
	if( apmCompare( Rlft_val, Rrght_val ) == 1 ) {
	    apmAssign( Rmax_cos, Rlft_val ) ;
	    apmAssign( Rmin_cos, Rrght_val ) ;
	}
	else {
	    apmAssign( Rmax_cos, Rrght_val ) ;
	    apmAssign( Rmin_cos, Rlft_val ) ;
	}
/*
	Check for extrema.
*/
	if( apmCompare( Rmax_x, one) == 1 ) 
	    apmAssign( Rmax_cos, one ) ;

	if( (apmCompare( Rmax_x, three_halfs) == 1) || 
	   ((apmCompare( Rmin_x, half) == -1) && 
	    (apmCompare( Rmax_x, half) ==  1))  ) apmNegate( Rmin_cos, one ) ;

	apmAdd( bound->ub, Rmax_cos, max_error ) ;
	apmSubtract( bound->lb, Rmin_cos, max_error ) ;
    }

    return ;
}
/* +++++++++++++++++++++++++++++ */

void 
Rbd_sin ( 
/*
    Use the relation sin( x - HALF_PI ) = cos( x ) 
    and the function bd_cos() to obtain a bound on 
    the sines of angles lying in a given range.
*/
    Bdd_apm *bound,
    Bdd_apm *theta
)
{
/*
	Rnew_theta is used here but is declared at the top of
	the file
*/
    apmSubtract( Rnew_theta.ub, theta->ub, half_pi ) ;
    apmSubtract( Rnew_theta.lb, theta->lb, half_pi ) ;

    Rbd_cos( bound, &Rnew_theta ) ;
    return ;
}
