# include  <stdio.h>
# include  <math.h>
# include  "apm.h"
# include  "converse.h"
# include  "bounding.h"

# define   DEBUG     NO

APM	Rextrema, Rextremb, Rub, Rlb ;
APM	Rprod[4], *Rlastp = (Rprod + 4) ;
double  prod[4], *lastp = (prod + 4) ;
/* ++++++++++++++++++++++++++++ */

void initBounding( void )
{
    int   j ;

    Rub = apmNew( BASE ) ;
    Rlb = apmNew( BASE ) ;

    Rextrema = apmNew( BASE ) ;
    Rextremb = apmNew( BASE ) ;

    for( j=0 ; j < 4 ; j++ )
	Rprod[j] = apmNew( BASE ) ;
}
/* +++++++++++++++++++++++++++ */

void bound_term( tpt )
/*
    Take a list of bounded factors and obtain a bound on their
    product.
*/

Bdd_term   *tpt ;
{
    double   extrema, extremb, *ppt ;
    Bdd_dbl  *facptr, **lastf, **fpt ;

# if DEBUG
    printf( "\tBounding a term. \n" ) ;
    fflush( stdout ) ;
# endif
/*
	If there is only one factor, deal with it directly.
*/
    if( tpt->nfactors == 1 ) {
	extrema = tpt->factors[0]->ub ;
	extremb = tpt->factors[0]->lb ;
    }

/*
	Handle terms with more than one factor.
	Since some of the factors may be negative we 
	can't just multiply together all the upper
	and lower bounds.
*/
    else {
	extrema = tpt->factors[0]->ub ;
	extremb = tpt->factors[0]->lb ;

	fpt = &tpt->factors[1] ;
	for( lastf = tpt->factors + tpt->nfactors ; fpt < lastf ; fpt++ ) {
	    facptr = *fpt ;
# if DEBUG
	    printf("\t\tConsidering a factor.\n" ) ;
	    fflush( stdout ) ;
# endif

	    prod[0] = facptr->ub * extrema ;
	    prod[1] = facptr->ub * extremb ;
	    prod[2] = facptr->lb * extrema ;
	    prod[3] = facptr->lb * extremb ;
# if DEBUG
	    printf( "\t\t\tevaluated the products.\n" ) ;
	    fflush( stdout ) ;
# endif
	    extrema = extremb = prod[0] ;
	    for( ppt = (prod+1) ; ppt < lastp ; ppt++ ) {
		if( *ppt > extrema )  
		    extrema = *ppt ;
		else if( *ppt < extremb )
		    extremb = *ppt ;
	    }
	}
    }

    extrema *= tpt->coef ;
    extremb *= tpt->coef ;
    if( extrema < extremb ) {
	tpt->bound.ub = extremb ;
	tpt->bound.lb = extrema ;
    }
    else {
	tpt->bound.ub = extrema ;
	tpt->bound.lb = extremb ;
    }
}
/* ++++++++++++++++++++++++++++++++++++ */

void bound_expr( ept ) 
/*
    Obtain bounds on the terms in a bounded expression, add them up,
    and so obtain a bound on the whole.
*/

Bdd_expr    *ept ;
{
    double    ub, lb ;
    Bdd_term  *tpt, *last_term ;

# if DEBUG
    printf( "Bounding an expression. \n" ) ;
    fflush( stdout ) ;
# endif

    ub = ept->const_term ;
    lb = ept->const_term ;

    tpt = ept->terms ;
    for( last_term = tpt + ept->nterms ; tpt < last_term ; tpt++ ) {
	bound_term( tpt ) ;
	ub += tpt->bound.ub ;
	lb += tpt->bound.lb ;
# if DEBUG
	printf( "term.ub = %.4f, tern.lb = %.4f \n", tpt->bound.ub,
						     tpt->bound.lb ) ;
# endif
    }

    ept->bound.ub = ub ;
    ept->bound.lb = lb ;
}
/* ++++++++++++++++++++++++++++ */

void Rbound_term( tpt )
/*
    Take a list of bounded factors and obtain a bound on their
    product.
*/

Bapm_term   *tpt ;
{
    APM	     *ppt ;
    Bdd_apm  *facptr, **lastf, **fpt ;

/*
	If there is only one factor, deal with it directly.
*/
    if( tpt->nfactors == 1 ) {
	apmAssign( Rextrema, tpt->factors[0]->ub ) ;
	apmAssign( Rextremb, tpt->factors[0]->lb ) ;
    }
/*
	Handle expressions with more than one factor.
	Since some of the factors may be negative we 
	can't just multiply to gether all the upper
	and lower bounds.
*/
    else {
	apmAssign( Rextrema, tpt->factors[0]->ub ) ;
	apmAssign( Rextremb, tpt->factors[0]->lb ) ;

	fpt = &tpt->factors[1] ;
	for( lastf = tpt->factors + tpt->nfactors ; fpt < lastf ; fpt++ ) {
	    facptr = *fpt ;

	    apmMultiply( Rprod[0], facptr->ub, Rextrema ) ;
	    apmMultiply( Rprod[1], facptr->ub, Rextremb ) ;
	    apmMultiply( Rprod[2], facptr->lb, Rextrema ) ;
	    apmMultiply( Rprod[3], facptr->lb, Rextremb ) ;

	    apmAssign( Rextrema, Rprod[0] ) ;
	    apmAssign( Rextremb, Rprod[0] ) ;
	    for( ppt = (Rprod+1) ; ppt < Rlastp ; ppt++ ) {
		if( apmCompare( *ppt, Rextrema ) == 1 )
		    apmAssign( Rextrema, *ppt ) ;
		else if( apmCompare( *ppt, Rextremb ) == -1 )
		    apmAssign( Rextremb,  *ppt ) ;
	    }
	}
    }

    apmCalc( Rextrema, Rextrema, tpt->coef, APM_MUL, NULL ) ;
    apmCalc( Rextremb, Rextremb, tpt->coef, APM_MUL, NULL ) ;
    if( apmCompare( Rextrema, Rextremb ) == -1 ) {
	apmAssign( tpt->bound.ub, Rextremb ) ;
	apmAssign( tpt->bound.lb, Rextrema ) ;
    }
    else {
	apmAssign( tpt->bound.ub, Rextrema ) ;
	apmAssign( tpt->bound.lb, Rextremb ) ;
    }
}
/* ++++++++++++++++++++++++++++++++++++ */

void Rbound_expr( ept ) 
/*
    Obtain bounds on the terms in a bounded expression, add them up,
    and so obtain a bound on the whole.
*/

Bapm_expr    *ept ;
{
    Bapm_term  *tpt, *last_term ;

    apmAssign( Rub, ept->const_term ) ;
    apmAssign( Rlb, ept->const_term ) ;

    tpt = ept->terms ;
    for( last_term = tpt + ept->nterms ; tpt < last_term ; tpt++ ) {
	Rbound_term( tpt ) ;
	apmCalc( Rub, Rub, tpt->bound.ub, APM_ADD, NULL ) ;
	apmCalc( Rlb, Rlb, tpt->bound.lb, APM_ADD, NULL ) ;
    }

    apmAssign( ept->bound.ub, Rub ) ;
    apmAssign( ept->bound.lb, Rlb ) ;
}
/* +++++++++++++++++++++++++++++++++ */

void RmaxAbs( result, x, y )

APM   result, x, y ;
{
    apmAbsoluteValue( Rub, x ) ;
    apmAbsoluteValue( Rlb, y ) ;

    if( apmCompare( Rub, Rlb ) == 1 ) 
	apmAssign( result, Rub ) ;
    else
	apmAssign( result, Rlb ) ;
}
/* +++++++++++++++++++++++++++++++ */

void printBdbl( bdbl, name )

char	*name ;
Bdd_dbl *bdbl ;
{
    printf( "\t %.4f < %s < %.4f \n", bdbl->lb, name, bdbl->ub ) ;
}
    
