# ifndef _BOUNDED_EXPRS_ 
# define _BOUNDED_EXPRS_ 
/*
	Data structures for calculating semi-rigorous bounds 
	on expressions. 
								*/

typedef   struct {  double    ub, lb ; } Bdd_dbl ;

typedef   struct {
			int		nfactors ;
		    double	coef ;
		    Bdd_dbl	**factors, bound ; } Bdd_term ;

typedef  struct {   
			int			nterms ;
		    double      const_term  ;
		    Bdd_dbl     bound  ;
		    Bdd_term    *terms ; }  Bdd_expr ;
/*  
	APM partners to the structures above
						*/

typedef   struct {  APM   ub, lb ; } Bdd_apm ;

typedef   struct {  
			int		nfactors ;
		    APM		coef ;
		    Bdd_apm	**factors, bound ; } Bapm_term ;

typedef  struct {  
			int		nterms ;
			APM		const_term  ;
			Bdd_apm	bound  ;
			Bapm_term	*terms ; }  Bapm_expr ;
/* +++++++++++++++++++++++ */

# define   assignBapm( empty, full )    (apmAssign( empty->ub, full->ub), \
					 apmAssign( empty->lb, full->lb) )
# define   assignBdbl( empty, full )	(empty->ub = full->ub, \
					 empty->lb = full->lb  )
# define   newBapm( new, base )		(new.ub = apmNew( base ),\
					 new.lb = apmNew( base ) )

# define   maxAbs( x, y )	( (fabs(x) > fabs(y)) ? fabs(x) : fabs(y) ) 

/* +++++++++++++++++++++++ */

/* bounding.c */
extern void initBounding( void ) ;
extern void bound_term( Bdd_term *tpt ) ;
extern void bound_expr( Bdd_expr *ept )  ;
extern void Rbound_term( Bapm_term *tpt ) ;
extern void Rbound_expr( Bapm_expr *ept )  ;
extern void RmaxAbs( APM result, APM x, APM y ) ;
extern void printBdbl( Bdd_dbl *bdbl, char *name ) ;

/* bd_trig.c */
extern void initTrigBd(void);
extern void bd_cos(Bdd_dbl *bound, Bdd_dbl *theta);
extern void bd_sin(Bdd_dbl *bound, Bdd_dbl *theta);
extern void Rbd_cos(Bdd_apm *bound, Bdd_apm *theta);
extern void Rbd_sin(Bdd_apm *bound, Bdd_apm *theta);

# endif /* _BOUNDED_EXPRS_ */