# include  <stdio.h>
# include  <math.h>
# include  "apm.h"
# include  "converse.h"
# include  "tree.h"

# define   DEBUG   NO 
# define   TATTLE  NO

/* Variables shared via converse.h */
int	do_graph, do_backup, restoration ;
int	precision, depth, furthest, stubborn, terse ;
APM	max_error, RSmBlock_err, RBgBlock_err ;
double  SmBlock_err = DF_SQ * DBL_ERR ;
double  BgBlock_err = DEG_FREE * N_PARMS * DBL_ERR ;
/* ++++++++++++++++++++++++++++ */

int 
main (int argc, char *argv[])
{
    int	     verdict, Rverdict, tree_verdict, nsteps ;
    Prism    *image_prism ;
    RPrism   *active_prism, *old_prism ;

    handle_opts( argc, argv ) ;
    active_prism = conjureRPrism() ;
    image_prism = conjurePrism() ;
    commence( active_prism ) ;

	/* Study the current prism, cutting it up if need be */
    while( active_prism != NULL ) {
    /*
	Try a preliminary, non-rigorous calculation to see if 
	prospects are good.  If they are, do a rigorous check.
	If they aren't, try to refine the prism.  If it has already
	been refined enough, just give up.
								    */
	if( do_graph == YES ) 
	    graphPrism( active_prism, ACTIVE ) ;
/*
	Check the tree to see if an equivalent prism 
	is already finished.  If so, record the result 
	and press on.  If not, do a detailed analysis.
*/
	tree_verdict = consultTree( active_prism ) ;

# if FANCY_TREE
	if( (tree_verdict == MAYBE) || (tree_verdict == NO_TORI) ) {
	    if( do_graph == YES ) 
	        graphPrism( active_prism, tree_verdict ) ;
	    if( do_backup == YES )
		make_backup( active_prism ) ;

	    old_prism = active_prism ;
	    active_prism = old_prism->next ;

	    old_prism->in_torus = tree_verdict ;
	    if( terse == NO ) 
	        printRPrism( old_prism, 0 ) ;
	    releaseRPrism( old_prism ) ;
	}
# else
	if( tree_verdict == MAY_SKIP ) {
	    if( do_graph == YES ) 
	        graphPrism( active_prism, SYMMTRC ) ;
	    if( do_backup == YES )
		make_backup( active_prism ) ;

	    old_prism = active_prism ;
	    active_prism = old_prism->next ;

	    releaseRPrism( old_prism ) ;
	}
# endif

	else {
	    prepare_trial( active_prism ) ;
	    verdict = try_prism( active_prism, image_prism, &nsteps ) ;

	    Rverdict = UNTRIED ;
# if DEBUG
	    if( active_prism->n_cuts == depth )
		verdict = NO_TORI ;
# endif
	    if( verdict == NO_TORI ) {
		Rverdict = Rtry_prism( active_prism, image_prism, &nsteps ) ;
		if( Rverdict == NO_TORI ) {
		    active_prism->in_torus = NO_TORI ;
# if FANCY_TREE
		    colorLeaf( active_prism ) ;
# endif
		    if( terse == NO ) 
		        printRPrism( active_prism, nsteps );
		    if( do_graph == YES )
			graphPrism( active_prism, NO_TORI ) ;
		    if( do_backup == YES )
			make_backup( active_prism ) ;

		    old_prism = active_prism ;
		    active_prism = old_prism->next ;
		    releaseRPrism( old_prism ) ;
		}
# if TATTLE
		else {
		    printf( 
			"Disagreement between try() and Rtry(). \n" ) ;
		    printf( "disputed prism : \n\t" ) ;
		    printRPrism( active_prism, nsteps ) ;
		    fflush( stdout ) ;
		}
# endif

	    }

	    if( (Rverdict == MAYBE) || (verdict == MAYBE) ) {

		    /* Either refine the prism . . . */
		if( may_refine(active_prism) == YES ) {
		    refinePrism( active_prism, image_prism ) ;
		    if( do_graph == YES ) {
			graphPrism( active_prism->next, UNTRIED ) ;
			graphPrism( active_prism, ACTIVE ) ;
		    }
	        }

		    /* . . . or give up and move on. */
		else {  
		    if( do_graph == YES ) 
		        graphPrism( active_prism, MAYBE ) ;
		    if( do_backup == YES )
			make_backup( active_prism ) ;

		    active_prism->in_torus = MAYBE ;
		    moveEdge_o_Chaos( active_prism, nsteps ) ;
		    if( terse == NO ) 
		        printRPrism( active_prism, nsteps ) ;

		    old_prism = active_prism ;
		    active_prism = old_prism->next ;
# if FANCY_TREE
		    colorLeaf( old_prism ) ;
# endif
		    releaseRPrism( old_prism ) ;
		}
	    }
	}
    }

    cease() ;
}
