# ifndef _CONVERSE_KAM_
# define _CONVERSE_KAM_
# ifndef YES
# define   YES	     1		/* Ever popular */
# define   NO        0
# endif

# ifndef WORKED
# define   WORKED    1
# define   FAILED    0
# endif

# define	NO_TORI   0		/* A prism of initial conditions    */
# define	UNTRIED   1		/* has one of these as its in_torus */
# define	MAYBE     2		/* attribute according to whether   */
# define	ACTIVE    3		/* it definitely does not contain   */
# define	SYMMTRC   4		/* any invariant tori, has not yet
				   been considered, is too hard to
				   decide, is under active 
				   consideration, or is  equivalent
				   to some symmetrically related,
				   other prism.			    */

# define	N_PARMS		3	/* Details of the particular map */
# define	DEG_FREE	2
# define	DF_SQ		(DEG_FREE * DEG_FREE) 
# define	TWO_DF		(2 * DEG_FREE) 

# define	MAT_DIM		(TWO_DF + N_PARMS)	/* Used in finding ones */
# define	STAID_LEN	(N_PARMS * MAT_DIM)	/* place in the matrix" */
# define	MERC_LEN	(TWO_DF * MAT_DIM)	/* part of a prism.     */
# define	MAT_SZ		(STAID_LEN + MERC_LEN)

# define	DBL_ERR		1.0e-13

/* Default values for various global variables */
# define	DFLT_DEPTH		10	
# define	DFLT_FRTHST		200
# define	DFLT_PRECIS		30
# define	DFLT_GRAPH		YES
# define	DFLT_GRAF_FILE	"ConverseKAM.ps"	
# define	DFLT_ERR		YES	
# define	DFLT_TERSE		YES
# define	STUBBORN 		NO
# define	SAFETY_PRECIS	5 

/* Numbers bearing on the accuracy of the starting point */
# define	START_SZ	6	
# define	NEWT_TOL	10e-8

# define	BASE		10000	/* The base used in APM calculations.  */

/*
	Data types for non-rigorous, rough calculations
*/

typedef	double   *Tor_pt, *Parm_pt ;

typedef	struct { Tor_pt	u, v ; } Embed_pt ;

typedef	struct { 
		Embed_pt	z ;
		Parm_pt		p ; } Xtnd_pt ;

typedef struct  prsm {  
		int			in_torus, n_cuts ;
		char		*cuts[N_PARMS+TWO_DF] ;
		double		*matrix ;
		Xtnd_pt		*center ;
		struct prsm	*next ;     } Prism ;
/*
	Data types for rigorous, arbitrary precision, calculations 
								*/

typedef	APM   *RTor_pt, *RParm_pt ;

typedef	struct { RTor_pt	u, v ; } REmbed_pt ;

typedef	struct { REmbed_pt	z ;
		  RParm_pt	p ; } RXtnd_pt ;

typedef	struct  Rprsm {  int		in_torus, n_cuts ;
			  APM		*matrix ;
			  char		*cuts[MAT_DIM] ;
			  RXtnd_pt	*center ;
			  struct Rprsm	*next ;     } RPrism ;

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
	Variables used throughout the converse KAM calculations, 
	along with the files where they are declared.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/* backup.c */
extern  char	*back_name, *rest_name, *parm_names[] ;

/* converse.c */
extern int 		do_graph, do_backup, restoration ;
extern int		precision, depth, furthest, terse, stubborn ;
extern  APM     max_error, RSmBlock_err, RBgBlock_err ;
extern  double  SmBlock_err, BgBlock_err ;

/* elxopt.c */
extern  char	*optarg ;
extern  int		optind, opterr ;

/* follow.c */
extern  int		quick_tries, tries, Rtries, max_steps, max_NTsteps ;
extern  int		HermSuccess, LLSuccess, ll_used[3], most_cuts ;

/* lambda.c */
extern APM		RminLeastLam, RmaxLeastLam, RfirstLeastLam, RsumTinyLams ;
extern double	firstLeastLam, minLeastLam, maxLeastLam, sumTinyLams ;

/* lipschitz.c */
extern double	first_slope, min_slope, max_slope ;
extern APM		Rfirst_slope, Rmin_slope, Rmax_slope, Rdf, Rdf_sq ;
extern APM		Rsqrt_disc ;

/* prismatic.c */
extern void		(* fatten)(), (* row_sums)() ;

/* PSgraph.c or graph.c */
extern  char	*graf_file ;

/* start.c */
extern  APM     Rstart_size ;

/* termina.c */
extern  double  parm_roof[], parm_floor[] ;

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
	Declarations of functions
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/* backup.c */
extern void startBackup(RPrism *priz);
extern void make_backup(RPrism *priz);
extern void restore_options(void);
extern void restore_start(void);
extern void restore_chain(RPrism *first);
extern void begin_restoration(void);
extern void restore_edge(void);
extern void restore_performance(void);

/* cr_fatten.c */
extern void initRotor(void);
extern void col_rotor(double *Amat, double *Deriv, double *Prizmat);
extern void Rcol_rotor(APM *Amat, APM *Deriv, APM *Prizmat);

/* cr_rows.c */
extern void cr_rows(double *w, double *Amat, double *Deriv, Prism *Priz);
extern void Rcr_rows(APM *Rw, APM *Amat, APM *Deriv, RPrism *Priz);

/* elxopt.c */
extern char getopt(int argc, char *argv[], char *opt_str);

/* ff_fatten.c */
extern void fxed_form(double *Amat, double *Deriv, double *Prizmat);
extern void Rfxed_form(APM *Amat, APM *Deriv, APM *Prizmat);

/* ff_rows.c */
extern void init_ffRows(void);
extern void ff_rows(double *w, double *Amat, double *Deriv, Prism *Priz);
extern void Rff_rows(APM *w, APM *Amat, APM *Deriv, RPrism *Priz);
extern void Rbeta_star(APM answer);

/* follow.c */
extern void prepare_trial(RPrism *priz);
extern int try_prism(RPrism *initial_priz, Prism *final_priz, int *nsteps);
extern void initFollowing(void);
extern int Rtry_prism(RPrism *initial_priz, Prism *final_priz, int *nsteps);
extern int big_Prism(Prism *Priz);
extern int big_RPrism(RPrism *Priz);
extern void print_perf(void);
extern double best_ll(double minBlam_ll, double maxBlam_ll, double trace_ll);
extern void Rbest_ll(APM best, APM minBlam_ll, APM maxBlam_ll, APM trace_ll);

/* gauss.c */
extern void initGauss(void);
extern void gauss(double **a);
extern void Rgauss(APM **a);

/* lambda.c */
# include "bounding.h"

extern void initLambda(void);
extern void bd_Blams(Bdd_dbl *leastBlam, Bdd_dbl *bigBlam, Bdd_dbl *trace);
extern void Rbd_Blams(Bdd_apm *leastBlam, Bdd_apm *bigBlam, Bdd_apm *trace);
extern void setLLbounds(RPrism *priz);
extern double calc_bML(double *beta, double b_trace);
extern void setLeastLam(RPrism *priz);

/* lipschitz.c */
extern void initLip(void);
extern void setCone(RPrism *priz);
extern void setSlopes(RPrism *priz);

/* memory.c */
# include "tree.h"

extern Prism *conjurePrism(void);
extern void releasePrism(Prism *ppt);
extern RPrism *conjureRPrism(void);
extern void releaseRPrism(RPrism *ppt);
extern void m_free(char *cpt);

# if FANCY_TREE /* set in tree.h */
	
	extern PrismLeaf *conjureLeaf(void);
# endif /* FANCY_TREE */

/* options.c */
extern void handle_opts(int argc, char *argv[]);

/* prismatic.c */
extern void prismatic_image(Prism *pz_prime, Prism *pz);
extern void Rprismatic_image(RPrism *pz_prime, RPrism *pz);
extern void initPrismatic(void);

/* PSgraph.c */
extern void startGraphics(void);
extern void restartGraphics(void);
extern void finishGraphics(void);
extern void graphPrism(RPrism *rpt, int style);

/* refine.c */
extern void initRefinement(void);
extern int may_refine(RPrism *priz);
extern void refinePrism(RPrism *priz, Prism *image);

/* start.c */
extern void setHermStart(RPrism *priz);
extern void setLLStart(RPrism *priz);
extern void shiftStart(RPrism *priz);

/* termina.c */
extern void commence (RPrism *first);
extern void cease(void);
extern void moveEdge_o_Chaos(RPrism *priz, int nsteps);
extern void setEdge_o_Chaos(RPrism *priz);
extern void report_performance(void);

/* trace.c */
extern void initTrace(void);
extern void bound_btrace(Bdd_dbl *result, Prism *priz);
extern void Rbound_btrace(Bdd_apm *result, RPrism *priz);

/* tree.c */
extern int consultTree(RPrism *rpt);
extern void initTree(void);
extern void colorLeaf(RPrism *rpt);

/* util.c */
extern void fprintRPrism(FILE *stream, RPrism *priz, int nsteps);
extern void printRPrism(RPrism *priz, int nsteps);
extern void copyRPrism(Prism *ppt, RPrism *Rpt);
extern void copyPrism(Prism *a_ppt, Prism *b_ppt);
extern void RcopyRPrism(RPrism *Rapt, RPrism *Rbpt);
extern void read_parameters(RPrism *first);
extern void echo_parameters(RPrism *first);
extern int errHndlr(int errcode, char *message, char *file, int line, char *function);
extern void printMat(double *mat);
extern void printRmat(APM *mat);
extern void initUtil(void);
extern void printXpt(Xtnd_pt *xpt);
extern void printRXpt(RXtnd_pt *xpt);
extern void truncateRPrism(RPrism *priz, int dp);
extern int areNewParms(RPrism *priz);

# endif /* _CONVERSE_KAM_ */
