# include  <stdio.h>
# include  <math.h>

# include  "apm.h"
# include  "apmSpecial.h"
# include  "converse.h"
# include  "bounding.h"
# include  "rows.h"
# include  "pi.h"

# define   m_sign( x )		( (x > 0.0) ? 1.0 : -1.0 )
# define   Rm_sign( x, a ) 	apmAssignLong((x),(long)apmSign((a)),0,BASE) 	


# define   DEBUG 	NO
# define   USE_ROT	YES

# define   THETA_ROT	0.5	/* The angle for rotations - it's 
				   recorded here in units of pi.    */
				   
/* Functions private to this file */
static void subspace_rot(double *Amat, int col_one, int col_two);
static void Rsubspace_rot(APM *Amat, int col_one, int col_two);

/* Variables private to this file */
static APM	 Rcthet, Rsthet, Rsmall_sinsq ;
static APM	 Rarea, Rsin_sq, Rnorm_one, Rnorm_two, Rsign ;
static APM	 Rnorm_prod, Rsign, Rx, Ry ;
static double   cthet, sthet, small_sinsq ;
/* ++++++++++++++++++++++++++++++++++++++++++ */

void 
col_rotor (double *Amat, double *Deriv, double *Prizmat)
/*
     Prepares the matrix called "A" in my notes.  Mostly we want to
     have A = DF*Priz, but  we want to ensure that A is not singular.
     In the interest of speed we have coded the calculations below with 
     pointers.  Our hope is that the resulting function will scream along 
     at ultrasonic speed.  Unfortunately it is quite unreadable.
*/
{
    int		     j, k ;
    double	     *Aend, *Dend, *Pend ;
    register double  *Apt, *Dpt, *Ppt ;

/*
	Copy the few terms which appear in the top rows of Amat.
*/
    Aend = Amat + N_PARMS * (MAT_DIM + 1) ;
    for( Apt = Amat, Ppt = Prizmat ; Apt < Aend ; Apt += (MAT_DIM + 1 ),
						  Ppt += (MAT_DIM + 1 ) ) 
	*Apt = *Ppt ;

# if DEBUG
    printf( "A(p,p) set \n" ) ;
# endif

/*
	Clear out those parts of Amat which change from iteration to 
	iteration. 
*/

    Aend = Amat + MAT_SZ ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt++ )
	*Apt = 0.0 ;

/*
	Set the (u,p) part of A
	It's equal to the (v,p) part of Prizmat.
*/
    
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt += TWO_DF ) {
	for( Pend = Ppt + N_PARMS ; Ppt < Pend ; Ppt++ )
	    *Apt++ = *Ppt ;

	Ppt += TWO_DF ;
    }

# if DEBUG
    printf( "A(u,p) set \n" ) ;
# endif
/*
	Set the (v,p) part - three terms.
*/
	/*  First term - equal to Deriv(v,p) * Prizmat(p,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Aend = Apt + (DEG_FREE*MAT_DIM) ; Apt < Aend ; Apt += TWO_DF ) {
	Ppt = Prizmat ;
	for( Dend = Dpt + N_PARMS ; Dpt < Dend ; Dpt++ ) {
	    *Apt++ += *Dpt * (*Ppt) ;
	    Ppt += MAT_DIM + 1 ;
	}

	Dpt += TWO_DF ;
    }

# if DEBUG
    printf( "A(v,p), first term, set \n" ) ;
# endif
	/*  Second term - equal to negative Prizmat(u,p) */

    Ppt = Prizmat + STAID_LEN ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Pend = Ppt + (DEG_FREE * MAT_DIM) ;  Ppt < Pend ; Ppt += TWO_DF ) {
	for( Aend = Apt + N_PARMS ; Apt < Aend ; Apt++ )
	    *Apt -= *Ppt++ ;

	Apt += TWO_DF ;
    }

# if DEBUG
    printf( "A(v,p), second term, set \n" ) ;
# endif
	/*  Third term - equal to Deriv(v,v) * Prizmat(v,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * (MAT_DIM + 1)) + N_PARMS ;
    Dend = Deriv + MAT_SZ ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    while( Dpt < Dend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
	Pend = Prizmat + MAT_SZ ;
	while( Ppt < Pend ) {
	    Aend = Apt + N_PARMS ;
	    while( Apt < Aend )
		*Apt++ += *Dpt * (*Ppt++) ;

	    Dpt++ ;
	    Ppt += TWO_DF ;
	    Apt -= N_PARMS ;
	}

	Dpt += N_PARMS + DEG_FREE ;
	Apt += MAT_DIM ;
    }

# if DEBUG
    printf( "A(v,p), third term, set \n" ) ;
# endif

/*
	(u,u) part
	equals Priz(v,u)
*/

    Apt = Amat + STAID_LEN + N_PARMS ;
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    while( Apt < Aend ) {
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) 
	    *Apt++ = *Ppt++ ;

	Apt  += N_PARMS + DEG_FREE ;
	Ppt  += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(u,u) set \n" ) ;
# endif

/*
	(u,v) part
	equals Priz(v,v)
*/

    Apt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    while( Apt < Aend ) {
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) 
	    *Apt++ = *Ppt++ ;

	Apt  += N_PARMS + DEG_FREE ;
	Ppt  += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(u,v) set \n" ) ;
# endif

/*
	The (v,u) part - equal to Deriv(v,v) * Priz(v,u) - Priz(u,u) ,
*/

	/* First term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    Aend = Apt + (DEG_FREE * MAT_DIM) ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    while( Apt < Aend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) {
	    Dend = Dpt + DEG_FREE ;
	    while( Dpt < Dend )  {
		*Apt += *Dpt++ * *Ppt ;

		Ppt += MAT_DIM ;
	    }
	    Apt++ ;
	    Dpt -= DEG_FREE ;
	    Ppt -= (DEG_FREE * MAT_DIM) - 1 ;
	}
	Dpt += MAT_DIM ;
	Apt += N_PARMS + DEG_FREE ;
	Ppt += N_PARMS + DEG_FREE ;
    }
	
	/* Second term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    Ppt = Prizmat + STAID_LEN + N_PARMS ;
    Pend = Ppt + (MAT_DIM * DEG_FREE) ;
    while( Ppt < Pend ) {
	Aend = Apt + DEG_FREE ;
	while( Apt < Aend ) 
	    *Apt++ -= *Ppt++ ;

	Ppt += N_PARMS + DEG_FREE ;
	Apt += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(v,u) set \n" ) ;
# endif

/*
	(v,v) part - equals Deriv(v,v) * Priz(v,v) - Priz(u,v) 
*/

	/* First term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Aend = Apt + (DEG_FREE * MAT_DIM) ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    while( Apt < Aend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE*MAT_DIM) + 
				    N_PARMS + DEG_FREE ;
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) {
	    Dend = Dpt + DEG_FREE ;
	    while( Dpt < Dend )  {
		*Apt += *Dpt++ * *Ppt ;

		Ppt += MAT_DIM ;
	    }
	    Apt++ ;
	    Dpt -= DEG_FREE ;
	    Ppt -= (DEG_FREE * MAT_DIM) - 1 ;
	}
	Dpt += MAT_DIM ;
	Apt += N_PARMS + DEG_FREE ;
	Ppt += N_PARMS + DEG_FREE ;
    }
	
	/* Second term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Ppt = Prizmat + STAID_LEN + N_PARMS + DEG_FREE ;
    Pend = Ppt + (MAT_DIM * DEG_FREE) ;
    while( Ppt < Pend ) {
	Aend = Apt + DEG_FREE ;
	while( Apt < Aend ) 
	    *Apt++ -= *Ppt++ ;

	Ppt += N_PARMS + DEG_FREE ;
	Apt += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(v,v) set \n" ) ;

    printf( "\n" ) ;
    printMat( Amat ) ;
    printf( "\n" ) ;
# endif

# if USE_ROT
/*
	Do up the rotations.
*/
    for( j=0 ; j < TWO_DF ; j++ )
	for( k=(j+1) ; k < TWO_DF ; k++ )
	    subspace_rot( Amat, j, k ) ;
# endif
}
/* +++++++++++++++++++++++++++++ */

void 
subspace_rot (double *Amat, int col_one, int col_two)
{
    double   *Apt, *Apt2 ;
    double   x, y ;
    double   norm_one, norm_two, norm_prod ;
    double   area, sin_sq, sign ;
/*
	Since this function only rotates two dimensional 
	subspaces it also only calculates angles and norms 
	in two dimensions.
*/

    Apt = Amat + STAID_LEN + N_PARMS + 
		 (col_two - col_one - 1) * MAT_DIM  +
		 col_one ;
    Apt2 = Apt + col_two - col_one ;

    area = *Apt * Apt2[MAT_DIM]  -  Apt[MAT_DIM] * *Apt2 ;
    norm_one = *Apt * *Apt + Apt[MAT_DIM] * Apt[MAT_DIM] ;
    norm_two = *Apt2 * *Apt2 + Apt2[MAT_DIM] * Apt2[MAT_DIM] ;
    norm_prod = norm_one * norm_two ;
    if( norm_prod != 0.0 ) { 
	sin_sq = area * area / norm_prod ;

	if( sin_sq < small_sinsq ) {
	    sign = m_sign( area ) ;

	    if( norm_two < norm_one ) {
		x = cthet * *Apt2 - sign * sthet * *(Apt2 + MAT_DIM) ;
		y = sign * sthet * *Apt2  +  cthet * *(Apt2 + MAT_DIM) ;

		Apt2[0] = x ;
		Apt2[MAT_DIM] = y ;	
	    }
	    else {
		sign = -sign ;
		x = cthet * *Apt - sign * sthet * *(Apt + MAT_DIM) ;
		y = sign * sthet * *Apt  +  cthet * *(Apt + MAT_DIM) ;

		Apt[0] = x ;
		Apt[MAT_DIM] = y ;	
	    }
# if DEBUG
	printf( "Did a (%d, %d) subspace rotation. \n", col_one, 
							col_two ) ;
	printf( "x = %.6f, y = %.6f \n", x, y ) ;
# endif
	}
    }	
}

/* ++++++++++++++++++++++++ */

void 
initRotor (void)
{
    Rcthet = apmNew( BASE ) ;
    Rsthet = apmNew( BASE ) ;

    Rx = apmNew( BASE ) ;
    Ry = apmNew( BASE ) ;
    Rarea = apmNew( BASE ) ;
    Rsign = apmNew( BASE ) ;
    Rsin_sq = apmNew( BASE ) ;
    Rnorm_one = apmNew( BASE ) ;
    Rnorm_two = apmNew( BASE ) ;
    Rnorm_prod = apmNew( BASE ) ;
    Rsmall_sinsq = apmNew( BASE ) ;

    cthet = cos( PI * THETA_ROT ) ;
    sthet = sin( PI * THETA_ROT ) ;
    small_sinsq = sthet * sthet ;

    dbltoapm( Rx, BASE, THETA_ROT ) ;
    apmMultiply( Ry, pi, Rx ) ;
    apmCos( Rcthet, Ry ) ;
    apmSin( Rsthet, Ry ) ;
    apmMultiply( Rsmall_sinsq, Rsthet, Rsthet ) ;
}
/* ++++++++++++++++++++++++++++++++++++++++++ */

void 
Rcol_rotor (APM *Amat, APM *Deriv, APM *Prizmat)
/*
     Prepares the matrix called "A" in my notes.  Mostly we want to
     have A = DF*Priz, but  we want to ensure that A is not singular.
     In the interest of speed we have coded the calculations below with 
     pointers.  Our hope is that the resulting function will scream along 
     at ultrasonic speed.  Unfortunately it is quite unreadable.
*/
{
    int		     j, k ;
    APM		     *Aend, *Dend, *Pend ;
    register APM     *Apt, *Dpt, *Ppt ;

/*
	Copy the few terms which appear in the top rows of Amat.
*/
    Aend = Amat + N_PARMS * (MAT_DIM + 1) ;
    for( Apt = Amat, Ppt = Prizmat ; Apt < Aend ; Apt += (MAT_DIM + 1 ),
						  Ppt += (MAT_DIM + 1 ) ) 
	apmAssign( *Apt, *Ppt ) ;

# if DEBUG
    printf( "A(p,p) set \n" ) ;
# endif

/*
	Clear out those parts of Amat which change from iteration to 
	iteration. 
*/

    Aend = Amat + MAT_SZ ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt++ )
	apmAssignLong( *Apt, 0L, 0, BASE ) ;

/*
	Set the (u,p) part of A
	It's equal to the (v,p) part of Prizmat.
*/
    
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt += TWO_DF ) {
	for( Pend = Ppt + N_PARMS ; Ppt < Pend ; Ppt++ )
	    apmAssign( *Apt++, *Ppt ) ;

	Ppt += TWO_DF ;
    }

# if DEBUG
    printf( "A(u,p) set \n" ) ;
# endif
/*
	Set the (v,p) part - three terms.
*/
	/*  First term - equal to Deriv(v,p) * Prizmat(p,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Aend = Apt + (DEG_FREE*MAT_DIM) ; Apt < Aend ; Apt += TWO_DF ) {
	Ppt = Prizmat ;
	for( Dend = Dpt + N_PARMS ; Dpt < Dend ; Dpt++ ) {
	    apmCalc( *Apt, *Apt, *Dpt, *Ppt, APM_MUL, APM_ADD, NULL ) ;
	    Apt++ ;
	    Ppt += MAT_DIM + 1 ;
	}

	Dpt += TWO_DF ;
    }

# if DEBUG
    printf( "A(v,p), first term, set \n" ) ;
# endif
	/*  Second term - equal to negative Prizmat(u,p) */

    Ppt = Prizmat + STAID_LEN ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Pend = Ppt + (DEG_FREE * MAT_DIM) ;  Ppt < Pend ; Ppt += TWO_DF ) {
	for( Aend = Apt + N_PARMS ; Apt < Aend ; Apt++ )
	    apmCalc( *Apt, *Apt, *Ppt++, APM_SUB, NULL ) ;

	Apt += TWO_DF ;
    }

# if DEBUG
    printf( "A(v,p), second term, set \n" ) ;
# endif
	/*  Third term - equal to Deriv(v,v) * Prizmat(v,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * (MAT_DIM + 1)) + N_PARMS ;
    Dend = Deriv + MAT_SZ ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    while( Dpt < Dend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
	Pend = Prizmat + MAT_SZ ;
	while( Ppt < Pend ) {
	    Aend = Apt + N_PARMS ;
	    while( Apt < Aend ) {
		apmCalc( *Apt, *Apt, *Dpt, *Ppt, APM_MUL, APM_ADD, NULL ) ;
		Apt++ ;
		Ppt++ ;
	    }

	    Dpt++ ;
	    Ppt += TWO_DF ;
	    Apt -= N_PARMS ;
	}

	Dpt += N_PARMS + DEG_FREE ;
	Apt += MAT_DIM ;
    }

# if DEBUG
    printf( "A(v,p), third term, set \n" ) ;
# endif

/*
	(u,u) part
	equals Priz(v,u)
*/

    Apt = Amat + STAID_LEN + N_PARMS ;
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    while( Apt < Aend ) {
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend )  {
	    apmAssign( *Apt++, *Ppt ++) ;
	}

	Apt  += N_PARMS + DEG_FREE ;
	Ppt  += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(u,u) set \n" ) ;
# endif

/*
	(u,v) part
	equals Priz(v,v)
*/

    Apt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    while( Apt < Aend ) {
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) 
	    apmAssign( *Apt++, *Ppt++ ) ;

	Apt  += N_PARMS + DEG_FREE ;
	Ppt  += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(u,v) set \n" ) ;
# endif

/*
	The (v,u) part - equal to Deriv(v,v) * Priz(v,u) - Priz(u,u) ,
*/

	/* First term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    Aend = Apt + (DEG_FREE * MAT_DIM) ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    while( Apt < Aend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) {
	    Dend = Dpt + DEG_FREE ;
	    while( Dpt < Dend )  {
		apmCalc( *Apt, *Apt, *Dpt++, *Ppt, APM_MUL, 
					     APM_ADD, NULL ) ;

		Ppt += MAT_DIM ;
	    }
	    Apt++ ;
	    Dpt -= DEG_FREE ;
	    Ppt -= (DEG_FREE * MAT_DIM) - 1 ;
	}
	Dpt += MAT_DIM ;
	Apt += N_PARMS + DEG_FREE ;
    }
	
	/* Second term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Ppt = Prizmat + STAID_LEN + N_PARMS ;
    Pend = Ppt + (MAT_DIM * DEG_FREE) ;
    while( Ppt < Pend ) {
	Aend = Apt + DEG_FREE ;
	while( Apt < Aend )  {
	    apmCalc( *Apt, *Apt, *Ppt, APM_SUB, NULL ) ;
	    Apt++ ;
	    Ppt++ ;
	}

	Ppt += N_PARMS + DEG_FREE ;
	Apt += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(v,u) set \n" ) ;
# endif

/*
	(v,v) part - equals Deriv(v,v) * Priz(v,v) - Priz(u,v) 
*/

	/* First term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Aend = Apt + (DEG_FREE * MAT_DIM) ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    while( Apt < Aend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE*MAT_DIM) + 
				    N_PARMS + DEG_FREE ;
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) {
	    Dend = Dpt + DEG_FREE ;
	    while( Dpt < Dend )  {
		apmCalc( *Apt, *Apt, *Dpt++, *Ppt, APM_MUL, 
					     APM_ADD, NULL ) ;

		Ppt += MAT_DIM ;
	    }
	    Apt++ ;
	    Dpt -= DEG_FREE ;
	    Ppt -= (DEG_FREE * MAT_DIM) - 1 ;
	}
	Dpt += MAT_DIM ;
	Apt += N_PARMS + DEG_FREE ;
    }
	
	/* Second term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Ppt = Prizmat + STAID_LEN + N_PARMS + DEG_FREE ;
    Pend = Ppt + (MAT_DIM * DEG_FREE) ;
    while( Ppt < Pend ) {
	Aend = Apt + DEG_FREE ;
	while( Apt < Aend ) {
	    apmCalc( *Apt, *Apt, *Ppt, APM_SUB, NULL ) ;
	    Apt++ ;
	    Ppt++ ;
	}

	Ppt += N_PARMS + DEG_FREE ;
	Apt += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(v,v) set \n" ) ;

    printf( "\n" ) ;
    printRmat( Amat ) ;
    printf( "\n" ) ;
# endif

# if USE_ROT
/*
	Do up the rotations.
*/
    for( j=0 ; j < TWO_DF ; j++ )
	for( k=(j+1) ; k < TWO_DF ; k++ )
	Rsubspace_rot( Amat, j, k ) ;
# endif
}
/* +++++++++++++++++++++++++++++ */

void 
Rsubspace_rot (APM *Amat, int col_one, int col_two)
{
    APM   *Apt, *Apt2 ;

    Apt = Amat + STAID_LEN + N_PARMS + 
		 (col_two - col_one - 1) * MAT_DIM  +
		 col_one ;
    Apt2 = Apt + col_two - col_one ;


    apmCalc( Rarea, *Apt, Apt2[MAT_DIM], APM_MUL,
		    Apt[MAT_DIM], *Apt2, APM_MUL, 
				  APM_SUB, NULL ) ;
    apmCalc( Rnorm_one, *Apt, APM_DUP, APM_MUL,
			Apt[MAT_DIM], APM_DUP, APM_MUL, 
					APM_ADD, NULL ) ;
    apmCalc( Rnorm_two, *Apt2, APM_DUP, APM_MUL,
			Apt2[MAT_DIM], APM_DUP, APM_MUL, 
					APM_ADD, NULL ) ;
    apmMultiply( Rnorm_prod, Rnorm_one, Rnorm_two ) ;
    if( apmCompare( Rnorm_prod, zero ) == 1 ) {
	apmMultiply( Rx, Rarea, Rarea ) ;
	apmDivide( Rsin_sq, precision, (APM) NULL, Rx, Rnorm_prod ) ;

	if( apmCompare( Rsin_sq, Rsmall_sinsq ) == -1 ) {
	    Rm_sign( Rsign, Rarea ) ;

	    if( apmCompare( Rnorm_two, Rnorm_one ) != 1 ) {
	    	apmCalc( Rx, Rcthet, *Apt2, APM_MUL,
		     Rsign, Rsthet, Apt2[MAT_DIM], APM_MUL, APM_MUL,
		     APM_SUB, NULL ) ;
	    	apmCalc( Ry, Rsthet, *Apt2, Rsign, APM_MUL, APM_MUL,
		     Rcthet, Apt2[MAT_DIM], APM_MUL, 
		     APM_ADD, NULL ) ;

	    	apmAssign( *Apt2, Rx ) ;
	    	apmAssign( Apt2[MAT_DIM], Ry ) ;
	    }
	    else {
		apmCalc( Rsign, Rsign, APM_NEG, NULL ) ;
	    	apmCalc( Rx, Rcthet, *Apt, APM_MUL,
		     Rsign, Rsthet, Apt[MAT_DIM], APM_MUL, APM_MUL,
		     APM_SUB, NULL ) ;
	    	apmCalc( Ry, Rsthet, *Apt, Rsign, APM_MUL, APM_MUL,
		     Rcthet, Apt[MAT_DIM], APM_MUL, 
		     APM_ADD, NULL ) ;

	    	apmAssign( *Apt, Rx ) ;
	    	apmAssign( Apt[MAT_DIM], Ry ) ;
	    }
# if DEBUG
	printf( "Did a (%d, %d) subspace rotation. \n", col_one, 
							col_two ) ;
	printf( "Rx = %.6f, Ry = %.6f \n", apmtodbl( Rx ), 
					   apmtodbl( Ry ) ) ;
# endif
	}
    }
}
