# include <stdio.h>
# include <math.h>
# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"

# define  DEBUG	      NO 

# define  NUM_FACTS   3 
# define  NUM_TERMS   3 
# define  DET_TOL     1e-13

/* Variables private to this file */
static APM		cr_scratch ;
static APM		RBmat[MAT_SZ], Rconst_mat[DF_SQ], Rcopy[4 * DF_SQ] ;
static APM		*Rcopy_rows[TWO_DF] ;
static APM		RBu_rows[DEG_FREE], RBv_rows[DEG_FREE] ;
static APM		Rbd_star, Rgd_star, Rstar, RPvp_star ;
static APM		Rcenter_err[MAT_DIM] ;
static APM		Rup_rows[DEG_FREE], Ruu_rows[DEG_FREE], Ruv_rows[DEG_FREE] ;
static APM		Rvp_rows[DEG_FREE], Rvu_rows[DEG_FREE], Rvv_rows[DEG_FREE] ;

static double	Bmat[MAT_SZ], const_mat[DF_SQ], copy[4 * DF_SQ] ;
static double	*copy_rows[TWO_DF] ;
static double	Bu_rows[DEG_FREE], Bv_rows[DEG_FREE] ;
static double	bd_star, gd_star, star, Pvp_star ;
static double	center_err[MAT_DIM] ;
static double	up_rows[DEG_FREE], uu_rows[DEG_FREE], uv_rows[DEG_FREE] ;
static double	vp_rows[DEG_FREE], vu_rows[DEG_FREE], vv_rows[DEG_FREE] ;

static Bdd_dbl		*cr_factors[NUM_FACTS] ;
static Bdd_term		cr_terms[NUM_TERMS] ;
static Bdd_expr		beta_prod ;

static Bdd_apm		*Rcr_factors[NUM_FACTS] ;
static Bapm_term	Rcr_terms[NUM_TERMS] ;
static Bapm_expr	Rbeta_prod ;

/* Functions private to this file */
static void bound_rows(double *rows, double *first_b, double *first_p, double *second_b, double *second_p);
static void Rbound_rows(APM *rows, APM *first_b, APM *first_p, APM *second_b, APM *second_p);
static void set_inverse(double *mat);
static void invert_corner(double *mat);
static void Rset_inverse(APM *mat);
static void Rinvert_corner(APM *mat);
static void MatMult(double *a, double *b);
static void Rcheck_inverse(APM *a, APM *b);
/* ++++++++++++++++++++++++++++++ */

void 
init_crRows (void)
/*
	Set up the expressions and terms as described in my notes
	from 12/3 and 12/4.
*/
{
    int	      j, k ;
    APM	      *Rcpt ;
    double    *cpt ;
    Bdd_dbl   **dpt ;
    Bdd_apm   **apt ;

/*
	Initialize a batch of APM's.
*/
    for(j=0 ; j < DEG_FREE ; j++ ) {
	Rvp_rows[j] = apmNew( BASE ) ;
	Rup_rows[j] = apmNew( BASE ) ;
	Ruu_rows[j] = apmNew( BASE ) ;
	Ruv_rows[j] = apmNew( BASE ) ;
	Rvu_rows[j] = apmNew( BASE ) ;
	Rvv_rows[j] = apmNew( BASE ) ;
	RBu_rows[j] = apmNew( BASE ) ;
	RBv_rows[j] = apmNew( BASE ) ;
    }

    Rstar = apmNew( BASE ) ;
    Rgd_star = apmNew( BASE ) ;
    Rbd_star = apmNew( BASE ) ;
    RPvp_star = apmNew( BASE ) ;
    cr_scratch = apmNew( BASE ) ;
    for( j=0 ; j < MAT_SZ ; j++ )  {
	Bmat[j] = 0.0 ;
	RBmat[j] = apmNew( BASE ) ;
    }

    for( j=0 ; j < DF_SQ ; j++ ) 
	Rconst_mat[j] = apmNew( BASE ) ;

    for( j=0 ; j < (4 * DF_SQ) ; j++ )
	Rcopy[j] = apmNew( BASE ) ;

    for( j=0 ; j < MAT_DIM ; j++ )
	Rcenter_err[j] = apmNew( BASE ) ;

    cpt = copy ;
    Rcpt = Rcopy ;
    for( j=0 ; j < TWO_DF ; j++ ) {
	copy_rows[j] = cpt ;
	Rcopy_rows[j] = Rcpt ;

	cpt += TWO_DF ;
	Rcpt += TWO_DF ;
    }

/*
	Set the number of terms in the bounded expressions
*/

    beta_prod.nterms = Rbeta_prod.nterms = 3 ;

/*	
	Assign terms
*/

    beta_prod.terms = cr_terms ;
    Rbeta_prod.terms = Rcr_terms ;

/*
	Set nfactors.
*/
    
    Rbeta_prod.terms[0].nfactors = beta_prod.terms[0].nfactors = 1  ;
    Rbeta_prod.terms[1].nfactors = beta_prod.terms[1].nfactors = 1  ;
    Rbeta_prod.terms[2].nfactors = beta_prod.terms[2].nfactors = 1  ;

/*
	Assign factors. 
*/

    dpt = cr_factors ;
    apt = Rcr_factors ;
    for( k=0 ; k < beta_prod.nterms ; k++ ) {
	beta_prod.terms[k].factors = dpt ;
	Rbeta_prod.terms[k].factors = apt ;

	dpt += beta_prod.terms[k].nfactors ;
	apt += Rbeta_prod.terms[k].nfactors ;
    }

/*
	Set up those of the "bound" attributes which are 
	bounded APM's.
*/

    newBapm( Rbeta_prod.bound, BASE ) ;
    for( j=0 ; j < NUM_TERMS ; j++ ) {
	newBapm( Rcr_terms[j].bound, BASE ) ;
    }

/*
	Set up the terms and expressions. 
*/

/* beta_prod */
 
    Rbeta_prod.const_term = apmNew( BASE ) ;
	Rbeta_prod.terms[0].coef = apmNew( BASE ) ;

	    beta_prod.terms[0].factors[0] = &a_sin.bound ;
	    Rbeta_prod.terms[0].factors[0] = &Ra_sin.bound ;

	Rbeta_prod.terms[1].coef = apmNew( BASE ) ;

	    beta_prod.terms[1].factors[0] = &c_sin.bound ;
	    Rbeta_prod.terms[1].factors[0] = &Rc_sin.bound ;

	Rbeta_prod.terms[2].coef = apmNew( BASE ) ;

	    beta_prod.terms[2].factors[0] = &b_sin.bound ;
	    Rbeta_prod.terms[2].factors[0] = &Rb_sin.bound ;
}
/* +++++++++++++++++++++++++++++++++ */

void 
cr_rows (double *w, double *Amat, double *Deriv, Prism *Priz)
/*
    Obtain bounds on the sums of the absolute values of
    the entries in the rows of 
       -1
    [A]   * Deriv * Pmat,

    put the results in w.
*/
{
    int		j ;
    double	*end_row, *end_mat, *Pmat, *inv_pt ;
    double	*p1pt, *p2pt, *b1pt, *b2pt, *wu_pt, *wv_pt ;

    Pmat = Priz->matrix ;
    set_inverse( Amat ) ;
# if DEBUG
	printf( "\n" ) ;
	printMat( Bmat ) ;
	printf( "\n" ) ;
	MatMult( Amat, Bmat ) ;
	printf( "\n" ) ;
# endif
/*
	Do up some row sums for the inverse; these
	are used to calculate center_err[].
*/
    b1pt = Bmat + STAID_LEN + N_PARMS ;
    b2pt = b1pt + MAT_DIM * DEG_FREE ;
    for( j=0  ; j < DEG_FREE ; j++ ) {
	Bu_rows[j] = 0.0 ;
	Bv_rows[j] = 0.0 ;

	for( end_row = b1pt + TWO_DF ; b1pt < end_row ; ) {
	    Bu_rows[j] += fabs( *b1pt++ ) ;
	    Bv_rows[j] += fabs( *b2pt++ ) ;
	}
    }
/*
	Call functions which calculate upper bound on the 
	sums of the elements of various matrices.
	Before any bounding of matrices, one must invoke
	global_bounds( Pmat ) to set such global variables,
	as cos_one, and sin_sum. This is assumed to have 
	already happened ( in, say, try_prism() ),
*/
    bd_star = beta_dif_star( Deriv ) ;
    gd_star = gamdif_star( Deriv, Pmat ) ;

/* 
	Calculate bounds on the sums of the absolute values
	of the elements in various blocks.
*/
    /* up & vp blocks */

    Pvp_star = 0.0 ;
    p1pt = Pmat + STAID_LEN + (MAT_DIM * DEG_FREE) ;
    end_mat = p1pt + (DEG_FREE * MAT_DIM) ;
    for( ; p1pt < end_mat ; p1pt += TWO_DF ) {
	for( end_row = p1pt + N_PARMS ; p1pt < end_row ; p1pt++ ) 
	    Pvp_star += fabs( *p1pt ) ;
    }

    star = gd_star + bd_star * Pvp_star ;
# if DEBUG
	printf( "gd** = %.4f, bd** = %.4f, Pvp** = %.4f \n", gd_star,
							     bd_star,
							     Pvp_star ) ;
	fflush( stdout ) ;
# endif

    b1pt = Bmat + STAID_LEN + N_PARMS + DEG_FREE ;
    b2pt = Bmat + STAID_LEN + N_PARMS + DEG_FREE + (MAT_DIM * DEG_FREE) ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	up_rows[j] = 0.0 ;
	vp_rows[j] = 0.0 ;
	for( end_row = b1pt + DEG_FREE ; b1pt < end_row ; 
					 b1pt++, b2pt++ ) {
	    up_rows[j] += fabs( *b1pt ) ;
	    vp_rows[j] += fabs( *b2pt ) ;
	}

	up_rows[j] *= star ;
	vp_rows[j] *= star ;

	b1pt += N_PARMS + DEG_FREE ;
	b2pt += N_PARMS + DEG_FREE ;
    }

/*
	Do the remaining blocks - those that actually arise 
	from the derivatives of the (u,v) -> (u',v') part of
	the map.  This section uses the mighty bound_rows(),
	which may be found below.
*/

	/* (u,u) block :
		B(u,u) * P(v,u) + B(u,v) * { beta * P(v,u) -
						    P(u,u) }
	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    p2pt = Pmat + STAID_LEN + N_PARMS ;

    b1pt = Bmat + STAID_LEN + N_PARMS ;
    b2pt = Bmat + STAID_LEN + N_PARMS + DEG_FREE ;
    bound_rows( uu_rows, b1pt, p1pt, b2pt, p2pt ) ;

	/* (u,v) block :
		B(u,u) * P(v,v) + B(u,v) * { beta * P(v,v) -
						    P(u,v) }

	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    p2pt = Pmat + STAID_LEN + N_PARMS + DEG_FREE ;

/* 	The same parts of Bmat as used to find uu_rows. */
    bound_rows( uv_rows, b1pt, p1pt, b2pt, p2pt ) ;

	/* (v,u) block :
		B(v,u) * P(v,u) + B(v,v) * { beta * P(v,u) -
						    P(u,u) }
	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS ;
    p2pt = Pmat + STAID_LEN + N_PARMS ;

    b1pt = Bmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS ;
    b2pt = Bmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    bound_rows( vu_rows, b1pt, p1pt, b2pt, p2pt ) ;

	/* (v,v) block :
		B(v,u) * P(v,v) + B(v,v) * { beta * P(v,v) -
						    P(u,v) }
	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    p2pt = Pmat + STAID_LEN + N_PARMS + DEG_FREE ;

/*	Same parts of Bmat as are used to find vu_rows. */
    bound_rows( vv_rows, b1pt, p1pt, b2pt, p2pt ) ;

/*
	Get the contributions to w[] arising from the error
	in computing the image of the prism's center.
*/
    for( j=0 ; j < DEG_FREE ; j++ ) {
	center_err[j+N_PARMS] = Bu_rows[j] * DBL_ERR ;
	center_err[j+N_PARMS+DEG_FREE] = Bv_rows[j] * DBL_ERR ;
    }

/*
	Compute the components of w[].
*/
    wu_pt = &w[N_PARMS] ;
    wv_pt = &w[N_PARMS + DEG_FREE] ;
    for( j=0 ; j < DEG_FREE ; j++, wu_pt++, wv_pt++ ) {
	*wu_pt = up_rows[j] + uu_rows[j] + uv_rows[j]  ;
	*wv_pt = vp_rows[j] + vu_rows[j] + vv_rows[j]  ;
    }

    /*
	Include center_err[].
    */

    for(j = N_PARMS ; j <MAT_DIM ; j++ )
	w[j] += center_err[j] ;

# if DEBUG
	printf("   p\t\t   u\t\t   v\n" ) ;
	for( j=0 ; j < DEG_FREE ; j++ ) 
	    printf( "%.4e\t%.4e\t%.4e\n", 
			up_rows[j],uu_rows[j],uv_rows[j] ) ;

	for( j=0 ; j < DEG_FREE ; j++ ) 
	    printf( "%.4e\t%.4e\t%.4e\n", 
			vp_rows[j],vu_rows[j],vv_rows[j] ) ;

	printf( "\n" ) ;
	fflush( stdout ) ;
# endif
    return ;
}
/* +++++++++++++++++++++++++++++++ */

void 
bound_rows (double *rows, double *first_b, double *first_p, double *second_b, double *second_p)
{
/*
	Obtain upper bounds on the sums of the absolute 
	values of rows of matricies given by expressions
	like:
	    B1 * S1 + B2 * ( [beta] * S1 - S2 ).

	Expressions like these arise in cr_rows() above.
	The idea is to cast these rows as bounded expressions
	and then use the usual machinery to find their limits.
*/
    int	    j, k ;
    double  *bpt_a, *bpt_b, *ppt_a, *ppt_b, *end_row, *cpt ;

/*
	Evaluate the constant part of the matrix expression.
	It's :
	    (B1 + 2.0 * B2) * S1  -  B2 * S2 
*/
    cpt = const_mat ;
    for( j=0 ; j < DEG_FREE ; j++ ) {

	bpt_a = first_b + j * MAT_DIM ;
	bpt_b = second_b + j * MAT_DIM ;
	for( k=0 ; k < DEG_FREE ; k++ ) {
	    *cpt = 0.0 ;

	    ppt_a = first_p + k ;
	    ppt_b = second_p + k ;
	    for( end_row = bpt_a + DEG_FREE ; bpt_a < end_row ; ) {
		*cpt += (*bpt_a + 2.0 * *bpt_b) * *ppt_a ;
		*cpt -= *bpt_b  *  *ppt_b ;

		bpt_a++, bpt_b++ ;
		ppt_a += MAT_DIM ;
		ppt_b += MAT_DIM ;
	    }
	    bpt_a -= DEG_FREE ;
	    bpt_b -= DEG_FREE ;
	    cpt++ ;
	}
    }

/*
	The more complicated part - it's :
		- B2 * H * S1

	where H is the Hessian of the perturbation to the generating 
	function.
*/
    cpt = const_mat ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	rows[j] = 0.0 ;

	bpt_a = second_b + j * MAT_DIM ;
	bpt_b = bpt_a + 1 ;
	for( k=0 ; k < DEG_FREE ; k++ ) {
	    ppt_a = first_p + k ;
	    ppt_b = ppt_a + MAT_DIM ;

/*   a * sin( v[0] ) term */
	    beta_prod.terms[0].coef =  - ( *bpt_a  *  *ppt_a ) ;

/*   c * sin( v[0] + v[1] )  term */
	    beta_prod.terms[1].coef = -( (*bpt_a + *bpt_b) *
					 (*ppt_a + *ppt_b) ) ;

/*   b * sin( v[0] + v[1] )  term */
	    beta_prod.terms[2].coef = - ( *bpt_b  *  *ppt_b ) ;

	    beta_prod.const_term = *cpt++ ;
	    bound_expr( &beta_prod ) ;

	    rows[j] += maxAbs(  beta_prod.bound.ub, 
				beta_prod.bound.lb ) ;
# if DEBUG
	    printf( "bound_rows( %d, %d) = %.4f \n", j, k, 
					maxAbs( beta_prod.bound.ub,
						beta_prod.bound.lb ) ) ;
	    fflush( stdout ) ;
# endif
	}
    }
# if DEBUG
    for( j=0 ; j < DEG_FREE ; j++ )
    printf( " %.3f \t %.3f \t \t%.3f \t %.3f \n", first_b[j*MAT_DIM], 
						  first_b[j*MAT_DIM+1],
						  first_p[j*MAT_DIM],
						  first_p[j*MAT_DIM+1] ) ;

    printf( "\n" ) ;
    for( j=0 ; j < DEG_FREE ; j++ )
    printf( " %.3f \t %.3f \t \t%.3f \t %.3f \n", second_b[j*MAT_DIM], 
						  second_b[j*MAT_DIM+1],
						  second_p[j*MAT_DIM],
						  second_p[j*MAT_DIM+1] ) ;
    printf( "\n" ) ;

    printf( "%.3f\t%.3f\n%.3f\t%.3f\n", const_mat[0], const_mat[1],
					const_mat[2], const_mat[3] ) ;
    fflush(stdout) ;
# endif
}
/* +++++++++++++++++++++++++++++++++++++++++ */

void 
Rcr_rows (APM *Rw, APM *Amat, APM *Deriv, RPrism *Priz)
/*
    Obtain bounds on the sums of the absolute values of
    the entries in the rows of 
       -1
    [A]   * Deriv * Pmat,

    put the results in w.
*/
{
    int		j ;
    APM		*end_row, *end_mat, *Pmat, *inv_pt ;
    APM		*p1pt, *p2pt, *b1pt, *b2pt, *wu_pt, *wv_pt ;

    Pmat = Priz->matrix ;
    Rset_inverse( Amat ) ;

/*
	Do up some row sums for the inverse; these
	are used to calculate center_err[].
*/
    b1pt = RBmat + STAID_LEN + N_PARMS ;
    b2pt = b1pt + MAT_DIM * DEG_FREE ;
    for( j=0  ; j < DEG_FREE ; j++ ) {
	apmAssign( RBu_rows[j], zero ) ;
	apmAssign( RBv_rows[j], zero ) ;

	for( end_row = b1pt + TWO_DF ; b1pt < end_row ; ) {
	    apmCalc( RBu_rows[j], RBu_rows[j], *b1pt++,
				  APM_ABS, APM_ADD, NULL ) ;
	    apmCalc( RBv_rows[j], RBv_rows[j], *b2pt++,
				  APM_ABS, APM_ADD, NULL ) ;
	}
    }

# if DEBUG
	printf( "\n" ) ;
	printRmat( RBmat ) ;
	Rcheck_inverse( Amat, RBmat ) ;
	printf( "\n" ) ;
# endif

/*
	Call functions which calculate upper bound on the 
	sums of the elements of various matrices.
	Before any bounding of matrices, one must invoke
	Rglobal_bounds( Pmat ) to set such global variables,
	as cos_one, and sin_sum.  This is done in Rtry_prism.
*/
    Rbeta_dif_star( Rbd_star, Deriv ) ;
    Rgamdif_star( Rgd_star, Deriv, Pmat ) ;

/* 
	Calculate bounds on the sums of the absolute values
	of the elements in various blocks.
*/
    /* up & vp blocks */

    apmAssignLong( RPvp_star, 0L, 0, BASE ) ;
    p1pt = Pmat + STAID_LEN + (MAT_DIM * DEG_FREE) ;
    end_mat = p1pt + (DEG_FREE * MAT_DIM) ;
    for( ; p1pt < end_mat ; p1pt += TWO_DF ) {
	for( end_row = p1pt + N_PARMS ; p1pt < end_row ; p1pt++ ) 
	    apmCalc( RPvp_star, RPvp_star, *p1pt, APM_ABS,
						  APM_ADD, NULL ) ;
    }

# if DEBUG
	printf( "Rgd** = %.4f, Rbd** = %.4f, RPvp** = %.4f \n", 
					apmtodbl( Rgd_star ),
					apmtodbl( Rbd_star ),
					apmtodbl( RPvp_star ) ) ;
	fflush( stdout ) ;
# endif

    apmCalc( Rstar, Rgd_star, Rbd_star, RPvp_star, 
			     APM_MUL, APM_ADD, NULL ) ;
    b1pt = RBmat + STAID_LEN + N_PARMS + DEG_FREE ;
    b2pt = RBmat + STAID_LEN + N_PARMS + DEG_FREE + (MAT_DIM * DEG_FREE) ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	apmAssignLong( Rup_rows[j], 0L, 0, BASE ) ;
	apmAssignLong( Rvp_rows[j], 0L, 0, BASE ) ; 
	for( end_row = b1pt + DEG_FREE ; b1pt < end_row ; 
					 b1pt++, b2pt++ ) {
	    apmCalc( Rup_rows[j], Rup_rows[j], *b1pt, APM_ABS,
						APM_ADD, NULL ) ;
	    apmCalc( Rvp_rows[j], Rvp_rows[j], *b2pt, APM_ABS,
						APM_ADD, NULL ) ;
	}

	apmCalc( Rup_rows[j], Rup_rows[j], Rstar, APM_MUL, NULL ) ;
	apmCalc( Rvp_rows[j], Rvp_rows[j], Rstar, APM_MUL, NULL ) ;

	b1pt += N_PARMS + DEG_FREE ;
	b2pt += N_PARMS + DEG_FREE ;
    }

/*
	Do the remaining blocks - those that actually arise 
	from the derivatives of the (u,v) -> (u',v') part of
	the map.  This section uses the mighty bound_rows(),
	which may be found below.
*/

	/* (u,u) block :
		B(u,u) * P(v,u) + B(u,v) * { beta * P(v,u) -
						    P(u,u) }
	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    p2pt = Pmat + STAID_LEN + N_PARMS ;

    b1pt = RBmat + STAID_LEN + N_PARMS ;
    b2pt = RBmat + STAID_LEN + N_PARMS + DEG_FREE ;
    Rbound_rows( Ruu_rows, b1pt, p1pt, b2pt, p2pt ) ;

	/* (u,v) block :
		B(u,u) * P(v,v) + B(u,v) * { beta * P(v,v) -
						    P(u,v) }

	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    p2pt = Pmat + STAID_LEN + N_PARMS + DEG_FREE ;

/* 	The same parts of RBmat as used to find uu_rows. */
    Rbound_rows( Ruv_rows, b1pt, p1pt, b2pt, p2pt ) ;

	/* (v,u) block :
		B(v,u) * P(v,u) + B(v,v) * { beta * P(v,u) -
						    P(u,u) }
	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS ;
    p2pt = Pmat + STAID_LEN + N_PARMS ;

    b1pt = RBmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS ;
    b2pt = RBmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    Rbound_rows( Rvu_rows, b1pt, p1pt, b2pt, p2pt ) ;

	/* (v,v) block :
		B(v,u) * P(v,v) + B(v,v) * { beta * P(v,v) -
						    P(u,v) }
	*/

    p1pt = Pmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;
    p2pt = Pmat + STAID_LEN + N_PARMS + DEG_FREE ;

/*	Same parts of RBmat as are used to find vu_rows. */
    Rbound_rows( Rvv_rows, b1pt, p1pt, b2pt, p2pt ) ;

/*
	Get the contibutions to Rw[] that arise from
	errors in the computation of the image of the
	prism's center.
*/
    for( j=0 ; j < DEG_FREE ; j++ ) {
	center_err[j+N_PARMS] = Bu_rows[j] * DBL_ERR ;
	center_err[j+N_PARMS+DEG_FREE] = Bv_rows[j] * DBL_ERR ;
	apmMultiply( Rcenter_err[j+N_PARMS], RBu_rows[j], max_error ) ;
	apmMultiply( Rcenter_err[j+N_PARMS+DEG_FREE], RBu_rows[j], 
						      max_error ) ;
    }

/*
	Compute the components of w[].
*/
    wu_pt = &Rw[N_PARMS] ;
    wv_pt = &Rw[N_PARMS + DEG_FREE] ;
    for( j=0 ; j < DEG_FREE ; j++, wu_pt++, wv_pt++ ) {
	apmCalc( *wu_pt, Rup_rows[j], Ruu_rows[j], Ruv_rows[j], max_error,
			 APM_ADD, APM_ADD, APM_ADD, NULL ) ;
	apmCalc( *wv_pt, Rvp_rows[j], Rvu_rows[j], Rvv_rows[j], max_error,
			 APM_ADD, APM_ADD, APM_ADD, NULL ) ;
    }

    /*
	Include errors due to miscalculation of
	prism's center.
    */

    for( j= N_PARMS ; j < MAT_DIM ; j++ ) 
	apmCalc( Rw[j], Rw[j], Rcenter_err[j], APM_ADD, NULL ) ;

# if DEBUG
	printf("   p\t\t   u\t\t   v\n" ) ;
	for( j=0 ; j < DEG_FREE ; j++ ) 
	    printf( "%.4e\t%.4e\t%.4e\n", apmtodbl( Rup_rows[j] ),
					  apmtodbl( Ruu_rows[j] ),
					  apmtodbl( Ruv_rows[j] ) ) ;

	for( j=0 ; j < DEG_FREE ; j++ ) 
	    printf( "%.4e\t%.4e\t%.4e\n", apmtodbl( Rvp_rows[j] ),
					  apmtodbl( Rvu_rows[j] ),
					  apmtodbl( Rvv_rows[j] ) ) ;

	printf( "\n" ) ;
	fflush( stdout ) ;
# endif
    return ;
}
/* +++++++++++++++++++++++++++++++ */

void 
Rbound_rows (APM *rows, APM *first_b, APM *first_p, APM *second_b, APM *second_p)
{
/*
	Obtain upper bounds on the sums of the absolute 
	values of rows of matricies given by expressions
	like:
	    B1 * S1 + B2 * ( [beta] * S1 - S2 ).

	Expressions like these arise in cr_rows() above.
	The idea is to cast these rows as bounded expressions
	and then use the usual machinery to find their limits.
*/
    int	    j, k ;
    APM     *bpt_a, *bpt_b, *ppt_a, *ppt_b, *end_row, *cpt ;

/*
	Evaluate the constant part of the matrix expression.
	It's :
	    (B1 + 2.0 * B2) * S1  -  B2 * S2 
*/
    cpt = Rconst_mat ;
    for( j=0 ; j < DEG_FREE ; j++ ) {

	bpt_a = first_b + j * MAT_DIM ;
	bpt_b = second_b + j * MAT_DIM ;
	for( k=0 ; k < DEG_FREE ; k++ ) {
	    apmAssignLong( *cpt, 0L, 0, BASE ) ;

	    ppt_a = first_p + k ;
	    ppt_b = second_p + k ;
	    for( end_row = bpt_a + DEG_FREE ; bpt_a < end_row ; ) {
		apmCalc( *cpt, *cpt, *bpt_a, 
				     *bpt_b, two, APM_MUL, 
				     APM_ADD,
				     *ppt_a, APM_MUL,
				     *bpt_b, *ppt_b,
				     APM_MUL, APM_SUB, 
				     APM_ADD, NULL ) ;

		bpt_a++, bpt_b++ ;
		ppt_a += MAT_DIM ;
		ppt_b += MAT_DIM ;
	    }

	    bpt_a -= DEG_FREE ;
	    bpt_b -= DEG_FREE ;
	    cpt++ ;
	}
    }

    cpt = Rconst_mat ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	apmAssignLong( rows[j], 0L, 0, BASE ) ;

	bpt_a = second_b + j * MAT_DIM ;
	bpt_b = bpt_a + 1 ;
	for( k=0 ; k < DEG_FREE ; k++ ) {
	    ppt_a = first_p + k ;
	    ppt_b = ppt_a + MAT_DIM ;

/*   a * sin( v[0] ) term */
	    apmMultiply( cr_scratch, *bpt_a, *ppt_a ) ;
	    apmNegate( Rbeta_prod.terms[0].coef, cr_scratch ) ;

/*   c * sin( v[0] + v[1] )  term */
	    apmCalc( cr_scratch, *bpt_a, *bpt_b, APM_ADD,
				 *ppt_a, *ppt_b, APM_ADD,
					APM_MUL, NULL ) ;
	    apmNegate( Rbeta_prod.terms[1].coef, cr_scratch ) ;

/*   b * sin( v[0] + v[1] )  term */
	    apmMultiply( cr_scratch, *bpt_b, *ppt_b ) ;
	    apmNegate( Rbeta_prod.terms[2].coef, cr_scratch ) ;

	    apmAssign( Rbeta_prod.const_term, *cpt++ ) ;
	    Rbound_expr( &Rbeta_prod ) ;

	    RmaxAbs( cr_scratch, Rbeta_prod.bound.ub, 
				 Rbeta_prod.bound.lb ) ;
# if DEBUG
	    printf( "bound_rows( %d, %d) = %.4f \n", j, k, 
					apmtodbl( cr_scratch ) ) ;
	    fflush( stdout ) ;
# endif
	    apmCalc( rows[j], rows[j], cr_scratch, APM_ADD, NULL ) ;
	}
    }
}
/* ++++++++++++++++++++++++++++++ */

void 
set_inverse (double *mat)
{
    double  *end_row, *end_block, *end_col ;
    double  *ipt_a, *ipt_b, *ipt_c, *ipt_set, *mpt_a, *mpt_b ;

/*
	If the matrix belongs to a new prism, set the components of
	the upper left corner, the part that corresponds to the
	variation in parameters.
*/
    if( isNewPrism == YES ) { 
		mpt_a = mat ;
		end_block = Bmat + N_PARMS * (MAT_DIM + 1) ;
		for( ipt_a=Bmat ; ipt_a < end_block ; ) {
			*ipt_a = 1.0 / *mpt_a ;

			mpt_a += MAT_DIM + 1 ;
			ipt_a += MAT_DIM + 1 ;
		}

		isNewPrism = NO ;
    }
    
    invert_corner( mat ) ;

/* 
    Set the (u,p) part of the inverse.
*/
    ipt_a = Bmat + STAID_LEN + N_PARMS ;
    ipt_b = Bmat + STAID_LEN + N_PARMS + DEG_FREE ;

    ipt_set = Bmat + STAID_LEN  ;
    end_block = ipt_set + (MAT_DIM * DEG_FREE) ;
    for( ; ipt_set < end_block ; ipt_set += TWO_DF ) {
	ipt_c = Bmat ;

	mpt_a = mat + STAID_LEN ;
	mpt_b = mat + STAID_LEN + (DEG_FREE * MAT_DIM) ;

	end_row = ipt_set + N_PARMS ;
	for( ; ipt_set < end_row ; ipt_set++ ) {
	    *ipt_set = 0.0 ;

	    end_col = mpt_a + (DEG_FREE * MAT_DIM) ;
	    for( ; mpt_a < end_col ; mpt_a += MAT_DIM ) {
		*ipt_set -= *ipt_a * *mpt_a  +  *ipt_b * *mpt_b ;

		ipt_a++ ;
		ipt_b++ ;
		mpt_b += MAT_DIM ;
	    }
	    *ipt_set *= *ipt_c ;

	    ipt_a -= DEG_FREE ;
	    ipt_b -= DEG_FREE ;
	    ipt_c += MAT_DIM + 1 ;

	    mpt_a -= (MAT_DIM * DEG_FREE) - 1 ;
	    mpt_b -= (MAT_DIM * DEG_FREE) - 1 ;
	}

	ipt_a += MAT_DIM ;
	ipt_b += MAT_DIM ;
	mpt_a -= DEG_FREE ;
	mpt_b -= DEG_FREE ;
    }

/* 
    Set the (v,p) part of the inverse.
*/
    ipt_a = Bmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS ;
    ipt_b = Bmat + STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS + DEG_FREE ;

    ipt_set = Bmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    end_block = ipt_set + (MAT_DIM * DEG_FREE) ;
    for( ; ipt_set < end_block ; ipt_set += TWO_DF ) {
	ipt_c = Bmat ;

	mpt_a = mat + STAID_LEN ;
	mpt_b = mat + STAID_LEN + (DEG_FREE * MAT_DIM) ;

	end_row = ipt_set + N_PARMS ;
	for( ; ipt_set < end_row ; ipt_set++ ) {
	    *ipt_set = 0.0 ;

	    end_col = mpt_a + (DEG_FREE * MAT_DIM) ;
	    for( ; mpt_a < end_col ; mpt_a += MAT_DIM ) {
		*ipt_set -= *ipt_a * *mpt_a  +  *ipt_b * *mpt_b ;

		ipt_a++ ;
		ipt_b++ ;
		mpt_b += MAT_DIM ;
	    }
	    *ipt_set *= *ipt_c ;

	    ipt_a -= DEG_FREE ;
	    ipt_b -= DEG_FREE ;
	    ipt_c += MAT_DIM + 1 ;

	    mpt_a -= (MAT_DIM * DEG_FREE) - 1 ;
	    mpt_b -= (MAT_DIM * DEG_FREE) - 1 ;
	}

	ipt_a += MAT_DIM ;
	ipt_b += MAT_DIM ;
	mpt_a -= DEG_FREE ;
	mpt_b -= DEG_FREE ;
    }
}
/* +++++++++++++++++++++ */

void 
invert_corner (double *mat)
{
/*
	Set up matrices to prepare 'em for use by gauss().
	Note that we use the matirx called const_mat[].  
	At the times this function is called const_mat[] 
	doesn't contain anything important.
*/
    int	     j ;
    double   *end_row, *mpt, *bpt, *cpt ;

/*
	Copy the matrix.
*/
    mpt = mat + STAID_LEN + N_PARMS ;
    for( j=0 ; j < TWO_DF ; j++ ) {

	cpt = copy_rows[j] ;
	end_row = mpt + TWO_DF ;
	while( mpt < end_row ) 
	    *cpt++ = *mpt++ ;

	mpt += N_PARMS ;
    }

/* 
	Do the inversion.
*/
    gauss( copy_rows ) ;

/*
	Copy the answer.
*/

    bpt = Bmat + STAID_LEN + N_PARMS ;
    for( j=0 ; j < TWO_DF ; j++ ) {
		cpt = copy_rows[j] ;
		end_row = bpt + TWO_DF ;
		while( bpt < end_row ) 
			*bpt++ = *cpt++ ;

		bpt += N_PARMS ;
    }
}
/* ++++++++++++++++++++++++++++++ */

void 
Rset_inverse (APM *mat)
{
    APM   *end_row, *end_block, *end_col ;
    APM   *ipt_a, *ipt_b, *ipt_c, *ipt_set, *mpt_a, *mpt_b ;

    if( isNewPrism == YES ) { 
		end_block = RBmat + N_PARMS * (MAT_DIM + 1) ;
		for( ipt_a=RBmat, mpt_a=mat ; ipt_a < end_block ; ) {
			apmDivide( *ipt_a, precision, (APM)NULL, one, *mpt_a ) ;

			mpt_a += MAT_DIM + 1 ;
			ipt_a += MAT_DIM + 1 ;
		}

		isNewPrism = NO ;
    }
    
    Rinvert_corner( mat ) ;

/* 
    Set the (u,p) part of the inverse.
*/
    ipt_a = RBmat + STAID_LEN + N_PARMS ;
    ipt_b = RBmat + STAID_LEN + N_PARMS + DEG_FREE ;

    ipt_set = RBmat + STAID_LEN  ;
    end_block = ipt_set + (MAT_DIM * DEG_FREE) ;
    for( ; ipt_set < end_block ; ipt_set += TWO_DF ) {
	ipt_c = RBmat ;

	mpt_a = mat + STAID_LEN ;
	mpt_b = mat + STAID_LEN + (DEG_FREE * MAT_DIM) ;

	end_row = ipt_set + N_PARMS ;
	for( ; ipt_set < end_row ; ipt_set++ ) {
	    apmAssignLong( *ipt_set, 0L, 0, BASE ) ; 

	    end_col = mpt_a + (DEG_FREE * MAT_DIM) ;
	    for( ; mpt_a < end_col ; mpt_a += MAT_DIM ) {
		apmCalc( *ipt_set, *ipt_a, *mpt_a, APM_MUL,
				   *ipt_b, *mpt_b, APM_MUL,
				   APM_ADD, APM_NEG, 
				   *ipt_set, APM_ADD, NULL ) ;

		ipt_a++ ;
		ipt_b++ ;
		mpt_b += MAT_DIM ;
	    }
	    apmCalc( *ipt_set, *ipt_set, *ipt_c, APM_MUL, NULL ) ;

	    ipt_a -= DEG_FREE ;
	    ipt_b -= DEG_FREE ;
	    ipt_c += MAT_DIM + 1 ;

	    mpt_a -= (MAT_DIM * DEG_FREE) - 1 ;
	    mpt_b -= (MAT_DIM * DEG_FREE) - 1 ;
	}

	ipt_a += MAT_DIM ;
	ipt_b += MAT_DIM ;
	mpt_a -= DEG_FREE ;
	mpt_b -= DEG_FREE ;
    }

/* 
    Set the (v,p) part of the inverse.
*/
    ipt_a = RBmat + STAID_LEN + N_PARMS + (DEG_FREE * MAT_DIM) ;
    ipt_b = RBmat + STAID_LEN + N_PARMS + (DEG_FREE*MAT_DIM) + DEG_FREE ;

    ipt_set = RBmat + STAID_LEN + (DEG_FREE * MAT_DIM)  ;
    end_block = ipt_set + (MAT_DIM * DEG_FREE) ;
    for( ; ipt_set < end_block ; ipt_set += TWO_DF ) {
	ipt_c = RBmat ;

	mpt_a = mat + STAID_LEN ;
	mpt_b = mat + STAID_LEN + (DEG_FREE * MAT_DIM) ;

	end_row = ipt_set + N_PARMS ;
	for( ; ipt_set < end_row ; ipt_set++ ) {
	    apmAssignLong( *ipt_set, 0L, 0, BASE ) ; 

	    end_col = mpt_a + (DEG_FREE * MAT_DIM) ;
	    for( ; mpt_a < end_col ; mpt_a += MAT_DIM ) {
		apmCalc( *ipt_set, *ipt_a, *mpt_a, APM_MUL,
				   *ipt_b, *mpt_b, APM_MUL,
				   APM_ADD, APM_NEG, 
				   *ipt_set, APM_ADD, NULL ) ;

		ipt_a++ ;
		ipt_b++ ;
		mpt_b += MAT_DIM ;
	    }
	    apmCalc( *ipt_set, *ipt_set, *ipt_c, APM_MUL, NULL ) ;

	    ipt_a -= DEG_FREE ;
	    ipt_b -= DEG_FREE ;
	    ipt_c += MAT_DIM + 1 ;

	    mpt_a -= (MAT_DIM * DEG_FREE) - 1 ;
	    mpt_b -= (MAT_DIM * DEG_FREE) - 1 ;
	}

	ipt_a += MAT_DIM ;
	ipt_b += MAT_DIM ;
	mpt_a -= DEG_FREE ;
	mpt_b -= DEG_FREE ;
    }
}
/* +++++++++++++++++++++ */

void 
Rinvert_corner (APM *mat)
{
/*
	Set up matrices to prepare 'em for use by Rgauss().
	Note that we use the matirx called const_mat[].  
	At the times this function is called const_mat[] 
	doesn't contain anything important.
*/
    int	     j ;
    APM      *end_row, *mpt, *bpt, *cpt ;
	

/*
	Copy the matrix.
*/
    mpt = mat + STAID_LEN + N_PARMS ;
    for( j=0 ; j < TWO_DF ; j++ ) {

	cpt = Rcopy_rows[j] ;
	end_row = mpt + TWO_DF ;
	while( mpt < end_row ) 
	    apmAssign( *cpt++, *mpt++ ) ;

	mpt += N_PARMS ;
    }

/* 
	Do the inversion.
*/
    Rgauss( Rcopy_rows ) ;

/*
	Copy the answer.
*/

    bpt = RBmat + STAID_LEN + N_PARMS ;
    for( j=0 ; j < TWO_DF ; j++ ) {

	cpt = Rcopy_rows[j] ;
	end_row = bpt + TWO_DF ;
	while( bpt < end_row ) 
	    apmAssign( *bpt++, *cpt++ ) ;

	bpt += N_PARMS ;
    }
}
/* +++++++++++++++++++++++++++ */

void 
MatMult (double *a, double *b)
{
    int     j, k, m ;
    double  *apt, *bpt, sum ;

    for( j=0 ; j < MAT_DIM ; j++ ) {
	for( k=0 ; k < MAT_DIM ; k++ ) {
	    sum = 0.0 ;
	    apt = a + j * MAT_DIM ;
	    bpt = b + k ;
	    for( m=0 ; m < MAT_DIM ; m++ ) {
		sum += *apt * *bpt ;
		apt++ ;
		bpt += MAT_DIM ;
	    }
	    printf( "%.2e  ", sum ) ;
	}
	printf( "\n" ) ;
    }
    printf( "\n" ) ;
    fflush( stdout ) ;
}
/* ++++++++++++++++++++++++++++++ */

void 
Rcheck_inverse (APM *a, APM *b)
{
    int     j, k, m ;
    APM  *apt, *bpt, sum ;

    sum = apmNew( BASE ) ;
    for( j=0 ; j < MAT_DIM ; j++ ) {
	for( k=0 ; k < MAT_DIM ; k++ ) {
	    apmAssignLong( sum, 0L, 0, BASE ) ;
	    apt = a + j * MAT_DIM ;
	    bpt = b + k ;
	    for( m=0 ; m < MAT_DIM ; m++ ) {
		apmCalc( sum, sum, *apt, *bpt, 
			 APM_MUL, APM_ADD, NULL ) ;
		apt++ ;
		bpt += MAT_DIM ;
	    }
	    printf( "%.2e  ", apmtodbl(sum) ) ;
	}
	printf( "\n" ) ;
    }
    printf( "\n" ) ;
    fflush( stdout ) ;

    apmDispose( sum ) ;
}
