# include  <stdio.h>
# include  <math.h>

# define  DELTA		0.01
# define  NEWT_TOL	1.0e-6
# define  MAX_JUMP	0.1

# define  PI  	     3.141592654
# define  HALF_PI    (PI / 2.0)

double  lam() ;
/* +++++++++++++++++++++++++++++++++ */

main()
{
    double   epsilon ; 
    double   a, b, c ;
    double   lam_min, lam_max ;

    for( epsilon = 0.035 ; epsilon < 0.05 ; epsilon += 0.001 ) {
	a = epsilon * 2.0 * PI * PI / 1.760173 ;
	b = a ;
	c = 2.0 * a ;

	Hessian( a, b, c, &lam_min, &lam_max ) ;
	printf( "epsilon = %.3f, lam_min = %.6f, lam_max = %.6f \n",
		epsilon, lam_min, lam_max ) ;
    }
}
/* +++++++++++++++++++++ */

setHermStart( a, b, c, x_start, y_start )

double	a, b, c, *x_start, *y_start ;
{
    double  two_c, x, y ;
    double  jump_sz, jump_scl, dx, dy ;
    double  gx, gy, hxx, hxy, hyy, hdet, tolerance ;

    two_c = 2.0 * c ;

    tolerance = NEWT_TOL * (a + b + c) ;

/*   
	Use Newton's method to try to find a minimum for the
	trace of the matrix beta.
*/

    x = HALF_PI ;
    y = HALF_PI ;

    do {
	/*	components of the gradient.	*/

	gx =  -a * cos( x )  -  two_c * cos( x + y ) ;
	gy =  -b * cos( y )  -  two_c * cos( x + y ) ;

	/* 	components of the Hessian 	*/

	hxx =  a * sin( x )  +  two_c * sin( x + y ) ;
	hxy =  two_c * sin( x + y ) ;
	hyy =  b * sin( y )  +  two_c * sin( x + y ) ;
	hdet = hxx * hyy  -  hxy * hxy ;

	/*  	A Newton's method step		*/
	if( hdet != 0.0 ) {
	    dx = (  gx * hyy  -  gy * hxy ) / hdet ;
	    dy = ( -gx * hxy  +  gy * hxx ) / hdet ;

	    if( (jump_sz = fabs(dx) + fabs(dy)) > MAX_JUMP ) {
		jump_scl = MAX_JUMP / jump_sz ;
		dx *= jump_scl ;
		dy *= jump_scl ;
	    }

	    x -= dx ;
	    y -= dy ;
		
	}
	else {
	    fprintf( stderr, "Death during Newton's method. \n" ) ;
	    exit(0) ;
	}
    } while( (fabs(gx) + fabs(gy))  > tolerance ) ;	

    *x_start = x ;
    *y_start = y ;
}
/* ++++++++++++++++++++++ */

Hessian( a , b, c, lam_min, lam_max )

double   a, b, c, *lam_min, *lam_max ;
{
    double  discrim, sqrt_disc, sqrt() ;
    double  a_sin, a_cos, b_sin, b_cos, c_sin, c_cos ;
    double  two_c, x, y ;
    double  jump_sz, jump_scl, dx, dy ;
    double  gx, gy, hxx, hxy, hyy, hdet, h_disc, htrc, tolerance ;
    double  dDisc_dx, dDisc_dy ;

    two_c = 2.0 * c ;

    tolerance = NEWT_TOL * (a + b + c) ;

/*   
	Start with a small displacement from the 
	position where the trace is maximized.
*/

    setHermStart( a, b, c, &x, &y ) ;
    do {
	/*	preliminaries			*/

	a_sin = a * sin( x ) ;
	b_sin = b * sin( y ) ;
	c_sin = two_c * sin( x + y ) ;

	a_cos = a * cos( x ) ;
	b_cos = b * cos( y ) ;
	c_cos = two_c * cos( x + y ) ;

	discrim = ( a_sin - b_sin ) * ( a_sin - b_sin ) +
					  c_sin * c_sin ;
	sqrt_disc = sqrt( discrim ) ;
	dDisc_dx = a_cos * (a_sin - b_sin) + c_cos * c_sin ;
	dDisc_dy = b_cos * (b_sin - a_sin) + c_cos * c_sin ;

	/*	components of the gradient.	*/

	gx = -a_cos - c_cos - dDisc_dx / sqrt_disc ;
	gy = -b_cos - c_cos - dDisc_dy / sqrt_disc ;

	/* 	components of the Hessian 	*/

	hxx = a_sin + c_sin +
	      ( a_sin * (a_sin - b_sin) - 
		a_cos * a_cos - c_cos * c_cos + 
		c_sin * c_sin ) / sqrt_disc
	      + dDisc_dx * dDisc_dx / (discrim * sqrt_disc) ;

	hxy = c_sin +
	      ( a_cos * b_cos + c_sin * c_sin -
				c_cos * c_cos ) / sqrt_disc
	      + dDisc_dx * dDisc_dy / (discrim * sqrt_disc) ;

	hyy = b_sin + c_sin +
	      ( b_sin * (b_sin - a_sin) - 
		b_cos * b_cos - c_cos * c_cos + 
		c_sin * c_sin ) / sqrt_disc
	      + dDisc_dy * dDisc_dy / (discrim * sqrt_disc) ;

	hdet = hxx * hyy  -  hxy * hxy ;
	htrc = hxx + hxy ;
	h_disc = htrc * htrc  - 4.0 * hdet ;

	/*  	A Newton's method step		*/
	if( hdet != 0.0 ) {
	    dx = (  gx * hyy  -  gy * hxy ) / hdet ;
	    dy = ( -gx * hxy  +  gy * hxx ) / hdet ;

	    if( (jump_sz = fabs(dx) + fabs(dy)) > MAX_JUMP ) {
		jump_scl = MAX_JUMP / jump_sz ;
		dx *= jump_scl ;
		dy *= jump_scl ;
	    }

	    x -= dx ;
	    y -= dy ;
	}
	else {
	    fprintf( stderr, "Death during Newton's method. \n" ) ;
	    exit(0) ;
	}
    } while( (fabs(gx) + fabs(gy))  > tolerance ) ;	

    *lam_min = (htrc - sqrt( h_disc )) / 2.0 ;
    *lam_max = (htrc + sqrt( h_disc )) / 2.0 ;
}
/* ++++++++++++++++++++++ */

