# include  <stdio.h>
# include  <math.h>
# include  "apm.h"
# include  "apmSpecial.h"
# include  "converse.h"

# define   DEBUG    NO
/* ++++++++++++++++++++++++++++++++++++++++++ */

void 
fxed_form (double *Amat, double *Deriv, double *Prizmat)
/*
     Prepares the matrix called "A" in my notes.  Eventually we want to
     have A = DF*Priz, but early in a calculation, when Priz is singular,
     we want to fatten A up by requiring it to have a certain fixed form.
     In the interest of speed we have coded the calculations below with 
     pointers.  Our hope is that the resulting function will scream along 
     at ultrasonic speed.  Unfortunately it is quite unreadable.
*/
{
    double	     *Aend, *Aend2, *Dend, *Pend, *Pend2  ;
    register double  *Apt, *Apt2, *Dpt, *Ppt, *Ppt2 ;

/*
	Copy the few terms which appear in the top rows of Amat.
*/
    Aend = Amat + N_PARMS * (MAT_DIM + 1) ;
    for( Apt = Amat, Ppt = Prizmat ; Apt < Aend ; Apt += (MAT_DIM + 1 ),
						  Ppt += (MAT_DIM + 1 ) ) 
	*Apt = *Ppt ;

# if DEBUG
    printf( "A(p,p) set \n" ) ;
# endif

/*
	Clear out those parts of Amat which change from iteration to 
	iteration. 
*/

    Aend = Amat + MAT_SZ ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt++ )
	*Apt = 0.0 ;

/*
	Set the (u,p) part of A
	It's equal to the (v,p) part of Prizmat.
*/
    
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt += TWO_DF ) {
	for( Pend = Ppt + N_PARMS ; Ppt < Pend ; Ppt++ )
	    *Apt++ = *Ppt ;

	Ppt += TWO_DF ;
    }

# if DEBUG
    printf( "A(u,p) set \n" ) ;
# endif
/*
	Set the (v,p) part - three terms.
*/
	/*  First term - equal to Deriv(v,p) * Prizmat(p,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Aend = Apt + (DEG_FREE*MAT_DIM) ; Apt < Aend ; Apt += TWO_DF ) {
	Ppt = Prizmat ;
	for( Dend = Dpt + N_PARMS ; Dpt < Dend ; Dpt++ ) {
	    *Apt++ += *Dpt * (*Ppt) ;
	    Ppt += MAT_DIM + 1 ;
	}

	Dpt += TWO_DF ;
    }

# if DEBUG
    printf( "A(v,p), first term, set \n" ) ;
# endif
	/*  Second term - equal to negative Prizmat(u,p) */

    Ppt = Prizmat + STAID_LEN ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Pend = Ppt + (DEG_FREE * MAT_DIM) ;  Ppt < Pend ; Ppt += TWO_DF ) {
	for( Aend = Apt + N_PARMS ; Apt < Aend ; Apt++ )
	    *Apt -= *Ppt++ ;

	Apt += TWO_DF ;
    }

# if DEBUG
    printf( "A(v,p), second term, set \n" ) ;
# endif
	/*  Third term - equal to Deriv(v,v) * Prizmat(v,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * (MAT_DIM + 1)) + N_PARMS ;
    Dend = Deriv + MAT_SZ ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    while( Dpt < Dend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
	Pend = Prizmat + MAT_SZ ;
	while( Ppt < Pend ) {
	    Aend = Apt + N_PARMS ;
	    while( Apt < Aend )
		*Apt++ += *Dpt * (*Ppt++) ;

	    Dpt++ ;
	    Ppt += TWO_DF ;
	    Apt -= N_PARMS ;
	}

	Dpt += N_PARMS + DEG_FREE ;
	Apt += MAT_DIM ;
    }

# if DEBUG
    printf( "A(v,p), third term, set \n" ) ;
# endif

/*
	(u,v) part
	equals Priz(v,u) + Priz(v,v)
*/

    Apt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    Ppt2 = Ppt + DEG_FREE ;
    while( Apt < Aend ) {
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend ) 
	    *Apt++ += ( *Ppt++ + *Ppt2++ ) ;

	Apt  += N_PARMS + DEG_FREE ;
	Ppt  += N_PARMS + DEG_FREE ;
	Ppt2 += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(u,v) set \n" ) ;
# endif

/*
	The (v,u) part
	equal to Deriv(v,v) * { Priz(v,u) + Priz(v,v) },
	which also equals Deriv(v, v) * A(u,v)
*/

    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Dend = Deriv + MAT_SZ ;
    while( Dpt < Dend ) {
        Apt2 = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
	Aend2 = Apt2 + (DEG_FREE * MAT_DIM) ;
	while( Apt2 < Aend2 ) {
	    Aend = Apt + DEG_FREE ;
	    while( Apt < Aend ) {
		*Apt++ += *Dpt * (*Apt2++) ;
	    }

	    Dpt++ ;
	    Apt -= DEG_FREE ;
	    Apt2 += DEG_FREE + N_PARMS ;
	}

	Apt += MAT_DIM ;
	Dpt += N_PARMS + DEG_FREE ;
    }
# if DEBUG
    printf( "A(v,u) set \n" ) ;
# endif

/*
	(v,v) part - equals Deriv(v,v) * Priz(v,v) - Priz(u,v) 
*/

	/* First term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Dend = Deriv + MAT_SZ ;
    while( Dpt < Dend ) {
        Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
	Pend = Prizmat + MAT_SZ ;
	while( Ppt < Pend ) {
	    Aend = Apt + DEG_FREE ;
	    while( Apt < Aend ) {
		*Apt++ += *Dpt * (*Ppt++) ;
	    }

	    Dpt++ ;
	    Apt -= DEG_FREE ;
	    Ppt += DEG_FREE + N_PARMS ;
	}

	Apt += MAT_DIM ;
	Dpt += N_PARMS + DEG_FREE ;
    }
	/* Second term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Ppt = Prizmat + STAID_LEN + N_PARMS + DEG_FREE ;
    Pend = Ppt + (MAT_DIM * DEG_FREE) ;
    while( Ppt < Pend ) {
	Aend = Apt + DEG_FREE ;
	while( Apt < Aend ) 
	    *Apt++ -= *Ppt++ ;

	Ppt += N_PARMS + DEG_FREE ;
	Apt += N_PARMS + DEG_FREE ;
    }

# if DEBUG
    printf( "A(v,v) set \n" ) ;
# endif
}
/* ++++++++++++++++++++++++ */

void 
Rfxed_form (APM *Amat, APM *Deriv, APM *Prizmat)
/*
     Prepares the matrix called "A" in my notes.  Eventually we want to
     have A = DF*Priz, but early in a calculation, when Priz is singular,
     we want to fatten A up by requiring it to have a certain fixed form.
     In the inerest of speed we have coded the calculations below in 
     terms of pointers.  Our hope is that the resulting function will
     scream along at ultrasonic speed.  Unfortunately it is quite
     unreadable.
*/
{
    APM		  *Aend, *Aend2, *Dend, *Pend, *Pend2  ;
    register APM  *Apt, *Apt2, *Dpt, *Ppt, *Ppt2 ;

/*
	Copy the few terms which appear in the top rows of Amat.
*/
    Aend = Amat + N_PARMS * (MAT_DIM + 1) ;
    for( Apt = Amat, Ppt = Prizmat ; Apt < Aend ; Apt += (MAT_DIM + 1 ),
						  Ppt += (MAT_DIM + 1 ) ) 
	apmAssign( *Apt, *Ppt ) ;

/*
	Clear out those parts of Amat which change from iteration to 
	iteration. 
*/

    Aend = Amat + MAT_SZ ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt++ )
	apmAssignLong( *Apt, 0L, 0, 0 ) ;

/*
	Set the (u,p) part of A
	It's equal to the (v,p) part of Prizmat.
*/
    
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) - TWO_DF ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Apt = Amat + STAID_LEN ; Apt < Aend ; Apt += TWO_DF ) {
	for( Pend = Ppt + N_PARMS ; Ppt < Pend ; Ppt++, Apt++ )
	    apmCalc( *Apt, *Apt, *Ppt, APM_ADD, NULL ) ;

	Ppt += TWO_DF ;
    }
/*
	Set the (v,p) part - three terms.
*/
	/*  First term - equal to Deriv(v,p) * Prizmat(p,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Aend = Apt + (DEG_FREE*MAT_DIM) ; Apt < Aend ; Apt += TWO_DF ) {
	Ppt = Prizmat ;
	for( Dend = Dpt + N_PARMS ; Dpt < Dend ; Dpt++ ) {
	    apmMultiply( *Apt++, *Dpt, *Ppt ) ;
	    Ppt += MAT_DIM + 1 ;
	}

	Dpt += TWO_DF ;
    }

	/*  Second term - equal to negative Prizmat(u,p) */

    Ppt = Prizmat + STAID_LEN ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( Pend = Ppt + (DEG_FREE * MAT_DIM) ; Ppt < Pend ; Ppt += TWO_DF ) {
	for( Aend = Apt + N_PARMS ; Apt < Aend ; Apt++, Ppt++ )
	    apmCalc( *Apt, *Apt, *Ppt, APM_SUB, NULL ) ;

	Apt += TWO_DF ;
    }

	/*  Third term - equal to Deriv(v,v) * Prizmat(v,p)  */

    Dpt = Deriv + STAID_LEN + (DEG_FREE * (MAT_DIM + 1)) + N_PARMS ;
    Dend = Deriv + MAT_SZ ;
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    while( Dpt < Dend ) {
	Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
	Pend = Prizmat + MAT_SZ - TWO_DF ;
	while( Ppt < Pend ) {
	    Aend = Apt + N_PARMS ;
	    while( Apt < Aend ) {
		apmCalc( *Apt, *Dpt, *Ppt, APM_MUL, *Apt, APM_ADD, NULL ) ;
		Apt++ ;
		Ppt++ ;
	    }

	    Dpt++ ;
	    Ppt += TWO_DF ;
	    Apt -= N_PARMS ;
	}

	Dpt += N_PARMS + DEG_FREE ;
	Apt += MAT_DIM ;
    }
/*
	(u,v) part
	equals Priz(v,u) + Priz(v,v)
*/

    Apt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    Aend = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    Ppt2 = Ppt + DEG_FREE ;
    while( Apt < Aend ) {
	Pend = Ppt + DEG_FREE ;
	while( Ppt < Pend )  {
	    apmCalc( *Apt, *Ppt, *Ppt2, APM_ADD, *Apt, APM_ADD, NULL ) ;
	    Apt++ ;
	    Ppt++ ;
	    Ppt2++ ;
	}

	Apt  += N_PARMS + DEG_FREE ;
	Ppt  += N_PARMS + DEG_FREE ;
	Ppt2 += N_PARMS + DEG_FREE ;
    }
/*
	The (v,u) part
	equal to Deriv(v,v) * { Priz(v,u) + Priz(v,v) },
	which also equals Deriv(v, v) * A(u,v)
*/

    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Dend = Deriv + MAT_SZ ;
    while( Dpt < Dend ) {
        Apt2 = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
	Aend2 = Apt2 + (DEG_FREE * MAT_DIM) ;
	while( Apt2 < Aend2 ) {
	    Aend = Apt + DEG_FREE ;
	    while( Apt < Aend ) {
		apmCalc( *Apt, *Apt, *Dpt, *Apt2, APM_MUL, APM_ADD, NULL ) ;
		Apt++ ;
		Apt2++ ;
	    }

	    Dpt++ ;
	    Apt -= DEG_FREE ;
	    Apt2 += DEG_FREE + N_PARMS ;
	}

	Apt += MAT_DIM ;
	Dpt += N_PARMS + DEG_FREE ;
    }

/*
	(v,v) part - equals Deriv(v,v) * Priz(v,v) - Priz(u,v) 
*/


	/* First term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Dpt = Deriv + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Dend = Deriv + MAT_SZ ;
    while( Dpt < Dend ) {
        Ppt = Prizmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
	Pend = Prizmat + MAT_SZ ;
	while( Ppt < Pend ) {
	    Aend = Apt + DEG_FREE ;
	    while( Apt < Aend ) {
		apmCalc( *Apt, *Apt, *Dpt, *Ppt, APM_MUL, APM_ADD, NULL ) ;

		Apt++ ;
		Ppt++ ;
	    }

	    Dpt++ ;
	    Apt -= DEG_FREE ;
	    Ppt += DEG_FREE + N_PARMS ;
	}

	Apt += MAT_DIM ;
	Dpt += N_PARMS + DEG_FREE ;
    }
	/* Second term */
    Apt = Amat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    Ppt = Prizmat + STAID_LEN + N_PARMS + DEG_FREE ;
    Pend = Ppt + (MAT_DIM * DEG_FREE) ;
    while( Ppt < Pend ) {
	Aend = Apt + DEG_FREE ;
	while( Apt < Aend ) {
	    apmCalc( *Apt, *Apt, *Ppt, APM_SUB, NULL ) ;

	    Apt++ ;
	    Ppt++ ;
	}

	Ppt += N_PARMS + DEG_FREE ;
	Apt += N_PARMS + DEG_FREE ;
    }
}
