# include <stdio.h>
# include <math.h>
# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"

# define  DEBUG	      NO 

# define  NUM_FACTS   5 
# define  NUM_TERMS   5 
# define  DET_TOL     1e-13

/* Functions private to this file */
static double beta_star(void);

/* Variables private to this file */
static APM	  Rerr_star ;
static APM	  ff_scratch ;
static APM	  Rcenter_err[MAT_DIM] ;
static APM	  Rdet_vu, Rdet_uv, Rstar ;
static APM	  RAvv_star, RAuvInv_star ;
static APM	  Rb_star, Rbd_star, Rgd_star ;
static APM	  RPvv_star, RPvp_star, RPvu_star ;

static double	center_err[MAT_DIM] ;
static Bdd_dbl	*ff_factors[NUM_FACTS] ;

static Bdd_term		ff_terms[NUM_TERMS] ;
static Bdd_expr		beta[3] ;

static Bdd_apm		*Rff_factors[NUM_FACTS] ;
static Bapm_term	Rff_terms[NUM_TERMS] ;
static Bapm_expr	Rbeta[3] ;
/* ++++++++++++++++++++++++++++++ */

void init_ffRows( void )
/*
	Set up the expressions and terms as described in my notes
	from 11/14.
*/
{
    int	      j, k ;
    Bdd_dbl   **dpt ;
    Bdd_apm   **apt ;
    Bdd_term  *tpt ;
    Bapm_term *Rtpt ;

/*
    Set up some APM's to be used to hold intermediate 
    results.
*/
    Rstar = apmNew( BASE ) ;
    Rdet_uv = apmNew( BASE ) ;
    Rdet_vu = apmNew( BASE ) ;
    Rb_star = apmNew( BASE ) ;
    Rbd_star = apmNew( BASE ) ;
    Rgd_star = apmNew( BASE ) ;
    Rerr_star = apmNew( BASE ) ;
    RAvv_star = apmNew( BASE ) ;
    RPvv_star = apmNew( BASE ) ;
    RPvp_star = apmNew( BASE ) ;
    RPvu_star = apmNew( BASE ) ;
    ff_scratch = apmNew( BASE ) ;
    RAuvInv_star = apmNew( BASE ) ;

    for( j = 0 ; j < MAT_DIM ; j++ )
	Rcenter_err[j] = apmNew( BASE ) ;

/*
	Set the number of terms in the bounded expressions
*/

    beta[0].nterms = Rbeta[0].nterms = 2 ;
    beta[1].nterms = Rbeta[1].nterms = 1 ;
    beta[2].nterms = Rbeta[2].nterms = 2 ;

/*	
	Assign terms
*/

    tpt = ff_terms ;
    Rtpt = Rff_terms ;
    for( j=0 ; j < 3 ; j++ ) {
	beta[j].terms = tpt ;
	Rbeta[j].terms = Rtpt ;
	tpt += beta[j].nterms ;
	Rtpt += Rbeta[j].nterms ;
    }

/*
	Set nfactors.
*/
    
    Rbeta[0].terms[0].nfactors = beta[0].terms[0].nfactors = 1  ;
    Rbeta[0].terms[1].nfactors = beta[0].terms[1].nfactors = 1  ;
    Rbeta[1].terms[0].nfactors = beta[1].terms[0].nfactors = 1  ;
    Rbeta[2].terms[0].nfactors = beta[2].terms[0].nfactors = 1  ;
    Rbeta[2].terms[1].nfactors = beta[2].terms[1].nfactors = 1  ;

/*
	Assign factors. 
*/

    dpt = ff_factors ;
    apt = Rff_factors ;
    for( j=0 ; j < 3 ; j++ ) {
		/* 
		beta
		*/
		for( k=0 ; k < beta[j].nterms ; k++ ) {
			beta[j].terms[k].factors = dpt ;
			Rbeta[j].terms[k].factors = apt ;

			dpt += beta[j].terms[k].nfactors ;
			apt += Rbeta[j].terms[k].nfactors ;
		}
    }

/*
	Set up those of the "bound" attributes which are 
	bounded APM's.
*/

    for( j=0 ; j < NUM_TERMS ; j++ ) {
		newBapm( Rff_terms[j].bound, BASE ) ;
    }

    for( j=0 ; j < 3 ; j++ ) {
		newBapm( Rbeta[j].bound, BASE ) ;
    }

/*
	Set up the terms and expressions. 
*/

/* beta */
 
	/* beta[0] = 2.0 - a * sin(v[0]) - c * sin(v[0] + v[1])  */
    beta[0].const_term = 2.0, Rbeta[0].const_term = two ;
	beta[0].terms[0].coef = -1.0 ;
	Rbeta[0].terms[0].coef = neg_one ;

	    beta[0].terms[0].factors[0] = &a_sin.bound ;
	    Rbeta[0].terms[0].factors[0] = &Ra_sin.bound ;

	beta[0].terms[1].coef = -1.0 ;
	Rbeta[0].terms[1].coef = neg_one ;

	    beta[0].terms[1].factors[0] = &c_sin.bound;
	    Rbeta[0].terms[1].factors[0] = &Rc_sin.bound;

	/* beta[1] = - 2.0 * c * sin( v[0] + v[1] )	*/
    beta[1].const_term = 0.0, Rbeta[1].const_term = zero ;
	beta[1].terms[0].coef = -2.0 ;
	Rbeta[1].terms[0].coef = neg_two ;

	    beta[1].terms[0].factors[0] = &c_sin.bound;
	    Rbeta[1].terms[0].factors[0] = &Rc_sin.bound;

	/* beta[2] = 2.0 - b * sin(v[1]) - c * sin(v[1] + v[0])  */
    beta[2].const_term = 2.0, Rbeta[2].const_term = two ;
	beta[2].terms[0].coef = -1.0 ;
	Rbeta[2].terms[0].coef = neg_one ;

	    beta[2].terms[0].factors[0] = &b_sin.bound;
	    Rbeta[2].terms[0].factors[0] = &Rb_sin.bound;

	beta[2].terms[1].coef = -1.0 ;
	Rbeta[2].terms[1].coef = neg_one ;

	    beta[2].terms[1].factors[0] = &c_sin.bound ;
	    Rbeta[2].terms[1].factors[0] = &Rc_sin.bound ;
}
/* +++++++++++++++++++++++++++++++++ */

void 
ff_rows (double *w, double *Amat, double *Deriv, Prism *Priz)
/*
    Obtain bounds on the sums of the absolute values of
    the entries in the rows of 
       -1
    [A]   * Deriv * Pmat,

    put the results in w.
*/
{
    double	*apt, *mpt, *end_row, *end_mat, *Pmat ;
    double	det_vu, det_uv, star ;
    double	Avv_star, AuvInv_star ;
    double	b_star, bd_star, gd_star ;
    double	Pvv_star, Pvp_star, Pvu_star ;

/*
	Check that A(u,v) is invertible.  If not, die.
*/
    Pmat = Priz->matrix ;

    apt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    det_uv = *apt * ( *(apt + MAT_DIM + 1)) ;
    apt++ ;
    det_uv -= *apt * ( *(apt + MAT_DIM -1)) ;

    if( fabs( det_uv ) < DET_TOL ) {
	fprintf( stderr, 
		 "The determinant of A(u,v) = %.14e. Died. \n",
							det_uv ) ;
	cease() ;
    }
/*
	Call functions which calculate upper bound on the 
	sums of the elements of various matrices.
	Before any bounding of matrices, one must invoke
	global_bounds( Pmat ) to set such global variables,
	as cos_one, and sin_sum.  It is called in try_prism().
*/
    b_star = beta_star() ;
    bd_star = beta_dif_star( Deriv ) ;
    gd_star = gamdif_star( Deriv, Pmat ) ;

# if DEBUG
    printf( "beta** = %.4f,  Dbeta** = %.4f, Dgamma** = %.4f \n",
							  b_star,
							  bd_star,
							  gd_star ) ;
# endif

/*
	Find sums of the absolute values of the entries
	of A(v,v), AInv(u,v), Pmat(v,v), Pmat(v,u), and Pmat(v,p)
*/
    end_mat = Pmat + MAT_SZ ;

    Pvv_star = 0.0 ;
    mpt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    for( ; mpt < end_mat ; mpt += (N_PARMS + DEG_FREE) ) {
	for( end_row = mpt + DEG_FREE ; mpt < end_row ; mpt++ ) {
	    Pvv_star += fabs( *mpt ) ;
	}
    }

    Pvu_star = 0.0 ;
    mpt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    for( ; mpt < end_mat ; mpt += (N_PARMS + DEG_FREE) ) {
	for( end_row = mpt + DEG_FREE ; mpt < end_row ; mpt++ ) {
	    Pvu_star += fabs( *mpt ) ;
	}
    }


    Pvp_star = 0.0 ;
    mpt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( ; mpt < end_mat ; mpt += TWO_DF ) {
	for( end_row = mpt + N_PARMS ; mpt < end_row ; mpt++ ) {
	    Pvp_star += fabs( *mpt ) ;
	}
    }

    Avv_star = SmBlock_err ;
    mpt = Amat + STAID_LEN + DEG_FREE * MAT_DIM + DEG_FREE + N_PARMS ;
    for( ; mpt < end_mat ; mpt += TWO_DF ) {
	for( end_row = mpt + N_PARMS ; mpt < end_row ; mpt++ ) {
	    Avv_star += fabs( *mpt ) ;
	}
    }

    AuvInv_star = SmBlock_err ;
    mpt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    for( ; mpt < end_mat ; mpt += TWO_DF ) {
	for( end_row = mpt + N_PARMS ; mpt < end_row ; mpt++ ) {
	    AuvInv_star += fabs( *mpt ) ;
	}
    }

    AuvInv_star /= det_uv ;

/*
	Check that A(v,u) is invertible.  If not, die.
	If it is, set the harder-to-compute elements of w.
*/

    apt = Amat + STAID_LEN + N_PARMS + (DEG_FREE * MAT_DIM) ;
    det_vu = *apt * ( *(apt + MAT_DIM + 1)) ;
    apt++ ;
    det_vu -= *apt * ( *(apt + MAT_DIM -1)) ;

    if( fabs( det_vu ) < DET_TOL ) {
	fprintf( stderr, 
		 "The determinant of A(v,u) = %.14e. Died. \n",
							det_vu ) ;
	cease() ;
    }
    else {
	w[3] =  fabs( Amat[MAT_SZ - DEG_FREE - 1] ) +
		fabs( Amat[STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + 1] ) +
		DBL_ERR ;
	w[4] =  fabs( Amat[MAT_SZ - TWO_DF] ) +
		fabs( Amat[STAID_LEN + (DEG_FREE*MAT_DIM) + N_PARMS] ) +
		DBL_ERR ;

	center_err[3] = w[3] * (1 + Avv_star * AuvInv_star) * DBL_ERR ; 
	center_err[4] = w[4] * (1 + Avv_star * AuvInv_star) * DBL_ERR ;
	center_err[5] = AuvInv_star * DBL_ERR ;
	center_err[6] = center_err[5] ;

	star =  (gd_star + bd_star * (Pvp_star + Pvv_star) +
			   b_star * Pvu_star) / det_vu ;
	w[3] *= star + center_err[3] / det_vu ;
	w[4] *= star + center_err[4] / det_vu ;
	w[5] = 1.0 + center_err[5] ;
	w[6] = 1.0 + center_err[6] ;
    }

# if DEBUG
    printf( "det( A(v,u) ) = %.4f \n", det_vu ) ;
    printf( "w[3] = %.4f \t w[4] = %.4f \n", w[3], w[4] ) ;
    printf( "Pvv = %.4f \t Pvu = %.4f \t Pvp = %.4f \n", Pvv_star,
							 Pvu_star,
							 Pvp_star ) ;
    fflush( stdout ) ;
# endif

    return ;
}
/* +++++++++++++++++++++++++++++++ */

double beta_star( void )
{
    double  answer ;

    bound_expr( &beta[0] ) ;
    bound_expr( &beta[1] ) ;
    bound_expr( &beta[2] ) ;

    answer = maxAbs( beta[0].bound.ub, beta[0].bound.lb ) +
	     maxAbs( beta[1].bound.ub, beta[1].bound.lb ) +
	     maxAbs( beta[2].bound.ub, beta[2].bound.lb )  ;

    return( answer ) ;
}
/* +++++++++++++++++++++++++++++ */

void 
Rff_rows (APM *w, APM *Amat, APM *Deriv, RPrism *Priz)
/*
    Obtain bounds on the sums of the absolute values of
    the entries in the rows of 
       -1
    [A]   * Deriv * Pmat,

    put the results in w.
*/
{
    APM   *apt, *mpt, *end_row, *end_mat, *Pmat ;

/*
	Check that A(u,v) is invertible.  If not, die.
*/
    Pmat = Priz->matrix ;

    apt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    apmMultiply( Rdet_uv, *apt, *(apt + MAT_DIM + 1) ) ;
    apt++ ;
    apmCalc( Rdet_uv, Rdet_uv, *apt, *(apt + MAT_DIM -1), 
				APM_MUL, APM_SUB, NULL ) ;
    apmAbsoluteValue( ff_scratch, Rdet_uv ) ;

    if( apmCompare( ff_scratch, max_error ) != 1 ) {
	fprintf( stderr, 
		 "The determinant of A(u,v) is too small. Died. \n" ) ;
	fprintf( stderr, "\t %.12e \n", apmtodbl( ff_scratch ) ) ;
	cease() ;
    }
/*
	Call functions which calculate upper bound on the 
	sums of the elements of various matrices.
	Before any bounding of matrices, one must invoke
	Rglobal_bounds( Pmat ) to set such global variables,
	as cos_one, and sin_sum.  It is called in Rtry_prism().
*/
    Rbeta_star( Rb_star ) ;
    Rbeta_dif_star( Rbd_star, Deriv ) ;
    Rgamdif_star( Rgd_star, Deriv, Pmat ) ;

# if DEBUG
    printf( "beta** = %.4f,  Dbeta** = %.4f, Dgamma** = %.4f \n",
						apmtodbl(Rb_star),
						apmtodbl(Rbd_star),
						apmtodbl(Rgd_star) ) ;
# endif

/*
	Find sums of the absolute values of the entries
	of Pmat(v,v), Pmat(v,u), and Pmat(v,p)
*/
    end_mat = Pmat + MAT_SZ ;

    apmAssign( RPvv_star, zero ) ; 
    mpt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS + DEG_FREE ;
    for( ; mpt < end_mat ; mpt += (N_PARMS + DEG_FREE) ) {
	for( end_row = mpt + DEG_FREE ; mpt < end_row ; mpt++ ) {
	    apmCalc( RPvv_star, RPvv_star, *mpt, APM_ABS, 
						   APM_ADD, NULL ) ;
	}
    }

    apmAssign( RPvu_star, zero ) ;
    mpt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) + N_PARMS ;
    for( ; mpt < end_mat ; mpt += (N_PARMS + DEG_FREE) ) {
	for( end_row = mpt + DEG_FREE ; mpt < end_row ; mpt++ ) {
	    apmCalc( RPvu_star, RPvu_star, *mpt, APM_ABS,
						   APM_ADD, NULL ) ;
	}
    }

    apmAssign( RPvp_star, zero ) ; 
    mpt = Pmat + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( ; mpt < end_mat ; mpt += TWO_DF ) {
	for( end_row = mpt + N_PARMS ; mpt < end_row ; mpt++ ) {
	    apmCalc( RPvp_star, RPvp_star, *mpt, APM_ABS,
						   APM_ADD, NULL ) ;
	}
    }

    apmAssign( RAvv_star, RSmBlock_err ) ;
    mpt = Amat + STAID_LEN + DEG_FREE * MAT_DIM + DEG_FREE + N_PARMS ;
    for( ; mpt < end_mat ; mpt += TWO_DF ) {
	for( end_row = mpt + N_PARMS ; mpt < end_row ; mpt++ ) {
	    apmCalc( RAvv_star, RAvv_star, *mpt, 
				APM_ABS, APM_ADD, NULL ) ;
	}
    }

    apmAssign( RAuvInv_star, RSmBlock_err ) ;
    mpt = Amat + STAID_LEN + N_PARMS + DEG_FREE ;
    for( ; mpt < end_mat ; mpt += TWO_DF ) {
	for( end_row = mpt + N_PARMS ; mpt < end_row ; mpt++ ) {
	    apmCalc( RAuvInv_star, RAuvInv_star, *mpt, 
				   APM_ABS, APM_ADD, NULL ) ;
	}
    }

    apmDivide( ff_scratch, precision, (APM) NULL, 
			   RAuvInv_star, Rdet_uv ) ;
    apmAssign( RAuvInv_star, ff_scratch ) ;

/*
	Check that A(v,u) is invertible.  If not, die.
	If it is, set the harder-to-compute elements of w.
*/

    apt = Amat + STAID_LEN + N_PARMS + (DEG_FREE * MAT_DIM) ;
    apmMultiply( Rdet_vu, *apt, *(apt + MAT_DIM + 1) ) ;
    apt++ ;
    apmCalc( Rdet_vu, Rdet_vu, *apt, *(apt + MAT_DIM - 1),
				     APM_MUL, APM_SUB, NULL ) ;
    apmAbsoluteValue( ff_scratch, Rdet_vu ) ;

    if( apmCompare( ff_scratch, max_error ) != 1 ) {
	fprintf( stderr, 
		 "The determinant of A(v,u) is too small. Died. \n") ;
	fprintf( stderr, "\t %.12e \n", apmtodbl( ff_scratch ) ) ;
	cease() ;
    }
/*
	Note that the sums below seem to contain some misplaced
	elements of Amat.  These are to be thought of as elements
	of A(v,u) inverse.
*/
    else {
	apmCalc( w[3], Amat[MAT_SZ-DEG_FREE-1], APM_ABS, 
		       Amat[STAID_LEN+(DEG_FREE*MAT_DIM)+N_PARMS+1],
		       APM_ABS, max_error, APM_ADD, APM_ADD, NULL ) ;

	apmCalc( w[4], Amat[MAT_SZ-TWO_DF], APM_ABS, 
		       Amat[STAID_LEN+(DEG_FREE*MAT_DIM)+N_PARMS],
		       APM_ABS, max_error, APM_ADD, APM_ADD, NULL ) ;

	apmCalc( Rerr_star, RAvv_star, RAuvInv_star, APM_MUL, 
				       one, APM_ADD, NULL );
	apmCalc( Rcenter_err[3], w[3], Rerr_star, max_error,
				       APM_MUL, APM_MUL, NULL ) ;
	apmCalc( Rcenter_err[4], w[4], Rerr_star, max_error, 
				       APM_MUL, APM_MUL, NULL ) ;
	apmMultiply( Rcenter_err[5], RAuvInv_star, max_error ) ;
	apmAssign( Rcenter_err[6], Rcenter_err[5] ) ;

	apmCalc( Rstar, RPvp_star, RPvv_star, APM_ADD,
			Rbd_star, APM_MUL,
			Rb_star, RPvu_star, APM_MUL,
			Rgd_star, APM_ADD, APM_ADD, NULL ) ;

	apmCalc( ff_scratch, Rcenter_err[3], Rstar, w[3], 
				 APM_MUL, APM_ADD, NULL ) ;
	apmDivide( w[3], precision, (APM) NULL, ff_scratch, Rdet_vu ) ;
	apmCalc( ff_scratch, Rcenter_err[4], Rstar, w[4], 
				 APM_MUL, APM_ADD, NULL ) ;
	apmDivide( w[4], precision, (APM) NULL, ff_scratch, Rdet_vu ) ;
	apmAdd( w[5], one, Rcenter_err[5] ) ;
	apmAdd( w[6], one, Rcenter_err[6] ) ;
    }
# if DEBUG
    printf( "det( A(v,u) ) = %.4f \n", apmtodbl(Rdet_vu) ) ;
    printf( "w[3] = %.4f \t w[4] = %.4f \n", apmtodbl(w[3]), 
					     apmtodbl(w[4]) ) ;
    printf( "Pvv = %.4f \t Pvu = %.4f \t Pvp = %.4f \n", 
						apmtodbl(RPvv_star),
						apmtodbl(RPvu_star),
						apmtodbl(RPvp_star) ) ;
    fflush( stdout ) ;
# endif

    return ;
}
/* +++++++++++++++++++++++++++++++ */

void 
Rbeta_star (APM answer)
{
    Rbound_expr( &Rbeta[0] ) ;
    Rbound_expr( &Rbeta[1] ) ;
    Rbound_expr( &Rbeta[2] ) ;

    RmaxAbs( answer, Rbeta[0].bound.ub, Rbeta[0].bound.lb )  ;
    RmaxAbs( Rrow_abs[0], Rbeta[1].bound.ub, Rbeta[1].bound.lb ) ;
    RmaxAbs( Rrow_abs[1], Rbeta[2].bound.ub, Rbeta[2].bound.lb ) ;

    apmCalc( answer, answer, Rrow_abs[0], Rrow_abs[1], 
					APM_ADD, APM_ADD, NULL ) ; 
}
/* +++++++++++++++++++++++++++++ */
