# include <stdio.h>
# include <math.h>
# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"

# define    DEBUG    NO	

/* Variables shared through converse.h */
APM		RfirstLeastLam, RminLeastLam, RmaxLeastLam, RsumTinyLams ;
double	firstLeastLam, minLeastLam, maxLeastLam, sumTinyLams ;

/* Variables private to this file */
static APM	    Ra_term, Rb_term, Rc_term ;
static APM	    Rtrace_ll, RminBlam_ll, RmaxBlam_ll, Rdenom ;

static Bdd_apm		RBtrace, RminLam, RmaxLam ;
static RPrism		*earliest ;

static Bdd_dbl		discrim ;
static Bdd_dbl		a_sq, b_sq, c_sq ;
static Bdd_dbl	    *lamFacts[2] ;
static Bdd_term		ab_term ;

static APM			lam_scratch ;
static Bdd_apm		Rdiscrim ;
static Bdd_apm		Ra_sq, Rb_sq, Rc_sq ;
static Bdd_apm		*RlamFacts[2] ;
static Bapm_term	Rab_term ;

/* Functions private to this file */
static void setSq(Bdd_dbl *xsq, Bdd_dbl *x);
static void RsetSq(Bdd_apm *xsq, Bdd_apm *x);
/*+++++++++++++++++++++++++++++++++ */

void 
initLambda (void)
{
/*
	Do up the APMs
*/
    Ra_term = apmNew( BASE ) ;
    Rb_term = apmNew( BASE ) ;
    Rc_term = apmNew( BASE ) ;

    Rdenom = apmNew( BASE ) ;
    Rtrace_ll = apmNew( BASE ) ;
    Rsqrt_disc = apmNew( BASE ) ;
    RminBlam_ll = apmNew( BASE ) ;
    RmaxBlam_ll = apmNew( BASE ) ;

    RminLeastLam = apmNew( BASE ) ;
    RmaxLeastLam = apmNew( BASE ) ;
    RsumTinyLams = apmNew( BASE ) ;
    RfirstLeastLam = apmNew( BASE ) ;

    newBapm( Ra_sq, BASE ) ;
    newBapm( Rb_sq, BASE ) ;
    newBapm( Rc_sq, BASE ) ;
    newBapm( RmaxLam, BASE ) ;
    newBapm( RminLam, BASE ) ;
    newBapm( RBtrace, BASE ) ;
    newBapm( Rdiscrim, BASE ) ;

    four = apmInit( 4L, 0, BASE ) ;
    lam_scratch = apmNew( BASE ) ;

    earliest = conjureRPrism() ;
/*
	Set up the terms.
*/

    ab_term.nfactors = Rab_term.nfactors = 2 ;
    ab_term.factors = lamFacts ;
    Rab_term.factors = RlamFacts ;
    ab_term.coef = -2.0 ;
    Rab_term.coef = apmInit( -2L, 0, BASE ) ;
    newBapm( Rab_term.bound, BASE ) ;
	ab_term.factors[0] = &a_sin.bound ;
	ab_term.factors[1] = &b_sin.bound ;
	Rab_term.factors[0] = &Ra_sin.bound ;
	Rab_term.factors[1] = &Rb_sin.bound ;
}
/* ++++++++++++++++++++++++ */

void 
bd_Blams (Bdd_dbl *leastBlam, Bdd_dbl *bigBlam, Bdd_dbl *trace)
{

	/* Bound the terms for the discriminant. */
    setSq( &a_sq, &a_sin.bound ) ;
    setSq( &b_sq, &b_sin.bound ) ;
    setSq( &c_sq, &c_sin.bound ) ;
    bound_term( &ab_term ) ;

	/* Bound the discriminant itself.  */
    discrim.lb = ab_term.bound.lb + a_sq.lb + 
			 b_sq.lb + 4.0 * c_sq.lb ;
    if( discrim.lb < 0.0 )
	discrim.lb = 0.0 ;
    discrim.ub = ab_term.bound.ub + a_sq.ub + 
			 b_sq.ub + 4.0 * c_sq.ub ;
    if( discrim.ub < 0.0 )
	discrim.ub = 0.0 ;

	/* Bound the least eigenvalue. */
    leastBlam->ub = ( trace->ub - sqrt(discrim.lb) ) / 2.0 ;
    leastBlam->lb = ( trace->lb - sqrt(discrim.ub) ) / 2.0 ;

    bigBlam->ub = ( trace->ub + sqrt(discrim.ub) ) / 2.0 ;
    bigBlam->lb = ( trace->lb + sqrt(discrim.lb) ) / 2.0 ;

# if DEBUG
    printf( "%.6f < leastBlam < %.6f \n", leastBlam->lb, leastBlam->ub ) ;
    printf( "%.6f < bigBlam < %.6f \n", bigBlam->lb, bigBlam->ub ) ;
    fflush( stdout ) ;
# endif
}
/* ++++++++++++++++++++++++++++++++++++++++++++ */

void 
Rbd_Blams (Bdd_apm *leastBlam, Bdd_apm *bigBlam, Bdd_apm *trace)
/*
	An APM partner to bd_Blams ;
*/
{
/* Bound the terms for the discriminant. */
    RsetSq( &Ra_sq, &Ra_sin.bound ) ;
    RsetSq( &Rb_sq, &Rb_sin.bound ) ;
    RsetSq( &Rc_sq, &Rc_sin.bound ) ;
    Rbound_term( &Rab_term ) ;

/* Bound the discriminant itself. */

	/* lower bound */
    apmCalc( Rdiscrim.lb, Ra_sq.lb, Rb_sq.lb, APM_ADD,
			  four, Rc_sq.lb, APM_MUL, APM_ADD,
			  Rab_term.bound.lb, APM_ADD, NULL ) ;

    if( apmCompare( Rdiscrim.lb, zero ) < 1 )
	apmAssign( Rdiscrim.lb, zero ) ;

	/* upper bound */
    apmCalc( Rdiscrim.ub, Ra_sq.ub, Rb_sq.ub, APM_ADD,
			  four, Rc_sq.ub, APM_MUL, APM_ADD,
			  Rab_term.bound.ub, APM_ADD, NULL ) ;

    if( apmCompare( Rdiscrim.ub, zero ) < 1 )
	apmAssign( Rdiscrim.ub, zero ) ;

/* Do up the final bounds on the eigenvalues. 
		First do those requiring 
		sqrt( discrim.lb ). 
*/
    apmSqrt( Rsqrt_disc, precision, Rdiscrim.lb ) ;
    apmCalc( lam_scratch, trace->ub, Rsqrt_disc, APM_SUB, 
					max_error, APM_ADD, NULL ) ;
    apmDivide( leastBlam->ub, precision, (APM)NULL, lam_scratch, two ) ;

    apmCalc( lam_scratch, trace->lb, Rsqrt_disc, APM_ADD, 
					max_error, APM_SUB, NULL ) ;
    apmDivide( bigBlam->lb, precision, (APM)NULL, lam_scratch, two ) ;


/*
		Next those requiring
		sqrt( discrim.lb )
*/
    apmSqrt( Rsqrt_disc, precision, Rdiscrim.ub ) ;
    apmCalc( lam_scratch, trace->lb, Rsqrt_disc, APM_SUB, 
					max_error, APM_SUB, NULL ) ;
    apmDivide( leastBlam->lb, precision, (APM)NULL, lam_scratch, two ) ;

    apmCalc( lam_scratch, trace->ub, Rsqrt_disc, APM_ADD, 
					max_error, APM_ADD, NULL ) ;
    apmDivide( bigBlam->ub, precision, (APM)NULL, lam_scratch, two ) ;

# if DEBUG
    printf( "%.6f < RleastBlam < %.6f \n", apmtodbl( leastBlam->lb ), 
					   apmtodbl( leastBlam->ub ) ) ;
    printf( "%.6f < RbigBlam < %.6f \n", apmtodbl( bigBlam->lb ), 
					 apmtodbl( bigBlam->ub ) ) ;
    fflush( stdout ) ;
# endif
}
/* ++++++++++++++++++++++++++++ */

void 
setLLbounds ( 
/*
    Get bounds on the least eigenvalue of the variation of the action
    functional.  This is equivalent to the summer's estimate of the 
    value of size of the perturbation for which no  minimizing state
    can include the maximum of the perturbation.
*/
    RPrism *priz
)
{
    int	    j, mat_pos ;
    APM	    *pmat_pos ;

    for( j=0 ; j < N_PARMS ; j++ ) 
	apmAssign( earliest->center->p[j], priz->center->p[j] ) ;

	mat_pos =  j * (MAT_DIM + 1) ;
	apmAssign( earliest->matrix[mat_pos],  priz->matrix[mat_pos] ) ;

    for( j=0 ; j < DEG_FREE ; j++ )  
	apmAssign( earliest->center->z.v[j], priz->center->z.u[j] ) ;

    Rglobal_bounds( earliest ) ;

/*
	Account for the imprecision of the starting point 
	and the variation of the parameters.
*/
    apmAssignLong( lam_scratch, 0L, 0, BASE ) ;
    pmat_pos = priz->matrix ;

    for( j=0 ; j < N_PARMS ; j++ ) {
	apmCalc( lam_scratch, lam_scratch, 
			      priz->center->p[j] , Rstart_size,
			      APM_MUL, APM_ADD,
			      *pmat_pos,
			      APM_ABS, APM_ADD, NULL ) ;
	pmat_pos += 1 + MAT_DIM ;
    }

    apmCalc( RminLam.lb, RminLam.lb, lam_scratch, APM_SUB, NULL ) ;
    apmCalc( RminLam.ub, RminLam.ub, lam_scratch, APM_ADD, NULL ) ;

/*
	Exploit the symmetry of the example.  The 
	largest value for an eigenvalue is 
	4.0 - (leastLam.lb).

	The calculation above assumes that the
	u part of the prism's center contains a
	starting point suitable for a least-eigenvalue
	kind of test, i.e. the point where the least ev
	attains its minimum.  The bdd_apm RmaxLam will
	contain information about the largest ev of beta 
	at the spot where leastLam is small.  To get the
	thing we really want for the calculations
	below we must exploit the symmetry described
	above.
*/
    apmSubtract( RmaxLam.ub, four, RminLam.lb ) ;
    apmCalc( Rdiscrim.ub, RmaxLam.ub, APM_DUP, APM_MUL,
			four, APM_SUB, NULL ) ;
    apmSqrt( Rsqrt_disc, precision, Rdiscrim.ub ) ;

/* 
	A global lower bound - if the least eigenvalue of
	one of the diagonal blocks (see notes, Jan 10 )
	slips below this value then the next block is 
	sure to have a negative eigenvalue.
*/
    apmSubtract( lam_scratch, RmaxLam.ub, Rsqrt_disc ) ;
    apmDivide( RminLeastLam, precision, (APM) NULL, lam_scratch, two ) ;
    apmCalc( RminLeastLam, RminLeastLam, max_error, APM_SUB, NULL ) ;
    minLeastLam = apmtodbl( RminLeastLam ) ;

	/* 
	    A lower bound on the sum of the non-maximal eigenvalues
	    of a diagonal block.
	*/
    sumTinyLams = minLeastLam ;
    apmAssign( RsumTinyLams, RminLeastLam ) ;

/*
	A global upper bound.
*/
    apmAdd( lam_scratch, RmaxLam.ub, Rsqrt_disc ) ;
    apmDivide( RmaxLeastLam, precision, (APM) NULL, lam_scratch, two ) ;
    apmCalc( RmaxLeastLam, RmaxLeastLam, max_error, APM_ADD, NULL ) ;
    maxLeastLam = apmtodbl( RmaxLeastLam ) ;

# if DEBUG
    printf( "globally :   %.6f < least ev. < %.6f \n",  minLeastLam, 
							maxLeastLam ) ;
    fflush( stdout ) ;
# endif
}
/* ++++++++++++++++++++++++++++++ */

double 
calc_bML (double *beta, double b_trace)
{
    double   discrim, det, sqrt() ;

    det =  beta[0] * beta[3] - beta[1] * beta[2] ;
    discrim = b_trace * b_trace - 4.0 * det ;

    if( discrim < 0.0 ) 
	discrim = 0.0 ;

    return( (b_trace + sqrt( discrim ) ) / 2.0 ) ;
}
/* +++++++++++++++++++++++++ */

void 
setSq (Bdd_dbl *xsq, Bdd_dbl *x)
{
    if( x->ub > 0.0 ) {
		if( x->lb > 0.0 ) {		/* Both positive */ 
			xsq->ub = x->ub * x->ub ;
			xsq->lb = x->lb * x->lb ;
		}
		else {				/* lb negative */
			if( x->ub > fabs(x->lb) ) {
				xsq->ub = x->ub * x->ub ;
				xsq->lb = 0.0 ;
			}
			else {
				xsq->ub = x->lb * x->lb ;
				xsq->lb = 0.0 ;
			}
		}
	}
	else {				/* Both negative */
		xsq->lb = x->ub * x->ub ;
		xsq->ub = x->lb * x->lb ;
	}
}
/* +++++++++++++++++++++++++++++++ */

void 
RsetSq (Bdd_apm *xsq, Bdd_apm *x)
{
    if( apmCompare( x->ub, zero ) > 0 ) {
	if( apmCompare( x->lb, zero ) > 0 ) {
	    apmMultiply( xsq->ub, x->ub, x->ub ) ;
	    apmMultiply( xsq->lb, x->lb, x->lb ) ;
	}
	else {
	    apmAbsoluteValue( lam_scratch, x->lb ) ;
	    if( apmCompare( x->ub, lam_scratch ) > 0 ) {
		apmMultiply( xsq->ub, x->ub, x->ub ) ;
		apmAssign( xsq->lb, zero ) ;
	    }
	    else {
		apmMultiply( xsq->ub, x->lb, x->lb ) ;
		apmAssign( xsq->lb, zero ) ;
	    }
	}
    }
    else {
	apmMultiply( xsq->ub, x->lb, x->lb ) ;
	apmMultiply( xsq->lb, x->ub, x->ub ) ;
    }
}
/* ++++++++++++++++++++++++++++++++ */

void 
setLeastLam (RPrism *priz)
/* 
	Calculate an upper bound on the largest eigenvalue of beta
	at the initial point, then use it and the global bound,
	maxLeastLam to set firstLeastLam.
*/
{
    int	    j, mat_pos ;

    for( j=0 ; j < N_PARMS ; j++ )  {
	earliest->center->p[j] = priz->center->p[j] ;

	mat_pos = j * (MAT_DIM + 1) ;
	earliest->matrix[mat_pos] = priz->matrix[mat_pos] ;
    }

    for( j=0 ; j < DEG_FREE ; j++ ) 
	earliest->center->z.v[j] = priz->center->z.u[j] ;

    Rglobal_bounds( earliest ) ;
    Rbound_btrace( &RBtrace, earliest ) ;
    Rbd_Blams( &RminLam, &RmaxLam, &RBtrace ) ;

/* 
	Obtain an upper bound on the least
	eigenvalue of the block of the Hessian of
	the action functional corresponding to the
	starting point. As in the functions in follow.c,
	compute a whole suite of estimates and choose
	the best one.

*/

	/*
		Rdenom is a global upper bound
		on the size of the largest eigevalue
		of a diagonal block.   
		    Rdenom = maximum trace - (n-1) * minimum ev.

		It's used together with the least eigenvalue
		of beta (evaluated at the starting point) :

		    LeastLam <= RminBlam.ub - 1.0 / Rdenom
	*/
    apmCalc( Rdenom, Rdf, one, APM_SUB, 
		     RminLeastLam, APM_MUL, APM_NEG, 
		     Rmax_slope, APM_ADD, NULL) ;
    apmDivide( lam_scratch, precision, (APM) NULL, one, Rdenom ) ;
    apmSubtract( RminBlam_ll, RminLam.ub, lam_scratch ) ;

	/*
		Here we try to attain a small estimate by 
		saying :
		    LeastLam <= RmaxBlam.ub - 1.0 / maxLeastLam.
	*/
    apmDivide( lam_scratch, precision, (APM) NULL, one, RmaxLeastLam ) ;
    apmSubtract( RmaxBlam_ll, RmaxLam.ub, lam_scratch ) ;

	/*
		Finally we make the estimate
		    LeastLam <= first_slope / DEG_FREE
	*/
    apmDivide( Rtrace_ll, precision, (APM)NULL, Rfirst_slope, Rdf ) ;

	/* 
		Choose the best (smallest) lower bound.
	*/
    apmAssign( RfirstLeastLam, RmaxBlam_ll ) ;
    if( apmCompare( RfirstLeastLam, RminBlam_ll ) == 1 ) 
	apmAssign( RfirstLeastLam, RminBlam_ll ) ;
    if( apmCompare( RfirstLeastLam, Rtrace_ll ) == 1 ) 
	apmAssign( RfirstLeastLam, Rtrace_ll ) ;

    firstLeastLam = apmtodbl( RfirstLeastLam ) ;

# if DEBUG
    printf( "firstLeastLam = %.6f \n", firstLeastLam ) ;
    printf( "the suite : maxBlam = %.6f, min_Blam = %.6f, trace = %.6f \n",
		apmtodbl( RmaxBlam_ll ) , apmtodbl( RminBlam_ll ),
					  apmtodbl( Rtrace_ll ) ) ;
    fflush( stdout ) ;
# endif 
}
