# include <stdio.h>
# include <string.h>
# include <stdlib.h>

# include "apm.h"
# include "converse.h"

/* Prototypes for functions private to this file */
static void set_defaults(void);
static void set_depth(char *argstr);
static void set_furthest(char *argstr);
static void set_precision(char *argstr);
static void set_graph(char *argstr);
static void set_terseness(char *argstr);
static void set_stubbornness(char *argstr);
static void set_restore(char *argstr);
static void set_backup(char *argstr);


/* A structure for managing command line options */
typedef struct {
	void   (*do_it)() ;
	char  name ;
} Option ;

static char    *optstr = "tsg:d:p:f:b:r:" ;
static Option  optlist[] = { 
		{ set_graph, 'g' }, 
		{ set_depth, 'd'}, 
		{ set_precision, 'p'},
		{ set_furthest, 'f' },
		{ set_terseness, 't'},
		{ set_stubbornness, 's'},
		{ set_backup, 'b'},
		{ set_restore, 'r'} 
	} ;

static int	num_opts = (sizeof(optlist) / sizeof(Option) ) ;
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
handle_opts (int argc, char *argv[])
{
    int    opt_num ;
    char   opt_char, getopt() ;

    set_defaults() ;

/*	
	For each command line option, run through the list of 
	recognized options and invoke the do_it() function of
	the one whose name matches the name form the command line.
									*/
    while( (opt_char = getopt( argc, argv, optstr )) != EOF ) { 
		for( opt_num=0 ; opt_num < num_opts ; opt_num++ ) {
			if( opt_char == optlist[opt_num].name )
			(*optlist[opt_num].do_it)( optarg ) ;
		}
    }

    if( restoration == YES ) 
		restore_options() ;
}
/* +++++++++++++++++++++++++++++++++ */

void 
set_defaults (void)
{
    furthest  = DFLT_FRTHST ;
    do_graph  = DFLT_GRAPH ;
    stubborn  = STUBBORN ;
    depth     = DFLT_DEPTH ;
    terse     = DFLT_TERSE ;

    do_backup = NO ;
    restoration = NO ;

    max_error = apmInit( 1L, -(DFLT_PRECIS), BASE ) ;

    RBgBlock_err = apmInit( (long)(DEG_FREE * N_PARMS), 
			     -DFLT_PRECIS, BASE ) ;
    RSmBlock_err = apmInit( (long)(DF_SQ), -DFLT_PRECIS, BASE ) ;
    precision = DFLT_PRECIS + SAFETY_PRECIS ;

    if( DFLT_ERR == YES )
		apmErrorFunc( errHndlr ) ;
}
/* +++++++++++++++++++++++++++++++ */

void 
set_depth (char *argstr)
{
/*
	Set the maximum number of subdividing steps the program
	will allow.  This number is called "depth" because it's
	the depth to which a tree structure of subdivided prisms
	would extend.
							*/
    depth = atoi( argstr ) ;
    if( depth < 0 ) {
		fprintf( stderr, "Won't set depth to a negative value. \n" ) ;
		fprintf( stderr, "Will use %d instead. \n", DFLT_DEPTH ) ;

		depth = DFLT_DEPTH ;
    }
}
/* ++++++++++++++++++++++++++++++++++++++++ */

void 
set_furthest (char *argstr)
{
    furthest = atoi( argstr ) ;
    if( furthest < 0 ) {
		fprintf( stderr, "Won't set furthest to a negative value. \n" ) ;
		fprintf( stderr, "Will use %d instead. \n", DFLT_FRTHST ) ;

		furthest = DFLT_FRTHST ;
    }
}
/* ++++++++++++++++++++++++++++++++++++++++ */

void 
set_precision (char *argstr)
{
/*
	Set the precision, ( the number of decimal places ), to
	which the APM calculations will be done.
	The number stored is actually two greater than the 
	desired precision.
								*/
    precision = atoi( argstr ) ;
    if( precision < 0 ) {
		fprintf( stderr, "Won't set precision to a negative value. \n" ) ;
		fprintf( stderr, "Will use %d instead. \n", DFLT_PRECIS ) ;

		precision = DFLT_PRECIS ; 
    }

    apmAssignLong( max_error, 1L, -precision, BASE ) ;
    apmAssignLong( RSmBlock_err, (long)(DF_SQ), -precision, BASE ) ;
    apmAssignLong( RBgBlock_err, (long)(DEG_FREE * N_PARMS), 
					 -precision, BASE ) ;
    precision += SAFETY_PRECIS ;
}
/* ++++++++++++++++++++++++++++++++++++++++ */

void 
set_graph (char *argstr)
{
/*
	If the default is to make a graphic display, choose
	no graphics.  If the default is to not have graphics, 
	turn them on.					*/
    if( DFLT_GRAPH == NO ) {
    	do_graph = YES ;
		graf_file = argstr ;
	}
	
	/* Deal with the special case of switching graphics off */
	if( strcmp( argstr, "off" ) == 0 ) {
		do_graph = NO ;
	}
}
/* +++++++++++++++++++++++ */

void 
set_terseness (char *argstr)
{
    if( terse == YES )
		terse = NO ;
    else
		terse = YES ;
}
/* +++++++++++++++++++++++ */

void 
set_stubbornness (char *argstr)
{
    if( stubborn == YES )
		stubborn = NO ;
    else
		stubborn = YES ;
}
/* ++++++++++++++++++++++++++ */

void 
set_restore (char *argstr)
{
    restoration = YES ;

    rest_name = argstr ;
    begin_restoration() ;
}
/* ++++++++++++++++++++++++++ */

void 
set_backup (char *argstr)
{
    do_backup = YES ;
    back_name = argstr ;
}
