# include <stdio.h>
# include <math.h>

# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "map.h"

# define DEBUG  NO	

/* Variables shared through converse.h */
void 	(* fatten)(), (* row_sums)() ;

/* Variables private to this file */
static APM		Rw[MAT_DIM] ;
static double	w[MAT_DIM]  ;
/* ++++++++++++++++++++++++++++++++ */

void 
prismatic_image (Prism *pz_prime, Prism *pz)
{
    int	      j ;
    double    *mpt, *end_mat, *wpt, *end_w ;

/*
	Find the image of the center of the prism.
*/
    image( pz_prime->center, pz->center ) ;

# if DEBUG
    printXpt( pz->center ) ;
    printXpt( pz_prime->center ) ;
# endif

	/* Calculate the derivative of the map. */
    jacobian( pz->center ) ;

/*
	Fatten the matrix Deriv * pz->matrix  so that it isn't too
	singular.
*/
# if DEBUG
    printMat( pz->matrix ) ;
    printMat( Deriv ) ;
# endif 
    (* fatten) ( pz_prime->matrix, Deriv, pz->matrix ) ;
# if DEBUG
    printMat( pz_prime->matrix ) ;
    fflush( stdout ) ;
# endif

/*
	Get upper bounds on the rows of the fattened matrix,
	and use them to get the matrix of a prism gauranteed 
	to enclose the image of pz.
*/
    (* row_sums)( w, pz_prime->matrix, Deriv, pz ) ;

    end_w = w + MAT_DIM ;
    end_mat = pz_prime->matrix + MAT_SZ ;
    for( mpt = pz_prime->matrix ; mpt < end_mat ; ) {
	for( wpt = w ; wpt < end_w ; wpt++ )
	    *mpt++ *= *wpt ;
    }
# if DEBUG
    printf( "\nw[]" ) ;
    for( j=0 ; j < MAT_DIM ; j++ )
	printf( "%.5f  ", w[j] ) ;
    printf( "\n" ) ;

    printMat( pz_prime->matrix ) ;
    printf( " ++++++++++++++++++++++++++++++ \n" );
    fflush( stdout ) ;
# endif
}
/* ++++++++++++++++++++++++ */

void 
Rprismatic_image (RPrism *pz_prime, RPrism *pz)
{
    int	   j ;
    APM    *mpt, *end_mat, *wpt, *end_w ;

/*
	Find the image of the center of the prism.
*/
    Rimage( pz_prime->center, pz->center ) ;

# if DEBUG
    printRXpt( pz->center ) ;
    printRXpt( pz_prime->center ) ;
# endif
	/* Calculate the derivative of the map. */
    Rjacobian( pz->center ) ;
    
# if DEBUG
    printf( "Rtry : \n" ) ;
    printRmat( pz->matrix ) ;
    printRmat( RDeriv ) ;
# endif 

/*
	Fatten the matrix Deriv * pz->matrix  so that it isn't too
	singular.
*/
    (* fatten) ( pz_prime->matrix, RDeriv, pz->matrix ) ;
# if DEBUG
    printRmat( pz_prime->matrix ) ;
    fflush( stdout ) ;
# endif

/*
	Get upper bounds on the rows of the fattened matrix,
	and use them to get the matrix of a prism gauranteed 
	to enclose the image of pz.
*/
    (* row_sums)( Rw, pz_prime->matrix, RDeriv, pz ) ;

    end_w = Rw + MAT_DIM ;
    end_mat = pz_prime->matrix + MAT_SZ ;
    for( mpt = pz_prime->matrix ; mpt < end_mat ; ) {
	for( wpt = Rw ; wpt < end_w ; wpt++, mpt++ )
	    apmCalc( *mpt, *mpt, *wpt, max_error, 
			   APM_ADD, APM_MUL, NULL ) ;
    }

    truncateRPrism( pz_prime, precision ) ;

# if DEBUG
    printf( "\nRw[]" ) ;
    for( j=0 ; j < MAT_DIM ; j++ )
	printf( "%.5f  ", apmtodbl( Rw[j] ) ) ;
    printf( "\n" ) ;

    printRmat( pz_prime->matrix ) ;
    printf( " ++++++++++++++++++++++++++++++ \n" );
    fflush( stdout ) ;

    if( row_sums == Rcr_rows )
	cease() ;
# endif
}
/* +++++++++++++++++++++ */

void 
initPrismatic (void)
{
    int   j ;

    for( j=0 ; j < N_PARMS ; j++ ) {
	Rw[j] = apmNew( BASE ) ;
	apmAssign( Rw[j], one ) ;
	w[j]  = 1.0 ;
    }

    for( j=N_PARMS ; j < (N_PARMS + DEG_FREE) ; j++ ) 
	Rw[j] = apmNew( BASE ) ;

    for( j=(N_PARMS + DEG_FREE) ; j < MAT_DIM ; j++ ) {
	w[j] = 1.0 + DBL_ERR ;
	Rw[j] = apmNew( BASE ) ;
	apmAdd( Rw[j], one, max_error ) ;
    }
}
