# include <stdio.h>
# include <string.h>
# include <math.h>

# include "apm.h"
# include "apmSpecial.h"/* two as an APM */
# include "apmUtil.h"	/* apmtodbl() */
# include "converse.h"

# define  DEBUG    NO 

/*
    Functions for refining prisms - for halving them along one of 
    the coordinate directions.  Since this is only done on 
    RPrisms the functions here only treat those.
*/

/* A variable private to this file */
static APM	delta ;
/* +++++++++++++++++++++++++ */

void 
initRefinement (void)
{
    delta = apmNew( BASE ) ;
}
/* ++++++++++++++++++++++++ */

int 
may_refine (RPrism *priz)
{
# if DEBUG
    printRPrism( priz, priz->n_cuts ) ;
# endif

    if( priz->n_cuts < depth ) {
        if( priz->n_cuts == most_cuts )
	    most_cuts = priz->n_cuts + 1 ;

		return( YES ) ;
    }
    else
		return( NO ) ;
}
/* +++++++++++++++++++++++++ */

void 
refinePrism (RPrism *priz, Prism *image)
{
    int	    col, max_col ;
    APM     *pmat_pt, *npmat_pt, *pcntr_pt, *npcntr_pt ;
    double  col_sum, *cpt, *end_col, max_sum ;
    RPrism  *new_priz ;

    new_priz = conjureRPrism() ;
    if( new_priz == NULL ) {
		fprintf( stderr, "Can't refine a prism. Died. \n" ) ;
		cease() ;
    }
    else {
	/* 	Copy the old prism into the new and 
		link the new one into the list.
		Note that the refinement will only
		affect two entries in each of priz 
		and new_priz.				*/

		RcopyRPrism( new_priz, priz ) ;

		new_priz->next = priz->next ;
		priz->next = new_priz ;

		/*	Decide how to make the cut.		*/

		max_sum = 0.0 ;
		max_col = 0 ;
		for( col=0 ; col < MAT_DIM ; col++ ) { 
			cpt = image->matrix + col ;
			col_sum = 0.0 ;
			for( end_col=cpt + MAT_SZ ; cpt < end_col ; cpt += MAT_DIM )
			col_sum += fabs( *cpt ) ;

			col_sum *= apmtodbl( priz->matrix[col * (MAT_DIM + 1)] ) ;
	/* 
		   Find the column for which the maximimum column
		   sum is attained.  We'll make our refining cut 
		   in the corresponding parameter.  Note that we
		   ignore the columns which correspond to the u
		   coordinate; the initial prism already has
		   zero width in these variables, we can't cut
		   it any more finely.
	*/
			if( (col_sum > max_sum) &&  
			((col < N_PARMS) || (col >= (N_PARMS + DEG_FREE))) ) {
				max_sum = col_sum ;
				max_col = col ;
			}
		}

# if DEBUG
	printf( "refine : about to cut along the %d axis.\n", max_col ) ;
# endif
   
		/*
		   Cut the original prism in half, putting
		   the bit which corresponds to larger 
		   parameter value in priz and the bit which
		   corresponds to smaller value in new_priz.
		   Record the change in the cut history of each 
		   prism.					*/

		pmat_pt  = priz->matrix + (max_col * (MAT_DIM + 1)) ;
		npmat_pt = new_priz->matrix + (max_col * (MAT_DIM + 1)) ;
		if( max_col < N_PARMS ) {
			pcntr_pt  = priz->center->p + max_col ;
			npcntr_pt = new_priz->center->p + max_col ;
		}
		else {
			pcntr_pt  = priz->center->z.v + (max_col - (DEG_FREE + N_PARMS)) ;
			npcntr_pt = new_priz->center->z.v + (max_col - (DEG_FREE+N_PARMS)) ;
		}

		apmDivide( delta, precision, (APM)NULL, *pmat_pt, two ) ;
		apmAssign( *pmat_pt, delta ) ;
		apmAssign( *npmat_pt, delta ) ;

		apmCalc( *pcntr_pt, *pcntr_pt, delta, APM_ADD, NULL ) ;
		apmCalc( *npcntr_pt, *npcntr_pt, delta, APM_SUB, NULL ) ;

		strcat( priz->cuts[max_col], "r" ) ;
		strcat( new_priz->cuts[max_col], "l" ) ;

		priz->n_cuts++ ;
		new_priz->n_cuts++ ;
    }
}
