# include <stdio.h>
# include <math.h>
# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"

# define  DEBUG	      NO 

# define  NUM_FACTS   14 
# define  NUM_TERMS   11 
# define  DET_TOL     1e-13


/* Variables shared through rows.h */
int	  isNewPrism ;

Bdd_expr  a_sin, b_sin, c_sin ;
Bapm_expr Ra_sin, Rb_sin, Rc_sin ;

APM	  neg_one, neg_two, Rrow_abs[DEG_FREE] ;

/* Variables private to this file */
static APM		Rrows[DEG_FREE] ;

static Bdd_dbl	a, b, c, theta ;
static Bdd_dbl	cos_zero, cos_one, cos_sum ;
static Bdd_dbl	sin_zero, sin_one, sin_sum ;
static Bdd_dbl	*row_factors[NUM_FACTS] ;

static Bdd_term		row_terms[NUM_TERMS] ;
static Bdd_expr		beta_dif[3], gamma_dif[3] ;
 
static Bdd_apm		Ra, Rb, Rc, Rtheta ;
static Bdd_apm		Rcos_zero, Rcos_one, Rcos_sum ;
static Bdd_apm		Rsin_zero, Rsin_one, Rsin_sum ;
static Bdd_apm		*Rrow_factors[NUM_FACTS] ;
static Bapm_term	Rrow_terms[NUM_TERMS] ;
static Bapm_expr	Rbeta_dif[3], Rgamma_dif[3] ;
/* ++++++++++++++++++++++++++++++ */

void 
initRowSums (void)
/*
	Set up the expressions and terms as described in my notes
	from 11/14.
*/
{
    int	      j, k ;
    Bdd_dbl   **dpt ;
    Bdd_apm   **apt ;
    Bdd_term  *tpt ;
    Bapm_term *Rtpt ;

/*
    Set up some APM's to be used to hold intermediate 
    results.
*/
    newBapm( Ra, BASE ) ;
    newBapm( Rb, BASE ) ;
    newBapm( Rc, BASE ) ;
    newBapm( Rtheta, BASE ) ;

    newBapm( Rcos_zero, BASE ) ;
    newBapm( Rcos_one, BASE ) ;
    newBapm( Rcos_sum, BASE ) ;
    newBapm( Rsin_zero, BASE ) ;
    newBapm( Rsin_one, BASE ) ;
    newBapm( Rsin_sum, BASE ) ;

    neg_one = apmInit( -1L, 0, BASE ) ;
    neg_two = apmInit( -2L, 0, BASE ) ;

    for( j=0 ; j <DEG_FREE ; j++ )  {
	Rrows[j] = apmNew( BASE ) ;
	Rrow_abs[j] = apmNew( BASE ) ;
    }

/*
	Set the number of terms in the bounded expressions
*/
    a_sin.nterms = Ra_sin.nterms = 1 ;
    b_sin.nterms = Rb_sin.nterms = 1 ;
    c_sin.nterms = Rc_sin.nterms = 1 ;

    beta_dif[0].nterms = Rbeta_dif[0].nterms = 2 ;
    beta_dif[1].nterms = Rbeta_dif[1].nterms = 1 ;
    beta_dif[2].nterms = Rbeta_dif[2].nterms = 2 ;

    gamma_dif[0].nterms = Rgamma_dif[0].nterms = 1 ;
    gamma_dif[1].nterms = Rgamma_dif[1].nterms = 1 ;
    gamma_dif[2].nterms = Rgamma_dif[2].nterms = 1 ;

/*	
	Assign terms
*/

    tpt = row_terms ;
    Rtpt = Rrow_terms ;
    for( j=0 ; j < 3 ; j++ ) {
	beta_dif[j].terms = tpt ;
	Rbeta_dif[j].terms = Rtpt ;
	tpt += beta_dif[j].nterms ;
	Rtpt += Rbeta_dif[j].nterms ;

	gamma_dif[j].terms = tpt ;
	Rgamma_dif[j].terms = Rtpt ;
	tpt += gamma_dif[j].nterms ;
	Rtpt += Rgamma_dif[j].nterms ;
    }

    a_sin.terms = tpt++ ;
    Ra_sin.terms = Rtpt++ ;

    b_sin.terms = tpt++ ;
    Rb_sin.terms = Rtpt++ ;

    c_sin.terms = tpt++ ;
    Rc_sin.terms = Rtpt++ ;

/*
	Set nfactors.
*/
    
    Rbeta_dif[0].terms[0].nfactors = beta_dif[0].terms[0].nfactors = 1 ;
    Rbeta_dif[0].terms[1].nfactors = beta_dif[0].terms[1].nfactors = 1 ;
    Rbeta_dif[1].terms[0].nfactors = beta_dif[1].terms[0].nfactors = 1 ;
    Rbeta_dif[2].terms[0].nfactors = beta_dif[2].terms[0].nfactors = 1 ;
    Rbeta_dif[2].terms[1].nfactors = beta_dif[2].terms[1].nfactors = 1 ;

    Rgamma_dif[0].terms->nfactors = gamma_dif[0].terms->nfactors = 1 ;
    Rgamma_dif[1].terms->nfactors = gamma_dif[1].terms->nfactors = 1 ;
    Rgamma_dif[2].terms->nfactors = gamma_dif[2].terms->nfactors = 1 ;

    a_sin.terms->nfactors = Ra_sin.terms->nfactors = 2 ;
    b_sin.terms->nfactors = Rb_sin.terms->nfactors = 2 ;
    c_sin.terms->nfactors = Rc_sin.terms->nfactors = 2 ;
/*
	Assign factors. 
*/

    dpt = row_factors ;
    apt = Rrow_factors ;
    for( j=0 ; j < 3 ; j++ ) {
    /*
	beta_dif
    */
	for( k=0 ; k < beta_dif[j].nterms ; k++ ) {
	    beta_dif[j].terms[k].factors = dpt ;
	    Rbeta_dif[j].terms[k].factors = apt ;

	    dpt += beta_dif[j].terms[k].nfactors ;
	    apt += Rbeta_dif[j].terms[k].nfactors ;
	}

    /*
	gamma_dif
    */
	for( k=0 ; k < gamma_dif[j].nterms ; k++ ) {
	    gamma_dif[j].terms[k].factors = dpt ;
	    Rgamma_dif[j].terms[k].factors = apt ;

	    dpt += gamma_dif[j].terms[k].nfactors ;
	    apt += Rgamma_dif[j].terms[k].nfactors ;
	}
    }

    a_sin.terms->factors = dpt ;
    Ra_sin.terms->factors = apt ;
    dpt += 2 ;
    apt += 2 ;

    b_sin.terms->factors = dpt ;
    Rb_sin.terms->factors = apt ;
    dpt += 2 ;
    apt += 2 ;

    c_sin.terms->factors = dpt ;
    Rc_sin.terms->factors = apt ;

/*
	Set up those of the "bound" attributes which are 
	bounded APM's.
*/

    for( j=0 ; j < NUM_TERMS ; j++ ) {
	newBapm( Rrow_terms[j].bound, BASE ) ;
    }

    for( j=0 ; j < 3 ; j++ ) {
	newBapm( Rbeta_dif[j].bound, BASE ) ;
	newBapm( Rgamma_dif[j].bound, BASE ) ;
    }

    newBapm( Ra_sin.bound, BASE ) ;
    newBapm( Rb_sin.bound, BASE ) ;
    newBapm( Rc_sin.bound, BASE ) ;

/*
	Set up the terms and expressions. 
*/

/*  a_sin */

    a_sin.const_term = 0.0 ;
    Ra_sin.const_term = apmNew( BASE ) ;
	a_sin.terms->coef = 1.0 ;
	Ra_sin.terms->coef = apmInit( 1L, 0, BASE ) ;
	    
	    a_sin.terms->factors[0] = &a ;
	    a_sin.terms->factors[1] = &sin_zero ;
	    Ra_sin.terms->factors[0] = &Ra ;
	    Ra_sin.terms->factors[1] = &Rsin_zero ;

/*  b_sin */

    b_sin.const_term = 0.0 ;
    Rb_sin.const_term = apmNew( BASE ) ;
	b_sin.terms->coef = 1.0 ;
	Rb_sin.terms->coef = apmInit( 1L, 0, BASE ) ;
	    
	    b_sin.terms->factors[0] = &b ;
	    b_sin.terms->factors[1] = &sin_one ;
	    Rb_sin.terms->factors[0] = &Rb ;
	    Rb_sin.terms->factors[1] = &Rsin_one ;

/*  c_sin */

    c_sin.const_term = 0.0 ;
    Rc_sin.const_term = apmNew( BASE ) ;
	c_sin.terms->coef = 1.0 ;
	Rc_sin.terms->coef = apmInit( 1L, 0, BASE ) ;
	    
	    c_sin.terms->factors[0] = &c ;
	    c_sin.terms->factors[1] = &sin_sum ;
	    Rc_sin.terms->factors[0] = &Rc ;
	    Rc_sin.terms->factors[1] = &Rsin_sum ;

/*  beta_dif */
 
	/* beta_dif[0] =  (2.0 - a * sin(v[0]) - c * sin(v[0] + v[1]) ) 
		   -{ 2.0 - ac * sin(vc[0]) - cc * sin(vc[0] + vc[1])  
	    Where ac, cc, vc[0], and vc[1] are the values of these
	    numbers at the center of the prism.  The whole second 
	    term ( enclosed in braces ) is an entry in the jacobian 
	    of the map							
								*/
    Rbeta_dif[0].const_term = apmNew( BASE ) ;
	beta_dif[0].terms[0].coef = -1.0 ;
	Rbeta_dif[0].terms[0].coef = neg_one ;

	    beta_dif[0].terms[0].factors[0] = &a_sin.bound ;
	    Rbeta_dif[0].terms[0].factors[0] = &Ra_sin.bound ;

	beta_dif[0].terms[1].coef = -1.0 ;
	Rbeta_dif[0].terms[1].coef = neg_one ;

	    beta_dif[0].terms[1].factors[0] = &c_sin.bound ;
	    Rbeta_dif[0].terms[1].factors[0] = &Rc_sin.bound ;

	/* beta_dif[1] =     -2.0 * c * sin.bound( v[0] + v[1] )
			 - { -2.0 * cc * sin.bound( vc[0] + vc[1] ) } 
	 */
    Rbeta_dif[1].const_term = apmNew( BASE ) ;
	beta_dif[1].terms[0].coef = -2.0 ;
	Rbeta_dif[1].terms[0].coef = neg_two ;

	    beta_dif[1].terms[0].factors[0] = &c_sin.bound ;
	    Rbeta_dif[1].terms[0].factors[0] = &Rc_sin.bound ;

	/* beta_dif[2] = 2.0 - b * sin.bound(v[1]) - c * sin(v[1] + v[0])  
		-{ 2.0 - bc * sin.bound(vc[1]) - cc * sin(vc[1] + vc[0]) }
	*/
    Rbeta_dif[2].const_term = apmNew( BASE ) ;
	beta_dif[2].terms[0].coef = -1.0 ;
	Rbeta_dif[2].terms[0].coef = neg_one ;

	    beta_dif[2].terms[0].factors[0] = &b_sin.bound ;
	    Rbeta_dif[2].terms[0].factors[0] = &Rb_sin.bound ;

	beta_dif[2].terms[1].coef = -1.0 ;
	Rbeta_dif[2].terms[1].coef = neg_one ;

	    beta_dif[2].terms[1].factors[0] = &c_sin.bound ;
	    Rbeta_dif[2].terms[1].factors[0] = &Rc_sin.bound ;

/* gamma_dif */

	/* gamma_dif[0] = da * ( cos(v[0]) - cos(vc[0]) ) 
	   Where da is half the prism's width as measured
	   along the a-axis and vc is as above.		  */

    Rgamma_dif[0].const_term = apmNew( BASE ) ;

	Rgamma_dif[0].terms[0].coef = apmNew( BASE ) ;
	    
	    gamma_dif[0].terms[0].factors[0] = &cos_zero ;
	    Rgamma_dif[0].terms[0].factors[0] = &Rcos_zero ;

	/* gamma_dif[1] = db * ( cos(v[1]) - cos(vc[1]) )  */

    Rgamma_dif[1].const_term = apmNew( BASE ) ;

	Rgamma_dif[1].terms[0].coef = apmNew( BASE ) ;
	    
	    gamma_dif[1].terms[0].factors[0] = &cos_one ;
	    Rgamma_dif[1].terms[0].factors[0] = &Rcos_one ;

	/* gamma_dif[2] = dc * ( cos(v[0]  + v[1]) - 
				 cos(vc[0] + vc[1]) )  */

    Rgamma_dif[2].const_term = apmNew( BASE ) ;

	Rgamma_dif[2].terms[0].coef = apmNew( BASE ) ;
	    
	    gamma_dif[2].terms[0].factors[0] = &cos_sum ;
	    Rgamma_dif[2].terms[0].factors[0] = &Rcos_sum ;
}
/* +++++++++++++++++++++++++++++++++ */

void 
global_bounds (Prism *pz)
{
    int	     j ;
    double   rows[DEG_FREE], *mpt, *end_row ;

    a.ub = pz->center->p[0] + pz->matrix[0] ;
    a.lb = pz->center->p[0] - pz->matrix[0] ;

    b.ub = pz->center->p[1] + pz->matrix[MAT_DIM+1] ;
    b.lb = pz->center->p[1] - pz->matrix[MAT_DIM+1] ;

    c.ub = pz->center->p[2] + pz->matrix[2*MAT_DIM+2] ;
    c.lb = pz->center->p[2] - pz->matrix[2*MAT_DIM+2] ;

    mpt = pz->matrix + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	rows[j] = 0.0 ;
	for( end_row=mpt + MAT_DIM ; mpt < end_row ; mpt++ ) {
	    rows[j] += fabs( *mpt ) ;
	}
    }

    theta.ub = pz->center->z.v[0] + rows[0] ;
    theta.lb = pz->center->z.v[0] - rows[0] ;
    bd_sin( &sin_zero, &theta ) ;
    bd_cos( &cos_zero, &theta ) ;
	
    theta.ub = pz->center->z.v[1] + rows[1] ;
    theta.lb = pz->center->z.v[1] - rows[1] ;
    bd_sin( &sin_one, &theta ) ;
    bd_cos( &cos_one, &theta ) ;
	
    theta.ub += pz->center->z.v[0] + rows[0] ;
    theta.lb += pz->center->z.v[0] - rows[0] ;
    bd_sin( &sin_sum, &theta ) ;
    bd_cos( &cos_sum, &theta ) ;

    bound_expr( &a_sin ) ;
    bound_expr( &b_sin ) ;
    bound_expr( &c_sin ) ;

# if DEBUG
    printf( "a_sin.ub = %.4f, b_sin.ub = %.4f, c_sin.ub = %.4f \n", 
							a_sin.bound.ub, 
							b_sin.bound.ub,
							c_sin.bound.ub ) ;
    printf( "a_sin.lb = %.4f, b_sin.lb = %.4f, c_sin.lb = %.4f \n", 
							a_sin.bound.lb, 
							b_sin.bound.lb,
							c_sin.bound.lb ) ;
# endif
}
/* +++++++++++++++++++++++++++++++++ */

double 
beta_dif_star (double *deriv)
{
     double   answer, *dpt ;

     dpt = deriv + STAID_LEN + (MAT_DIM*DEG_FREE) + N_PARMS + DEG_FREE ;
     beta_dif[0].const_term = 2.0 - *dpt++ ; 
     beta_dif[1].const_term = -2.0 * (*dpt) ;
     dpt += MAT_DIM ;
     beta_dif[2].const_term = 2.0 - *dpt ;

    bound_expr( &beta_dif[0] ) ;
    bound_expr( &beta_dif[1] ) ;
    bound_expr( &beta_dif[2] ) ;

    answer = maxAbs( beta_dif[0].bound.ub, beta_dif[0].bound.lb ) +
	     maxAbs( beta_dif[1].bound.ub, beta_dif[1].bound.lb ) +
	     maxAbs( beta_dif[2].bound.ub, beta_dif[2].bound.lb )  ;

# if DEBUG
    printf( "beta_dif_star : %.4f + %.4f + %.4f \n",
    	maxAbs( beta_dif[0].bound.ub, beta_dif[0].bound.lb ),
	maxAbs( beta_dif[1].bound.ub, beta_dif[1].bound.lb ),
	maxAbs( beta_dif[2].bound.ub, beta_dif[2].bound.lb ) )  ;
# endif

    return( answer ) ;
}
/* +++++++++++++++++++++ */

double 
gamdif_star (double *deriv, double *pmat)
{
    double  *dpt, da, db, dc, answer ;

    da = pmat[0] ;
    db = pmat[MAT_DIM + 1] ;
    dc = pmat[2 * MAT_DIM + 2] ;

    dpt = deriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    gamma_dif[0].const_term = -da * (*dpt) ;
    dpt += MAT_DIM + 1 ; 
    gamma_dif[1].const_term = -db * (*dpt) ;
    dpt++ ;
    gamma_dif[2].const_term = -2.0 * dc * (*dpt) ;

    gamma_dif[0].terms[0].coef = da ;
    gamma_dif[1].terms[0].coef = db ;
    gamma_dif[2].terms[0].coef = 2.0 * dc ;

    bound_expr( &gamma_dif[0] ) ;
    bound_expr( &gamma_dif[1] ) ;
    bound_expr( &gamma_dif[2] ) ;

    answer = maxAbs( gamma_dif[0].bound.ub, gamma_dif[0].bound.lb ) +
	     maxAbs( gamma_dif[1].bound.ub, gamma_dif[1].bound.lb ) +
	     maxAbs( gamma_dif[2].bound.ub, gamma_dif[2].bound.lb )  ;

    return( answer ) ;
}
/* +++++++++++++++++++++++++++++++++ */

void 
Rglobal_bounds (RPrism *pz)
{
    int	     j ;
    APM	     *apt, *end_row ;

    apmAdd( Ra.ub, pz->center->p[0], pz->matrix[0] ) ;
    apmSubtract( Ra.lb, pz->center->p[0], pz->matrix[0] ) ;

    apmAdd( Rb.ub, pz->center->p[1], pz->matrix[MAT_DIM+1] ) ;
    apmSubtract( Rb.lb, pz->center->p[1], pz->matrix[MAT_DIM+1] ) ;

    apmAdd( Rc.ub, pz->center->p[2], pz->matrix[2*MAT_DIM+2] ) ;
    apmSubtract( Rc.lb, pz->center->p[2], pz->matrix[2*MAT_DIM+2] ) ;

    apt = pz->matrix + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	apmAssign( Rrows[j], zero ) ;
	for( end_row=apt + MAT_DIM ; apt < end_row ; apt++ ) {
	    apmCalc( Rrows[j], Rrows[j], *apt, 
					APM_ABS, APM_ADD, NULL ) ;
	}
    }

    apmAdd( Rtheta.ub, pz->center->z.v[0], Rrows[0] ) ;
    apmSubtract( Rtheta.lb, pz->center->z.v[0], Rrows[0] ) ;
    Rbd_sin( &Rsin_zero, &Rtheta ) ;
    Rbd_cos( &Rcos_zero, &Rtheta ) ;
	
    apmAdd( Rtheta.ub, pz->center->z.v[1], Rrows[1] ) ;
    apmSubtract( Rtheta.lb, pz->center->z.v[1], Rrows[1] ) ;
    Rbd_sin( &Rsin_one, &Rtheta ) ;
    Rbd_cos( &Rcos_one, &Rtheta ) ;
	
    apmCalc( Rtheta.ub, Rtheta.ub, pz->center->z.v[0], Rrows[0],
					APM_ADD, APM_ADD, NULL ) ;
    apmCalc( Rtheta.lb, Rtheta.lb, pz->center->z.v[0], Rrows[0],
					APM_SUB, APM_ADD, NULL ) ;
    Rbd_sin( &Rsin_sum, &Rtheta ) ;
    Rbd_cos( &Rcos_sum, &Rtheta ) ;

    Rbound_expr( &Ra_sin ) ;
    Rbound_expr( &Rb_sin ) ;
    Rbound_expr( &Rc_sin ) ;

# if DEBUG
    printf( "Ra_sin.ub = %.4f, Rb_sin.ub = %.4f, Rc_sin.ub = %.4f \n", 
					apmtodbl(Ra_sin.bound.ub), 
					apmtodbl(Rb_sin.bound.ub),
					apmtodbl(Rc_sin.bound.ub) ) ;
    printf( "Ra_sin.lb = %.4f, Rb_sin.lb = %.4f, Rc_sin.lb = %.4f \n", 
					apmtodbl(Ra_sin.bound.lb), 
					apmtodbl(Rb_sin.bound.lb),
					apmtodbl(Rc_sin.bound.lb) ) ;
# endif
}
/* +++++++++++++++++++++++++++++++++ */

void 
Rbeta_dif_star (APM answer, APM *deriv)
{
     APM    *dpt ;

     dpt = deriv + STAID_LEN + (MAT_DIM*DEG_FREE) + N_PARMS + DEG_FREE ;
     apmSubtract( Rbeta_dif[0].const_term, two, *dpt++ ) ; 
     apmMultiply( Rbeta_dif[1].const_term, neg_two, *dpt ) ;
     dpt += MAT_DIM ;
     apmSubtract( Rbeta_dif[2].const_term, two, *dpt ) ;

    Rbound_expr( &Rbeta_dif[0] ) ;
    Rbound_expr( &Rbeta_dif[1] ) ;
    Rbound_expr( &Rbeta_dif[2] ) ;

    RmaxAbs( answer, Rbeta_dif[0].bound.ub, Rbeta_dif[0].bound.lb )  ;
    RmaxAbs( Rrow_abs[0], Rbeta_dif[1].bound.ub, Rbeta_dif[1].bound.lb ) ;
    RmaxAbs( Rrow_abs[1], Rbeta_dif[2].bound.ub, Rbeta_dif[2].bound.lb ) ;

# if DEBUG
    printf( "beta_dif_star : %.4f + %.4f + %.4f \n",
					apmtodbl(answer),
					apmtodbl(Rrow_abs[0]),
					apmtodbl(Rrow_abs[1]) ) ;
# endif

/*
	Add max_error to the answer to account for the uncertainties
	in beta**(center).
*/
    apmCalc( answer, answer, Rrow_abs[0], Rrow_abs[1], max_error, 
				APM_ADD, APM_ADD, APM_ADD, NULL ) ; 
}
/* +++++++++++++++++++++ */

void 
Rgamdif_star (APM answer, APM *deriv, APM *pmat)
{
    APM    *apt, *Rda, *Rdb, *Rdc ;

    Rda = pmat ;
    Rdb = pmat + MAT_DIM + 1 ;
    Rdc = pmat+ (2 * MAT_DIM) + 2 ;

    apmAssign( Rgamma_dif[0].terms[0].coef, *Rda ) ;
    apmAssign( Rgamma_dif[1].terms[0].coef, *Rdb ) ;
    apmMultiply( Rgamma_dif[2].terms[0].coef, two, *Rdc ) ;

    apt = deriv + STAID_LEN + (DEG_FREE * MAT_DIM) ;
    apmCalc( Rgamma_dif[0].const_term, *Rda, APM_NEG, *apt, APM_MUL, NULL ) ;
    apt += MAT_DIM + 1 ; 
    apmCalc( Rgamma_dif[1].const_term, *Rdb, APM_NEG, *apt, APM_MUL, NULL ) ;
    apt++ ;
    apmCalc( Rgamma_dif[2].const_term, two, APM_NEG, *Rdc, *apt, 
				       APM_MUL, APM_MUL, NULL ) ;

    Rbound_expr( &Rgamma_dif[0] ) ;
    Rbound_expr( &Rgamma_dif[1] ) ;
    Rbound_expr( &Rgamma_dif[2] ) ;

    RmaxAbs( answer, Rgamma_dif[0].bound.ub, Rgamma_dif[0].bound.lb )  ;
    RmaxAbs( Rrow_abs[0], Rgamma_dif[1].bound.ub, Rgamma_dif[1].bound.lb ) ;
    RmaxAbs( Rrow_abs[1], Rgamma_dif[2].bound.ub, Rgamma_dif[2].bound.lb ) ;
/*
	Add max_error to the answer to account for the uncertainties
	in beta**(center).
*/
    apmCalc( answer, answer, Rrow_abs[0], Rrow_abs[1], max_error,
				APM_ADD, APM_ADD, APM_ADD, NULL ) ; 
}
