# include <stdio.h>
# include <stdlib.h>
# include <sys/time.h>
# include <sys/resource.h>

# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"
# include "map.h"

# define  DEBUG		NO

# define  USE_EDGE	YES

/* Variables shared via converse.h */
double	parm_roof[N_PARMS], parm_floor[N_PARMS] ;

/* Variables private to this file */
static int	  	found_doom ;
static struct	rusage  rsrcs ;
/* ++++++++++++++++++++++++++ */

void 
commence (RPrism *first)
{
/*
	Beware : these initialization functions sometimes depend on 
	each other; the order in which they are called may be important.
	For example, initApmSpecials() is important for most of the 
	others.  initRowSums() should come before initffRows() or 
	initcrRows().
*/
    initApmSpecials( BASE ) ;
    setTrigDp( precision ) ;

# if DEBUG 
	printf( "Many initializations : backup = %d, restoration = %d \n",
						do_backup, restoration ) ;
# endif
    initUtil() ;
    initMap() ;
    initLip() ;
    initTree() ;
    initGauss() ;
    initRefinement() ;
    initFollowing() ;
    initBounding() ;
    initTrigBd() ;
    initTrace() ;
    initLambda() ;
    initPrismatic() ;
    initRotor() ;
    initRowSums() ;
    init_ffRows() ;
    init_crRows() ;

    if( do_graph == YES )  {
		if( restoration == YES )
			restartGraphics() ;
		else
			startGraphics() ;
    }

    if( restoration == NO ) {
		read_parameters( first ) ;
		echo_parameters( first ) ;
		setEdge_o_Chaos( first ) ;
    } 
    else  {
		restore_start() ;		/*  The order in which  */
# if USE_EDGE					/*  the various 	*/
		restore_edge() ;		/*  restorations are	*/
# endif							/*  made matters.	*/
		restore_performance() ;
		restore_chain( first ) ;
    }

    if( do_backup == YES ) 
		startBackup( first ) ;

    first->in_torus = ACTIVE ;
    found_doom = NO ;
}
/* ++++++++++++++++++++++++ */

void 
cease (void)
{
    if( do_graph == YES )
		finishGraphics() ;

    report_performance() ;
    exit( 0 ) ;
}
/* ++++++++++++++++++++++ */

void 
moveEdge_o_Chaos (RPrism *priz, int nsteps)
{
    int      j ;
    double   center, width, new_floor ;

    for( j=0 ; j < N_PARMS ; j++ ) {
	center = apmtodbl( priz->center->p[j] ) ;
	width  = apmtodbl( priz->matrix[j*(MAT_DIM + 1)] ) ;

	new_floor = center + width ;
	if( new_floor == parm_roof[j] ) {
	    if( found_doom == NO ) {
	        printRPrism( priz, nsteps ) ;
	        printf( "Dooms hopes for a converse KAM " ) ;
	        printf( "theorem on these parameters. \n" ) ;

		found_doom = YES ;
	    }

	    if( stubborn == NO ) 
	        cease() ;
	}
    }
}
/* +++++++++++++++++++++++++ */

void 
setEdge_o_Chaos (RPrism *priz)
{
    int      j ;
    double   center, width ;

    for( j=0 ; j < N_PARMS ; j++ ) {
		center = apmtodbl( priz->center->p[j] ) ;
		width  = apmtodbl( priz->matrix[j*(MAT_DIM + 1)] ) ;

		parm_roof[j] = center + width ;
		parm_floor[j] = center - width ;
    }
}
/* ++++++++++++++++++++++++++++ */

void 
report_performance (void)
{
    int     j ;
    double  ll_total, elapsed ;

	/* Report various facts about the run. */
    if ( found_doom == NO ) {
        printf( "I find no invariant tori for the range of parameters :\n" ) ;
	for( j=0 ; j < N_PARMS ; j++ ) 
	    printf( "%.6f < %s < %.6f \n", parm_floor[j], parm_names[j], 
							  parm_roof[j] ) ;
    }
    printf( "\nDid %d quick checks, %d semi-rigorous bounding tries,\n", 
							    quick_tries,
							         tries ) ;
    printf( "and %d rigorous bounding tries.\n", Rtries ) ;
    printf( "The most deeply refined prism was cut %d times. \n", most_cuts ) ;
    printf( "The longest semi-rigorous orbit ran for %d iterations, \n", 
							    max_steps ) ;
    printf( "the longest successful orbit, %d iterations. \n", 
						 max_NTsteps ) ;
    printf( "Of the %d successful prisms, %d fell to the", 
					(HermSuccess + LLSuccess),
						     HermSuccess ) ;
    printf( " trace criterion, \n%d to the least eigenvalue test. \n",
							   LLSuccess ) ;

	/* Report on the puissance of the various leastlam estimates */
    ll_total = ll_used[0] + ll_used[1] + ll_used[2] ;
    if( ll_total != 0.0 ) {
	printf( "The best upper bound on the least eigenvalue came from \n" ) ;
	printf( "the maxBlam criterion %.1f%% of the time,\n",
					(100.0 * ll_used[0] / ll_total) ) ;
	printf( "the minBlam criterion %.1f%% of the time,\n",
					(100.0 * ll_used[1] / ll_total) ) ;
	printf( "and from the trace criterion %.1f%% of the time. \n\n",
					(100.0 * ll_used[2] / ll_total) ) ;
    }

	/*  Report the time spent examining prisms */
    getrusage( RUSAGE_SELF, &rsrcs ) ;
    elapsed = rsrcs.ru_stime.tv_sec + 0.000001 * rsrcs.ru_stime.tv_usec +
	      rsrcs.ru_utime.tv_sec + 0.000001 * rsrcs.ru_utime.tv_usec ;
    printf( "\nThis investigation took %.2f seconds.\n", elapsed ) ;
}
