# include <stdio.h>
# include <math.h>
# include "apm.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"

# define    DEBUG  NO	

# define    NUM_FACTS   3 
# define    NUM_TERMS   3

/* Variables private to this file */
static Bdd_dbl		*fact_buf[NUM_FACTS] ;
static Bdd_expr		b_trc ;
static Bdd_term		trace_terms[NUM_TERMS] ;

static Bdd_apm		*Rfact_buf[NUM_FACTS] ;
static Bapm_expr	Rb_trc ;
static Bapm_term	Rtrace_terms[NUM_TERMS] ;
/* ++++++++++++++++++++++++++++++ */

void 
initTrace (void)
{
    int		j ;		
    Bdd_dbl	**fpt ;
    Bdd_apm	**Rfpt ;
/*
	Set up the expressions.
*/
    b_trc.nterms  = NUM_TERMS ;
    Rb_trc.nterms = NUM_TERMS ;

    b_trc.const_term = 4.0 ;
    Rb_trc.const_term = apmInit( 4L, 0, BASE ) ;

    newBapm( Rb_trc.bound, BASE ) ;

    b_trc.terms  = trace_terms ;
    Rb_trc.terms = Rtrace_terms ;
 
 /*
	Set up their terms.
*/
    fpt = fact_buf ;
    Rfpt = Rfact_buf ;
    for( j=0 ; j < NUM_TERMS ; j++ ) {
	trace_terms[j].nfactors = 1 ;
	trace_terms[j].coef = -1.0 ;
	trace_terms[j].factors = fpt ;

	Rtrace_terms[j].nfactors = 1 ;
	Rtrace_terms[j].coef = apmInit( -1L, 0, BASE ) ;
	Rtrace_terms[j].factors = Rfpt ;

	newBapm( Rtrace_terms[j].bound, BASE ) ;

	fpt++ ;
	Rfpt++ ;
    }
/*
	Fix up the constant in the third term . . . it should be
	-2.0.
*/
    trace_terms[2].coef = -2.0 ;
    apmAssignLong( Rtrace_terms[2].coef, -2L, 0, BASE ) ;
/*
	Associate the factors - which are only pointers 
	to bounded objects - to genuine, properly initialized objects.
*/
    /* first term */
    b_trc.terms[0].factors[0] = &a_sin.bound ;
    Rb_trc.terms[0].factors[0] = &Ra_sin.bound ;

    /* second term */
    b_trc.terms[1].factors[0] = &b_sin.bound ;
    Rb_trc.terms[1].factors[0] = &Rb_sin.bound ;

    /* third term */
    b_trc.terms[2].factors[0] = &c_sin.bound ;
    Rb_trc.terms[2].factors[0] = &Rc_sin.bound ;
}
/* ++++++++++++++++++++++++ */

void 
bound_btrace (Bdd_dbl *result, Prism *priz)
{
	/* Bound the expression for the trace */
    bound_expr( &b_trc ) ;
# if DEBUG
    printf( "trace = %.4f %.4f \n", b_trc.bound.ub, b_trc.bound.lb ) ;
# endif
    result->ub = b_trc.bound.ub ;
    result->lb = b_trc.bound.lb ;
}
/* ++++++++++++++++++++++++++++++++++++++++++++ */

void 
Rbound_btrace (Bdd_apm *result, RPrism *priz)
/*
	An APM partner to bound_btrace.  Some of the variables 
	used here are defined above.
*/
{
	/* Bound the expression */
    Rbound_expr( &Rb_trc ) ;
    apmCalc( Rb_trc.bound.ub, Rb_trc.bound.ub, max_error, APM_ADD, NULL ) ;
    apmCalc( Rb_trc.bound.lb, Rb_trc.bound.lb, max_error, APM_SUB, NULL ) ;

# if DEBUG
    printf( "Rtrace = %.4f %.4f\n", apmtodbl( Rb_trc.bound.ub ),
				    apmtodbl( Rb_trc.bound.lb ) ) ;
# endif
    apmAssign( result->ub, Rb_trc.bound.ub ) ;
    apmAssign( result->lb, Rb_trc.bound.lb ) ;
}
