# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <math.h>

# include "apm.h"
# include "converse.h"
# include "tree.h"

# define   DEBUG   NO 

# define   LEFT   -1
# define   HERE    0
# define   RIGHT   1

# define   EQUIVS  2
# define   final_color( c )    ( (((c)==NO_TORI) || ((c)==MAYBE)) ? YES : NO )

int	    equiv_class[EQUIVS][MAT_DIM] = {
					{0, 1, 2, 3, 4, 5, 6 },
					{1, 0, 2, 3, 4, 6, 5 } } ;
char	     *LRcorner[2] ;
PrismLeaf    tree, equiv_leaves[EQUIVS] ;

/* Functions private to this file */
static PrismLeaf *seekLeaf(PrismLeaf *lpt);
static int leafCmpr(PrismLeaf *lapt, PrismLeaf *lbpt);
static void loadLeaf(PrismLeaf *lpt, RPrism *rpt);
static void printLeaf(PrismLeaf *lpt);
static void printTree(PrismLeaf *tpt);
static void makeEquivs(PrismLeaf *lpt);
static int LowerRight(char **result, char **cuts);
static int cutCmpr(char *capt, char *cbpt);
/* +++++++++++++++++++++++++++++++++++ */

# if FANCY_TREE

int 
consultTree (RPrism *rpt)
{
    int		shade, j ;
    PrismLeaf   *lpt, copy ;

    loadLeaf( &copy, rpt ) ;
    makeEquivs( &copy ) ;

/*
	Run through the list of equivalent leaves and 
	see what their colors are. If any is an interesting, final
	color set all the rest to the same shade.
*/
    shade = UNTRIED ;
    for( j=0 ; (j < EQUIVS) && (final_color( shade ) == NO) ; j++ )  {
		lpt = seekLeaf( &equiv_leaves[j]) ;
		shade = lpt->color ;
    }

    if( final_color( shade ) == YES )  {
		for( j=0 ; j < EQUIVS ; j++ )  {
			lpt = seekLeaf( &equiv_leaves[j] ) ;
			lpt->color = shade ;
		}
    }

    return( shade ) ;
}
/* ++++++++++++++++++++++++++++++++++ */

void 
initTree (void)
{
    int j ;

    tree.color = UNTRIED ;

    for( j=0 ; j < MAT_DIM ; j++ ) 
	tree.cuts[j] = "." ;

    tree.left = NULL ;
    tree.right = NULL ;
}
/* +++++++++++++++++++++++++++++++++ */

PrismLeaf 
*seekLeaf (PrismLeaf *lpt)
{
    int		j, branch ;
    PrismLeaf   *tpt ;

    tpt = &tree ;
    while( (branch = leafCmpr( lpt, tpt )) != HERE ) { 
		if( branch == LEFT ) {
			if( tpt->left == NULL ) {
				tpt->left = conjureLeaf() ;
			if( tpt->left == NULL ) {
				fprintf( stderr, "Couldn't turn a new leaf. Died.\n" ) ;
				cease() ;
			}

			tpt->left->color = lpt->color ;
			for( j=0 ; j < MAT_DIM ; j++ ) 
				strcpy( tpt->left->cuts[j], lpt->cuts[j] ) ;
			}

			tpt = tpt->left ;
		}
		else {
			if( tpt->right == NULL ) {
				tpt->right = conjureLeaf() ;
				if( tpt->right == NULL ) {
					fprintf( stderr, "Couldn't turn a new leaf. Died.\n" ) ;
					cease() ;
				}

				tpt->right->color = lpt->color ;
				for( j=0 ; j < MAT_DIM ; j++ ) 
					strcpy( tpt->right->cuts[j], lpt->cuts[j] ) ;
			}

			tpt = tpt->right ;
		}
    }

    return( tpt ) ;
}
/* ++++++++++++++++++++++++++++++ */

int 
leafCmpr (PrismLeaf *lapt, PrismLeaf *lbpt)
{
   int   answer, j ;

/*
    Check and see whether two leaves have the same
    cut histories.
*/
    answer = 0 ;
    for( j=0 ; (j < MAT_DIM) && (answer == 0) ; j++ )
	answer  =  cutCmpr( lapt->cuts[j], lbpt->cuts[j] ) ;

    if( answer == 0 )   return( HERE ) ;
    if( answer <  0 )   return( LEFT ) ;
    if( answer >  0 )   return( RIGHT ) ;
}
/* ++++++++++++++++++++++++++++++++++++ */

void 
colorLeaf (RPrism *rpt)
{
    PrismLeaf    copy, *equiv_leaf ;

    loadLeaf( &copy, rpt ) ;
    equiv_leaf = seekLeaf( &copy ) ;

    equiv_leaf->color = rpt->in_torus ;
}
/* ++++++++++++++++++++++++++++++++++ */

void 
loadLeaf (PrismLeaf *lpt, RPrism *rpt)
{
    int j ;

    lpt->color = rpt->in_torus ;

    for( j=0 ; j < MAT_DIM ; j++ )
		lpt->cuts[j] = rpt->cuts[j] ;
}
/* +++++++++++++++++++ */

void 
printLeaf (PrismLeaf *lpt)
{
    int  offset, j ;

    offset = 0 ;
    for( j=5 ; j < MAT_DIM ; j++ )
		offset += strlen( lpt->cuts[j] ) - 1 ;
		
    for( j=0 ; j < offset ; j++ ) 
		printf( "\t" ) ;

    for( j=5 ; j < MAT_DIM ; j++ ) 
		printf( "%s ", lpt->cuts[j] ) ;

    printf( ": %d \n", lpt->color ) ;
}
/* +++++++++++++++++++++++++++ */

int 
printTree (PrismLeaf *tpt)
{
    if( tpt != NULL ) {
	printTree( tpt->left ) ;
	printLeaf( tpt ) ;
	printTree( tpt->right ) ;
    }
}
/* +++++++++++++++++++++++++++++++++++ */

void 
makeEquivs (PrismLeaf *lpt)
{
    int    class, j ;

    for( class=0 ; class < EQUIVS ; class++ ){
		equiv_leaves[class].color = lpt->color ;
		for( j=0 ; j < MAT_DIM ; j++ )
			equiv_leaves[class].cuts[j] = lpt->cuts[ equiv_class[class][j] ] ;
    }
}
/* ++++++++++++++++++++++++++++ */

# else

int 
consultTree (RPrism *priz)
{
    int   at_edge ;

    at_edge = LowerRight( LRcorner, priz->cuts ) ;
    if( at_edge == YES ) {
# if DEBUG
	printf( "At an edge.\n" ) ;
# endif
	return( MUST_DO ) ;
    }

    else if( cutCmpr( LRcorner[0], LRcorner[1] ) > 0 )
	return( MUST_DO ) ;
    else
	return( MAY_SKIP ) ;
}
/* +++++++++++++++ */

int 
LowerRight (
/*
	Get a pair of strings which describe the lower right corner
	of the prism.  This involves certain manipulations on the
	strings which represent the center.  The function also 
	returns either YES or NO according to whether the the 
	lower right corner is on an edge of the whole fundamental domain.
*/
    char **result,
    char **cuts
)
{
    int   horiz_len, vert_len ;
    char  *hpt, *vpt ;

# if DEBUG
    printf( "LowerRight : %s %s \n", cuts[5], cuts[6] ) ;
# endif

    horiz_len = strlen( cuts[5] ) ;
    vert_len = strlen( cuts[6] ) ;

    strcpy( result[0], cuts[5] ) ;
    strcpy( result[1], cuts[6] ) ;

/*
	Horizontal part.  If the center's horizontal string ends in
	an "l", all is easy; the rectangle is a left-horizontal
	daughter of a parent whose center lies on the same vertical line
	as the lower right corner of our rectangle.  If the last character
	is an "r" we must run backward through the string until we find an
	"l".  If none occurs the right edge of our rectangle lies on the
	vertical line theta0 = two_pi.
*/
    hpt = result[0] + horiz_len - 1 ;
    if( *hpt == 'l')
	*hpt = '\0' ;
    else while( *hpt == 'r' )
	*hpt-- = '\0' ;
    if( *hpt == '.' )
	return( YES ) ;
    else
	*hpt = '\0' ;
/*
	Vertical part.  Same as the horizontal except that the
	roles of "l" and "r" are interchanged because we want a 
	lower bound on this coordinate.
*/
    vpt = result[1] + vert_len - 1 ;
    if( *vpt == 'r')
	*vpt = '\0' ;
    else while( *vpt == 'l' )
	*vpt-- = '\0' ;
    if( *vpt == '.' )
	return( YES ) ;
    else
	*vpt = '\0' ;

# if DEBUG
    printf( "corner at %s %s \n\n", result[0], result[1] ) ;
# endif

    return( NO ) ;
}
/* +++++++++++++++ */

void 
initTree (void)
{
    LRcorner[0] = malloc( 4 * (depth + 2 ) ) ;
    if( LRcorner[0] == NULL ) {
		fprintf( stderr, 
			"Can't allocate strings for consultTree(). Died. \n" ) ;
		cease() ;
    }
    else
		LRcorner[1] = LRcorner[0] + 2 * (depth + 2) ;
}
# endif
/* ++++++++++++++++++++ */

int 
cutCmpr (char *capt, char *cbpt)
{
/*  find the first point at which the strings disagree */

    while( (*capt == *cbpt) && (*capt) && (*cbpt) ){
	capt++ ;
	cbpt++ ;
    }

    if( *capt ) {
	if ( *cbpt )
	    return( *capt - *cbpt ) ;
	else {
	    if( *capt == 'r' )
		return( 1 ) ;
	    else
		return( -1 ) ;
	}
    }
    else {
	if( *cbpt ) {
	    if( *cbpt == 'r' )
		return( -1 ) ;
	    else
		return( 1 ) ;
	}
	else
	    return( 0 ) ;
    }
}
