# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <math.h>

# include "apm.h"
# include "apmUtil.h"
# include "apmSpecial.h"
# include "apmPrint.h"
# include "converse.h"

# define  DEBUG			NO 
# define  IN_BUF_SZ		20

/* Variables private to this file */
static APM	u_scratch ;
static char	buf[BUF_SZ], trunc_buf[BUF_SZ], mant_str[DBL_DIGITS+3] ; 
static char	*lastParmStrs[N_PARMS] ;
static char	*verdict_str[] = { "no tori", "untried", "maybe", "still checking" } ;
/* +++++++++++++++++++++++++++++ */

void 
fprintRPrism (FILE *stream, RPrism *priz, int nsteps)
{
    int  j ;

    for( j=0 ; j < MAT_DIM ; j++ )
	fprintf( stream, "%s  ", priz->cuts[j] ) ;

    fprintf( stream, "%d  steps give : %s.\n", nsteps, 
					       verdict_str[priz->in_torus] ) ;
    fflush( stream ) ;
}
/* +++++++++++++++++++++++++++ */

void 
printRPrism (RPrism *priz, int nsteps)
{
     fprintRPrism( stdout, priz, nsteps ) ;
}
/* +++++++++++++++++++++ */

void 
copyRPrism (Prism *ppt, RPrism *Rpt)
{
    int     j ;
    APM	    *apt ;
    double  *mpt, *end_mat ;

    ppt->in_torus = Rpt->in_torus ;	/* Copy the integer attributes. */
    ppt->n_cuts = Rpt->n_cuts ;

    for( j=0 ; j < MAT_DIM ; j++ )		/* Copy the cut history */
	strcpy( ppt->cuts[j], Rpt->cuts[j] ) ;

    for( j=0 ; j < N_PARMS ; j++ )		/* Copy the center */
	ppt->center->p[j] = apmtodbl( Rpt->center->p[j] ) ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	ppt->center->z.u[j] = apmtodbl( Rpt->center->z.u[j] ) ;
	ppt->center->z.v[j] = apmtodbl( Rpt->center->z.v[j] ) ;
    }

    mpt = ppt->matrix ;				/* Copy the matrix */
    apt = Rpt->matrix ;
    for( end_mat = mpt + MAT_SZ ; mpt < end_mat ;  )
	*mpt++ = apmtodbl( *apt++ ) ;
}
/* ++++++++++++++++++++++++ */

void 
copyPrism (Prism *a_ppt, Prism *b_ppt)
{
    int     j ;
    double  *ampt, *bmpt, *end_amat ;

    a_ppt->in_torus = b_ppt->in_torus ;	/* Copy the integer attributes. */
    a_ppt->n_cuts = b_ppt->n_cuts ;

    for( j=0 ; j < MAT_DIM ; j++ )		/* Copy the cut history */
	strcpy( a_ppt->cuts[j], b_ppt->cuts[j] ) ;

    for( j=0 ; j < N_PARMS ; j++ )		/* Copy the center */
		a_ppt->center->p[j] = b_ppt->center->p[j] ;
		for( j=0 ; j < DEG_FREE ; j++ ) {
			a_ppt->center->z.u[j] = b_ppt->center->z.u[j] ;
			a_ppt->center->z.v[j] = b_ppt->center->z.v[j] ;
    }

    ampt = a_ppt->matrix ;			/* Copy the matrix */
    bmpt = b_ppt->matrix ;
    for( end_amat = ampt + MAT_SZ ; ampt < end_amat ;  )
	*ampt++ = *bmpt++ ;
}
/* +++++++++++++++++++++ */

void 
RcopyRPrism (RPrism *Rapt, RPrism *Rbpt)
{
    int     j ;
    APM	    *apt, *bpt, *end_mat ;

    Rapt->in_torus = Rbpt->in_torus ;	/* Copy the integer attributes. */
    Rapt->n_cuts = Rbpt->n_cuts ;

    for( j=0 ; j < MAT_DIM ; j++ )		/* Copy the cut history */
	strcpy( Rapt->cuts[j], Rbpt->cuts[j] ) ;

    for( j=0 ; j < N_PARMS ; j++ )		/* Copy the center */
	apmAssign( Rapt->center->p[j], Rbpt->center->p[j] ) ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	apmAssign( Rapt->center->z.u[j], Rbpt->center->z.u[j] ) ;
	apmAssign( Rapt->center->z.v[j], Rbpt->center->z.v[j] ) ;
    }
    
    apt = Rapt->matrix ;			/* Copy the matrix */
    bpt = Rbpt->matrix ;
    for( end_mat = bpt + MAT_SZ ; bpt < end_mat ;  )
	apmAssign( *apt++, *bpt++ ) ;
}
/* ++++++++++++++++++++++++ */

void 
read_parameters (RPrism *first)
{
    int     j ;
    APM     *mpt ;
    char    p_center[IN_BUF_SZ], p_width[IN_BUF_SZ] ;
    char    v_center[IN_BUF_SZ], v_width[IN_BUF_SZ] ;

    mpt = first->matrix ;
    for( j=0 ; j < N_PARMS ; j++ ) {
	scanf("%s %s", p_center, p_width ) ;
	apmAssignString( first->center->p[j], p_center, BASE ) ;
	apmAssignString( *mpt, p_width, BASE ) ;

	mpt += MAT_DIM + 1 ;
    }

    mpt = first->matrix + STAID_LEN + (DEG_FREE * MAT_DIM) + 
					N_PARMS + DEG_FREE ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
	scanf( "%s %s", v_center, v_width ) ;

	apmAssignString( u_scratch, v_center, BASE ) ;
	apmCalc( u_scratch, u_scratch, max_error, APM_ADD, NULL ) ;
	apmMultiply( first->center->z.v[j], u_scratch, pi ) ;

	apmAssignString( u_scratch, v_width, BASE ) ;
	apmCalc( u_scratch, u_scratch, max_error, APM_ADD, NULL ) ;
	apmMultiply( *mpt, u_scratch, pi ) ;

	mpt += MAT_DIM + 1 ;
    }
}
/* +++++++++++++++++++++++++++ */	

void 
echo_parameters (RPrism *first)
{
    int     j ;
    APM     *mpt ;
    char	comment_buf[BUF_SZ] ;
    double  center, width ;

    mpt = first->matrix ;
    printf( "Parameters : \n" ) ;
    for( j=0 ; j < N_PARMS ; j++ ) {
		center = apmtodbl( first->center->p[j] ) ;
		width = apmtodbl( *mpt ) ; 
		printf("%s : %.14e \t %.14e \n", parm_names[j], 
						 center, width ) ;
		mpt += MAT_DIM + 1 ;
    }

    mpt = first->matrix + STAID_LEN + MAT_DIM * DEG_FREE + 
					N_PARMS + DEG_FREE ;
    printf( "\nInitial region : \n" ) ;
    for( j=0 ; j < DEG_FREE ; j++ ) {
		center = apmtodbl( first->center->z.v[j] ) ;
		width = apmtodbl( *mpt ) ; 
		printf("v[%d] : %.14e \t %.14e \n", j, center, width ) ;

		mpt += MAT_DIM + 1 ;
    }

    printf( "\nComments :" ) ;
    if( restoration == NO ) {
		while( fgets(comment_buf, BUF_SZ, stdin) != NULL ) {
			printf( "%s", comment_buf ) ;
		}
    }
    else {
		printf( "restarted from %s.\n", rest_name ) ;
	}

    printf( "\n\n++++++++++++++++++++++++++++++++++++++++++++++++\n\n" ) ;
    fflush( stdout ) ;
}
/* +++++++++++++++++++++++ */

int 
errHndlr (int errcode, char *message, char *file, int line, char *function)
{
    fprintf( stderr, "%s : %s in %s at line %d. \n", function, message,
							   file, line ) ;
    return( errcode ) ;
}
/* +++++++++++++++++++++++ */

void 
printMat (double *mat)
{
    int	    row ;
    double  *mpt, *end_row ;

    mpt = mat ;
    for( row=0 ; row < MAT_DIM ; row++ )  {
	for( end_row = mpt + MAT_DIM ; mpt < end_row ; mpt++ )
	    printf( "%.4f\t", *mpt ) ;

	printf( "\n" ) ;
    }

    printf( "\n" ) ;
}
/* +++++++++++++++++++++++ */

void 
printRmat (APM *mat)
{
    int	    row ;
    APM     *mpt, *end_row ;

    mpt = mat ;
    for( row=0 ; row < MAT_DIM ; row++ )  {
	for( end_row = mpt + MAT_DIM ; mpt < end_row ; mpt++ ) {
	    apmConvert( buf, BUF_SZ, (short)4, DO_ROUND, LEFT_JUST, *mpt ) ;
	    printf( "%s\t", buf ) ;
	}
	printf( "\n" ) ;
    }

    printf( "\n" ) ;
}
/* ++++++++++++++++++++++++++ */

void 
initUtil (void)
{
    int   j ;

    u_scratch = apmNew( BASE ) ;

    lastParmStrs[0] = malloc( (depth + 2) * N_PARMS ) ;
    if( lastParmStrs[0] == NULL ) {
	fprintf( stderr, 
		 "initUtil : Can't allocate space for parameter strings. \n" ) ;
	cease() ;
    }

    for( j=1 ; j < N_PARMS ; j++ )
	lastParmStrs[j] = lastParmStrs[j-1] + depth + 2 ;
     
/* 	
	Take care to initialize the strings so that the first
	call to areNewParms() will return YES.
*/
    for( j=0 ; j < N_PARMS ; j++ ) 
	strcpy( lastParmStrs[j], "?" ) ;
}
/* +++++++++++++++++++++++++ */

void 
printXpt (Xtnd_pt *xpt)
{
    int	     j ;
    double   *cpt ;

/*	Print the parameters 	*/
    printf( "{ ( " ) ;
    cpt = xpt->p ;
    for( j=0 ; j < N_PARMS ; j++ )
	printf( "%.4f  ", *cpt++ ) ;
    printf( "), " )  ;

/*	Print the components of u	*/
    printf( "( " ) ;
    cpt = xpt->z.u ;
    for( j=0 ; j < DEG_FREE ; j++ )
	printf( "%.4f  ", *cpt++ ) ;
    printf( "), " )  ;

/*	Print the components of v	*/
    printf( "( " ) ;
    cpt = xpt->z.v ;
    for( j=0 ; j < DEG_FREE ; j++ )
	printf( "%.4f  ", *cpt++ ) ;
    printf( ") } \n" )  ;
}
/* +++++++++++++++++++++++++ */

void 
printRXpt (RXtnd_pt *xpt)
{
    int	     j ;
    APM      *cpt ;

/*	Print the parameters 	*/
    printf( "{ ( " ) ;
    cpt = xpt->p ;
    for( j=0 ; j < N_PARMS ; j++ )
	printf( "%.4f  ", apmtodbl(*cpt++) ) ;
    printf( "), " )  ;

/*	Print the components of u	*/
    printf( "( " ) ;
    cpt = xpt->z.u ;
    for( j=0 ; j < DEG_FREE ; j++ )
	printf( "%.4f  ", apmtodbl(*cpt++) ) ;
    printf( "), " )  ;

/*	Print the components of v	*/
    printf( "( " ) ;
    cpt = xpt->z.v ;
    for( j=0 ; j < DEG_FREE ; j++ )
	printf( "%.4f  ", apmtodbl(*cpt++) ) ;
    printf( ") } \n" )  ;
}
/* ++++++++++++++++++++++++++++ */

void 
truncateRPrism (RPrism *priz, int dp)
{
    int   j ;
    APM   *mpt, *end_mat ;

    for( j=0 ; j < N_PARMS ; j++ ) 
	apmTruncate( priz->center->p[j], dp ) ;

    for( j=0 ; j < DEG_FREE ; j++ ) {
	apmTruncate( priz->center->z.u[j], dp ) ;
	apmTruncate( priz->center->z.v[j], dp ) ;
    }

    end_mat = priz->matrix + MAT_SZ ;
    for( mpt = priz->matrix, j=0 ; j < N_PARMS ; j++, mpt += MAT_DIM + 1 )
	apmTruncate( *mpt, dp ) ;
    for( mpt -= N_PARMS ; mpt < end_mat ; mpt++ )
	apmTruncate( *mpt, dp ) ;

# if DEBUG
    fflush( stdout ) ;
    fflush( stderr ) ;
# endif
}
/* ++++++++++++++++++ */

int 
areNewParms (RPrism *priz)
{
    int  j, answer ;

    answer = NO ;
    for( j=0 ; j < N_PARMS ; j++ ) {
		if( strcmp( lastParmStrs[j], priz->cuts[j] ) != 0 ) {
			strcpy( lastParmStrs[j], priz->cuts[j] ) ;
			answer = YES ;
		}
    }
    
    return( answer ) ;
}
