# include <stdio.h>
# include <math.h>

# include "apm.h"
# include "apmSpecial.h"
# include "converse.h"
# include "bounding.h"
# include "rows.h"
# include "pi.h"

# define  DEBUG	    NO	

/* Variables shared through converse.h */
APM		Rsqrt_disc ;
APM		Rdf_sq, Rdf ;
APM		Rmax_slope, Rmin_slope, Rfirst_slope ;
double	max_slope, min_slope, first_slope ;

/* Variables private to this file */
static APM	lip_scratch, Rdscrm ;

static RPrism	*earliest ;
static Bdd_apm	Rmax_btrace, Rmin_btrace, Rfirst_btrace ;
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void initLip (void)
{
/*
    This function depends in detail on the choice of map.
*/

/*
	APM stuff
*/
    Rmin_slope = apmNew( BASE ) ;	/*  The external APMs */
    Rmax_slope = apmNew( BASE ) ;
    Rfirst_slope = apmNew( BASE ) ;
    Rdf = apmInit( (long)(DEG_FREE), 0, BASE ) ;
    Rdf_sq = apmInit( (long)(DF_SQ), 0, BASE ) ;
    Rstart_size = apmInit( 1L, -START_SZ, BASE ) ;

    Rdscrm = apmNew( BASE ) ;
    Rsqrt_disc  = apmNew( BASE ) ;
    lip_scratch = apmNew( BASE ) ;

    newBapm( Rmax_btrace, BASE ) ;
    newBapm( Rmin_btrace, BASE ) ;
    newBapm( Rfirst_btrace, BASE ) ;

    earliest = conjureRPrism() ;
}
/* +++++++++++++++++++++++ */

void 
setCone (RPrism *priz)
/*
	Get the minimum and maximum values for the
	trace of the slope object.  Note that we 
	exploit the symmetry of the potential; the minimum
	and maximum values of the trace of (beta - 2I) have
	the same absolute value.
*/
{
    int  j ;
    APM  *mat_pos ;

    for( j=0 ; j < N_PARMS ; j++ ) 
	apmAssign( earliest->center->p[j], priz->center->p[j] ) ;

    for( j=0 ; j < DEG_FREE ; j++ ) {
	apmAssign( earliest->center->z.v[j],  priz->center->z.u[j] ) ;

    }

    Rglobal_bounds( earliest ) ;
    Rbound_btrace( &Rmin_btrace, earliest ) ;
/*
	Account for the imprecision of the starting point 
	and the variation of the parameters.
*/
    apmAssignLong( lip_scratch, 0L, 0, BASE ) ;
    mat_pos = priz->matrix ;

    for( j=0 ; j < N_PARMS ; j++ ) {
	apmCalc( lip_scratch, lip_scratch, 
			      priz->center->p[j] , Rstart_size,
			      APM_MUL, APM_ADD,
			      *mat_pos,
			      APM_ABS, APM_ADD, NULL ) ;
	mat_pos += 1 + MAT_DIM ;
    }

    apmCalc( Rmin_btrace.lb, Rmin_btrace.lb, lip_scratch, 
			     APM_SUB, NULL ) ;
    apmCalc( Rmin_btrace.ub, Rmin_btrace.ub, lip_scratch, 
			     APM_ADD, NULL ) ;

	/* exploit the symmetry */
    apmSubtract( Rmax_btrace.ub, eight, Rmin_btrace.lb ) ;
    apmSubtract( Rmax_btrace.lb, eight, Rmin_btrace.ub ) ;

    apmCalc( Rdscrm,  Rmax_btrace.lb, APM_DUP, APM_MUL,
			four, Rdf_sq, APM_MUL, APM_SUB, NULL ) ;

    apmSqrt( Rsqrt_disc, precision, Rdscrm ) ;
# if DEBUG
	printf( "sqrt_disc : %.6f,  %.6f \n", apmtodbl( Rsqrt_disc ),
					sqrt( apmtodbl( Rdscrm ) ) ) ;
# endif 
    apmAdd( lip_scratch, Rmax_btrace.lb, Rsqrt_disc ) ;
    apmDivide( Rmax_slope, precision, (APM)NULL, lip_scratch, two ) ;

    apmSubtract( lip_scratch, Rmax_btrace.lb, Rsqrt_disc ) ;
    apmDivide( Rmin_slope, precision, (APM)NULL, lip_scratch, two ) ;

    min_slope = apmtodbl( Rmin_slope ) ;
    max_slope = apmtodbl( Rmax_slope ) ;

# if DEBUG
    printf( "setCone : %.6f < trace < %.6f \n", min_slope, max_slope ) ;
# endif
}
/* +++++++++++++++++++++++++++++++++++++++++++ */

void 
setSlopes (RPrism *priz)
/*
	Recall that our orbit will, at the beginning of
	a round of orbit-following, have just passed through a
	point on the torus whose beta will diminish the 
	slope. This implies that the slope is already smaller 
	than the value of max_slope found above.  Calculate
	a better upper bound on what the slope could be and
	store it in first_slope and Rfirst_slope.
*/
{
    int  j, mat_pos ;

    for( j=0 ; j < N_PARMS ; j++ )  {
	apmAssign( earliest->center->p[j], priz->center->p[j] ) ;

	mat_pos = j * (MAT_DIM + 1) ;
	apmAssign( earliest->matrix[mat_pos], priz->matrix[mat_pos] ) ;
    }

    for( j=0 ; j < DEG_FREE ; j++ ) {
	apmAssign( earliest->center->z.v[j], priz->center->z.u[j] ) ;

	/* 
		Account for imprecision in the starting point.
 	*/
	mat_pos = STAID_LEN + TWO_DF*MAT_DIM + 
		  N_PARMS + DEG_FREE + j * (MAT_DIM + 1) ;
	apmAssign( earliest->matrix[mat_pos], Rstart_size ) ;
    }
    
    Rglobal_bounds( earliest ) ;
    Rbound_btrace( &Rfirst_btrace, earliest ) ;
    
    apmDivide( lip_scratch, precision, (APM)NULL, Rdf_sq, Rmax_slope ) ;
    apmCalc( Rfirst_slope, Rfirst_btrace.ub, lip_scratch, APM_SUB,
					    max_error, APM_ADD, NULL ) ;

    first_slope = apmtodbl( Rfirst_slope ) + DBL_ERR ;

# if DEBUG
    printf( "setSlope : first_slope = %.6f \n", first_slope ) ;
# endif
}
